# ===================================================================================
#                            meshconverter.py                                       |
# ===================================================================================
#                                                                                   |
# Programmer:     Sandy Wilson (KnightErrant)                                       |
#                                                                                   |
# Creation Date:  16 February 2007                                                  |
# Revision Dates: 16 February 2007 -                                                |
#                 23 February 2007 - haven't kept up with all the changes. At this  |
#                                    point the script makes it all the way through  |
#                                    the file and prints everything out to a .txt   |
#                                    file for study.  It also outputs a file in     |
#                                    .obj format that Casuir has been able to       |
#                                    import and skin in 3Dexplorer.  I've been able |
#                                    to import this into Milkshape 3D but haven't   |
#                                    skinned it yet.  The script accumulates data   |
#                                    into three lists: trianglelist for the faces,  |
#                                    vertexdatalist for the uv coordinates, vertex  |
#                                    weights, vertex vectors, and vertex bone group |
#                                    assignments, and finally bonestringslist for   |
#                                    a list of the ASCII bone names and hierarchy   |
#                                    numbers, the model name, and the bounding      |
#                                    sphere floats.                                 |
#                                                                                   |
#                 07 March 2007    - Radically redid all the header/footer stuff    |
#                                    to split them where it seems reasonable.  Also |
#                                    updated to read the Aztec variant mesh that    |
#                                    Casuir found.  This has made this code a real  |
#                                    spaghetti mess so hope to clean it up          |
#                                    sometime when I get a chance.                  |
#                                                                                   |
#                 08 March 2007    - Included conversion from .mesh to .ms3d but    |
#                                    this meant hard coding the hierarchy tree      |
#                                    the .ms3d wants.                               |
#                                                                                   |
#                 09 March 2007    - Got the conversion from .ms3d back to .mesh    |
#                                    done. Must test it tonight.                    |
#                                                                                   |
#                 18 March 2007    - Removed hard-coded skeleton and put in ASCII   |
#                                    skeleton files.  Seems to be a standard        |
#                                    skeleton and one for horse, camel, and         |
#                                    elephant.  Put boundingsphere in the model     |
#                                    comments. Also included the bone strings as    |
#                                    they come out of the mesh including char       |
#                                    counts and tree index(?).                      |
#                                                                                   |
#                 19 March 2007    - Changed bone strings write out to be optional. |
#                                    GOM doesn't use them in his .ms3d format.      |
#                                    He includes materials but I'll skip this.      |
#                                    Rewrote to process or discard any section so   |
#                                    I can process any .ms3d file (hopefully)       |
#                                    without crashing.                              |
#                                                                                   |
#                 20 March 2007    - Tested on multiple units.  Wrote ROADMAP       |
#                                    comments below on how the program is           |
#                                    structured.  Will test more tonight.           |
#                                                                                   |
#                 17 April 2007    - Missed the 4 0 0 0 int after the weights       |
#                                    for variant siege engine meshes.  Put in       |
#                                    bounding sphere data.                          |
#                                                                                   |
#                 18 April 2007    - Switched UV coords to match the triangle       |
#                                    swap.                                          |
#                                                                                   |
#                 15 May   2007    - Did this earlier but forgot to comment.        |
#                                    Wrote out an extra int for subversion 2 at     |
#                                    the secondary bone Id assignment and vertex    |
#                                    weighting.  Also read it back in from .ms3d    |
#                                    to .mesh back conversion.                      |
#                                                                                   |
#                 15 June  2007    - Modified the hard-coding of the human          |
#                                    standard skeleton to allow for extra bones     |
#                                    before the weapon/shield bones.  This only     |
#                                    concerns the repeating node numbers.           |
#                                                                                   |
#                                                                                   |
# NOTES: (1)  There are still discrepancies between CA's normal vectors and those   |
#             computed by this program.  Can't see this in-game so not worrying     |
#             about it.  There are also runs of zero vectors in the tangents and    |
#             binormals.                                                            |
#                                                                                   |
#        (2)  Very important to remember:  x -> -x and flip two triangle indices    |
#             to get .mesh and Milkshape to play nicely together.                   |
#                                                                                   |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# A forwards and backwards mesh converter.  Converts .mesh files to .ms3d format    |
# and .ms3d format files back into .mesh files.                                     |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
#                     ROADMAP THROUGH THIS SPAGHETTI CODE                           |
#                                                                                   |
# .mesh to .ms3d:                                                                   |
#                                                                                   |
#   meshtoms3dconverter()      - called from main if user clicks on a .mesh file.   |
#                                                                                   |
#   Reading data out of the .mesh file is done by these routines:                   |
#                                                                                   |
#   readmainheader()           - reads header and gets num_groups of triangles.     |
#   processfirstpart()         - reads irregular first triangle group.              |
#   processregularpart()       - reads rest of triangle groups.                     |
#   processtexturecoordinates()- reads uv coordinates.                              |
#   processvertexweights()     - reads vertex weights.                              |
#   processvertexvectors()     - reads vertex vectors.                              |
#                                                                                   |
#   computetangentspacebasis() - stops and does this to get vertex normals to give  |
#                                to .ms3d file.  Thought we needed the reference    |
#                                counts but it works with 1's.                      |
#                                                                                   |
#   processvertexbones()       - bone assignments, primary and secondary.           |
#   processmysteryblock()      - actually the tangent basis in RGB encoding.        |
#                                a file with a _lod0 suffix has normals, tangents,  |
#                                and binormals, lod1, lod2, or lod3 has only the    |
#                                normals.  This is hard coded into the LOD0FLAG.    |
#   processbonestrings()       - gets bones strings and node indices (?) as well as |
#                                characterlod0 string and the bounding sphere       |
#                                floats and works through the final granny block.   |
#                                                                                   |
#   All the data obtained by the above are saved in global Python lists:            |
#       trianglelist           - all the triangle group types, names, and triangle  |
#                                data arrays.                                       |
#       vertexdatalist         - uvcoords   - nvert x 2 floats, uv coordinates      |
#                                vwts       - nvert x 2 floats, vertex weights      |
#                                vvecs      - nvert x 3 floats, vertex vectors      |
#                                vbones     - nvert x 4 bytes, bone assignments     |
#       bonestringslist        - bones names and bounding sphere data.              |
#                                                                                   |
#                                                                                   |
#   Writing data to the .ms3d format is done by:                                    |
#                                                                                   |
#   writems3dfile()            - uses the global Python lists to write the data     |
#                                appropriately.                                     |
#   writeskeleton2ms3dfile()   - reads the appropriate external ASCII skeleton      |
#                                file to put the joints into Mete's ms3d format.    |
#                                These external files have parent/child relations   |
#                                as well as position and rotation information along |
#                                with two time values (set to 1.0).                 |
#                                                                                   |
#                                Currently we have the standard skeleton for units  |
#                                plus horse, camel, and elephant.  They are         |
#                                chosen if the input filename has                   |
#                                    "mount_camel", "mount_elephant", or "mount_"   |
#                                in them, if not, the standard skeleton is used.    |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# .ms3d to .mesh:                                                                   |
#                                                                                   |
#   ms3dtomeshconverter()      - called from main if user clicks on a .ms3d file.   |
#                                                                                   |
#   This is just one big long section of straight code.  It first pulls all the     |
#   data out of the .ms3d file, it computes normals, tangents, and binormals and    |
#   RGB converts them.  It then rearranges the data a little.  Then it opens a      |
#   .mesh file with "_converted" appended to it to prevent overwriting and just     |
#   writes the data out with hard-coded granny headers and footers.                 |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
#   There are lots of little getters, putters, and formatters to ease coding.       |
#   These helper defs or functions are found in the front part of this file.        |
#   After these the mesh reading functions begin.  These are followed by the        |
#   higher level routines mesh2ms3dconverter and ms3d2meshconverter.  There are     |
#   some obsolete defs still remaining that pertain to the .obj and .smd formats    |
#   as well as our diversion onto variant meshes.                                   |
#                                                                                   |
# ===================================================================================

# Imports. 
import os
import struct                  # This is for reading binary files.
import array                   # This is for arrays of data read from binary mesh file.
import string                  # String utilities, don't know if this is necessary.
import listdir                 # This is from the Python documentation, Just a GUI file chooser.
from Tkinter import *          # GUI support for the file chooser.

# Globals for counters and flags.
global bytecount
global ms3dbytecount
global WRITEDATA
global WRITEBYTECOUNT
global WRITEFOOTERS
global WRITESTRINGS
global LOD0FLAG      
global VARIANTMESHFLAG      
global TANGENTSPACEFLAG     

# Globals for data.
global num_groups
global trianglelist
global vertexdatalist
global tangentspacedatalist
global bonestringslist
global texturefilelist
global meshtrianglelist
global meshvertexdatalist
global meshbonestringslist
global variantboundingsphere

# Initialize globals.
WRITEDATA                  = True 
WRITEBYTECOUNT             = True 
WRITEFOOTERS               = True 
WRITESTRINGS               = True
TANGENTSPACEFLAG           = 2
trianglelist               = []
vertexdatalist             = []
bonestringslist            = []

# ===================================================================================
#                     Getters(), Putters(), and Formatting Utilities                |
# getbyte()                                                                         |
# getubyte()                                                                        |
# getshort()                                                                        |
# getushort()                                                                       |
# getint()                                                                          |
# getuint()                                                                         |
# getfloat()                                                                        |
# putbyte()                                                                         |
# putubyte()                                                                        |
# putshort()                                                                        |
# putushort()                                                                       |
# putint()                                                                          |
# putuint()                                                                         |
# putfloat()                                                                        |
# aby()              - Aligns an output byte value.                                 |
# ash()              - Aligns an output short value.                                |
# hexdigit()         - Returns hex representation of a 4-bit nibble.                |
# zipstrip()         - Removes trailing 0's from Mete's 32 byte character strings.  |
# rgb2f()            - Inverse cube normalization.                                  |
# f2rgb()            - Cube normalization.                                          |
# mag()              - Magnitude of float triple.                                   |
# formatfloattriples()  - Formats floats to a given width.                          |
# formatfloatvector()   - Formats float vectors and put out length of vector.       |
# formatbytetriples()- Formats bytes to a fixed width field.                        |
#                                                                                   |
# ===================================================================================

# -----------------------------------------------------------------------------------
def getbyte( fidin ) :
    global bytecount
    (thebyte,)             = struct.unpack( 'b', fidin.read(1) )
    bytecount              = bytecount + 1
    return thebyte

# -----------------------------------------------------------------------------------
def getubyte( fidin ) :
    global bytecount
    (thebyte,)             = struct.unpack( 'B', fidin.read(1) )
    bytecount              = bytecount + 1
    return thebyte

# -----------------------------------------------------------------------------------
def getshort( fidin ) :
    global bytecount
    (theshort,)            = struct.unpack( 'h', fidin.read(2) )
    bytecount              = bytecount + 2
    return theshort

# -----------------------------------------------------------------------------------
def getushort( fidin ) :
    global bytecount
    (theshort,)            = struct.unpack( 'H', fidin.read(2) )
    bytecount              = bytecount + 2
    return theshort

# -----------------------------------------------------------------------------------
def getint( fidin ) :
    global bytecount
    (theint,)              = struct.unpack( 'i', fidin.read(4) )
    bytecount              = bytecount + 4
    return theint

# -----------------------------------------------------------------------------------
def getuint( fidin ) :
    global bytecount
    (theint,)              = struct.unpack( 'I', fidin.read(4) )
    bytecount              = bytecount + 4
    return theint

# -----------------------------------------------------------------------------------
def getfloat( fidin ) :
    global bytecount
    (thefloat,)            = struct.unpack( 'f', fidin.read(4) )
    bytecount              = bytecount + 4
    return thefloat

# -----------------------------------------------------------------------------------
def putbyte( thebyte, fidout ) :
    global ms3dbytecount
    fidout.write( struct.pack( 'b', thebyte ) )
    ms3dbytecount          = ms3dbytecount + 1
    return  

# -----------------------------------------------------------------------------------
def putubyte( thebyte, fidout ) :
    global ms3dbytecount
    fidout.write( struct.pack( 'B', thebyte ) )
    ms3dbytecount          = ms3dbytecount + 1
    return  

# -----------------------------------------------------------------------------------
def putshort( theshort, fidout ) :
    global ms3dbytecount
    fidout.write( struct.pack( 'h', theshort ) )
    ms3dbytecount          = ms3dbytecount + 2
    return  

# -----------------------------------------------------------------------------------
def putushort( theshort, fidout ) :
    global ms3dbytecount
    fidout.write( struct.pack( 'H', theshort ) )
    ms3dbytecount          = ms3dbytecount + 2
    return  

# -----------------------------------------------------------------------------------
def putint( theint, fidout ) :
    global ms3dbytecount
    fidout.write( struct.pack( 'i', theint ) )
    ms3dbytecount          = ms3dbytecount + 4
    return  

# -----------------------------------------------------------------------------------
def putuint( theint, fidout ) :
    global ms3dbytecount
    fidout.write( struct.pack( 'I', theint ) )
    ms3dbytecount          = ms3dbytecount + 4
    return  

# -----------------------------------------------------------------------------------
def putfloat( thefloat, fidout ) :
    global ms3dbytecount
    fidout.write( struct.pack( 'f', thefloat ) )
    ms3dbytecount          = ms3dbytecount + 4
    return  

# -----------------------------------------------------------------------------------
def putzerobytes( n, fidout ) :
    global ms3dbytecount
    for ii in range(n) :
        putubyte( 0, fidout )
        ms3dbytecount      = ms3dbytecount + 1
    return  


# -----------------------------------------------------------------------------------
#    Right aligns a decimal byte into 4 spaces.
# -----------------------------------------------------------------------------------
def aby( thebyte ) :
    bytestring             = str( thebyte )
    if len( bytestring ) == 1 :
        bytestring         = '   ' + bytestring
    elif len( bytestring ) == 2 :
        bytestring         = '  ' + bytestring
    else :
        bytestring         = ' ' + bytestring

    return bytestring


# -----------------------------------------------------------------------------------
#    Right aligns a decimal byte into 3 spaces, or 4 if bigger than 99.
# -----------------------------------------------------------------------------------
def alt( thebyte ) :
    bytestring             = str( thebyte )
    if len( bytestring ) == 1 :
        bytestring         = '  ' + bytestring
    elif len( bytestring ) == 2 :
        bytestring         = ' ' + bytestring
    else :
        bytestring         = ' ' + bytestring

    return bytestring


# -----------------------------------------------------------------------------------
#    Right aligns a decimal short into 6 spaces.
# -----------------------------------------------------------------------------------
def ash( theshort ) :
    shortstring            = str( theshort )
    if len( shortstring ) == 1 :
        shortstring        = '     ' + shortstring
    elif len( shortstring ) == 2 :
        shortstring        = '    ' + shortstring
    elif len( shortstring ) == 3 :
        shortstring        = '   ' + shortstring
    elif len( shortstring ) == 4 :
        shortstring        = '  ' + shortstring
    else :
        shortstring        = ' ' + shortstring

    return shortstring


# -----------------------------------------------------------------------------------
#    Returns a hex representation of a nibble.  Returns a '-' if user inputs
#    something bigger than 15.
# -----------------------------------------------------------------------------------
def hexdigit( n ) :
    if n < 10 :
        dig                = str( n )
    elif n == 10 :
        dig                = 'A'
    elif n == 11 :
        dig                = 'B'
    elif n == 12 :
        dig                = 'C'
    elif n == 13 :
        dig                = 'D'
    elif n == 14 :
        dig                = 'E'
    elif n == 15 :
        dig                = 'F'
    else :
        dig                = '-'

    return dig


# -----------------------------------------------------------------------------------
#    Strips trailing \0's. (Mete's fault.)
# -----------------------------------------------------------------------------------
def zipstrip( string ) :
    nch                    = len( string )
    for ii in range( nch ) :
        (val,)             = struct.unpack( 'b', string[ii] )
        if val == 0 :
            break

    stripstring            = string[0:ii]

    return stripstring


# -----------------------------------------------------------------------------------
#    Convert RGB triples back to normal vectors through the inverse cube normalization.
# -----------------------------------------------------------------------------------
def rgb2f( R, G, B ) :
    x                      = float( R ) / 255.0
    y                      = float( G ) / 255.0
    z                      = float( B ) / 255.0
    x                      = 2.0 * x - 1.0
    y                      = 2.0 * y - 1.0
    z                      = 2.0 * z - 1.0

    triple                 = [ x, y, z ]

    return triple


# -----------------------------------------------------------------------------------
#    Convert float normal vectors into RGB triples by cube normalization.
# -----------------------------------------------------------------------------------
def f2rgb( x, y, z ) :
    nx                     = 255.0 * 0.5 * ( x + 1.0 )
    ny                     = 255.0 * 0.5 * ( y + 1.0 )
    nz                     = 255.0 * 0.5 * ( z + 1.0 )

    R                      = int( round( nx ) )
    G                      = int( round( ny ) )
    B                      = int( round( nz ) )

    triple                 = [ R, G, B ]

    return triple   


# -----------------------------------------------------------------------------------
#    Compute magnitude of a float triple.                                 
# -----------------------------------------------------------------------------------
def mag( x, y, z ) :

    magn                   = x*x + y*y + z*z
    magn                   = magn**0.5

    return magn     


# -----------------------------------------------------------------------------------
#    Format triples of floats to have a consistant output string.
# -----------------------------------------------------------------------------------
def formatfloattriples( x, y, z ) :

    xstr                   = "%+7.5f" % x
    ystr                   = "%+7.5f" % y
    zstr                   = "%+7.5f" % z

    outstring              = ' ( ' + xstr + ', ' + ystr + ', ' + zstr + ' ) '
    return outstring


# -----------------------------------------------------------------------------------
#    Format vectors with magnitudes first.
# -----------------------------------------------------------------------------------
def formatfloatvector( x, y, z ) :

    magn                   = x*x + y*y + z*z
    magn                   = magn**0.5
    magnstr                = "%+7.5f" % magn
    xstr                   = "%+7.5f" % x
    ystr                   = "%+7.5f" % y
    zstr                   = "%+7.5f" % z

    outstring              = ' ' + magnstr + ' ( ' + xstr + ', ' + ystr + ', ' + zstr + ' ) '
    return outstring


# -----------------------------------------------------------------------------------
#    Format triples of bytes to have a consistant output string.
# -----------------------------------------------------------------------------------
def formatbytetriples( x, y, z ) :

    xstr                   = "%4d" % x
    ystr                   = "%4d" % y
    zstr                   = "%4d" % z

    outstring              = xstr + ystr + zstr + ' '
    return outstring

# ===================================================================================
#                            Reading procedures begin here.                         |
# ===================================================================================


# -----------------------------------------------------------------------------------
#    This def is used for grabbing header/footer lines of a given input size.
#    It aligns the bytes using the def alt.  It DOESN'T group them into blocks of 4.
#    It returns the formatted string for the caller to print out.
# -----------------------------------------------------------------------------------
def formatbytes( fidin, n ) :

    string                 = ''
    for ii in range(n) :
        thebyte            = getubyte( fidin )
        string             = string + alt( thebyte ) 

    return string


# -----------------------------------------------------------------------------------
#    This def is used for grabbing header/footer lines of a given input size.
#    It aligns the bytes using the def aby, and groups them into blocks of 4 and 
#    then writes to fidout as well as prints to the DOS box.
#    NOTE: Commented out the hex version; hex didn't help me.
# -----------------------------------------------------------------------------------
def gobblebytes( fidin, fidout, n ) :
    global bytecount
    global WRITEFOOTERS

    string                 = ''
    hexstr                 = ''
    for ii in range(n) :
        thebyte            = getubyte( fidin )
        string             = string + aby( thebyte ) 
        if ii % 4 == 3 :
            string         = string + '  '

#        # Now do it hex style.
#        lonib              = thebyte % 16 
#        hinib              = thebyte / 16         
#        hexstr             = hexstr + hexdigit( hinib ) + hexdigit( lonib ) + ' '
#        if ii % 4 == 3 :
#            hexstr         = hexstr + '  '

    if WRITEFOOTERS == True :
        fidout.write( string + '\n' )

    print string

#    fidout.write( hexstr + '\n' )
#    print hexstr

    return


# -----------------------------------------------------------------------------------
#    Writes the current byte count into the file both with .tell and our own 
#    counter.  Doesn't seem like .mesh files use a byte count anywhere so this
#    superfluous code.
# -----------------------------------------------------------------------------------
def writebytecount( fidin, fidout ) :
    global bytecount
    global WRITEBYTECOUNT

    if WRITEBYTECOUNT == True :
        nbytes                 = fidin.tell()
        print 'number of bytes from tell()    ' + str( nbytes )
        print 'number of bytes from bytecount ' + str( bytecount )
        fidout.write( 'number of bytes from tell()    ' + str( nbytes ) + '\n' )
        fidout.write( 'number of bytes from bytecount ' + str( bytecount ) + '\n' )
        fidout.flush()

    return


# -----------------------------------------------------------------------------------
#    Reads the main header and gets the num_groups variable at byte 69.  It leaves
#    the file indicator at byte 83 where the triangle or face data starts.
# -----------------------------------------------------------------------------------
def readmainheader( fidin, fidout ) :
    global bytecount
    global num_groups
    global WRITEDATA
    global WRITESTRINGS
    global VARIANTMESHFLAG

    # Print out first 87-26=61 bytes for reference then reset file indicator back to 0 to start fresh.
    fidin.seek( 26, 0 )                                     
    gobblebytes( fidin, fidout, 61 )
    fidin.seek( 0, 0 )                                     
    bytecount              = 0

    # Get a uint string count.
    nchars                 = getuint( fidin )
    # Read the string "serialization::archive".  This is always the same.
    string                 = fidin.read(nchars)
    bytecount              = bytecount + nchars

    # Output the strings to fidout and to the DOS box.
    if WRITESTRINGS == True :
        print str(nchars) + ' ' + string
        fidout.write( str( nchars ) + ' ' + string + '\n' )

    # Read and write 8 bytes.  These should be the same even for variant meshes.
    gobblebytes( fidin, fidout, 8 )

    # Read a uint. If 0 then it's a variant mesh, if it's 256 it's a regular mesh.
    meshtypeint            = getuint( fidin )
    print 'meshtypeint = ' + str( meshtypeint )
    fidout.write( 'meshtypeint = ' + str( meshtypeint ) + '\n' )

    if meshtypeint == 0 :
        VARIANTMESHFLAG    = True
        print 'Variant mesh, NOT reading 4 bytes'
        fidout.write( 'Variant mesh, NOT reading 4 bytes\n' )
    else :
        VARIANTMESHFLAG    = False
        print 'Regular mesh, reading 4 bytes'
        fidout.write( 'Regular mesh, reading 4 bytes\n' )
        gobblebytes( fidin, fidout, 4 )        # Gobble 4 bytes for regular meshes.

    # Read and write 27 bytes. So far we've read 26+8+4=38 for variants and
    # 26+8+4+4=42 for regular. After this call we should be at byte 38+27=65
    # for variants and byte 42+27=69 for regular.  This is where the number
    # of groups integer begins.
    gobblebytes( fidin, fidout, 27 )

    nbytes                 = fidin.tell()
    fidout.write( 'number of bytes from tell()    ' + str( nbytes ) + '\n' )
    fidout.write( 'number of bytes from bytecount ' + str( bytecount ) + '\n' )
    print 'At byte number ' + str( nbytes )

    # get the num_groups number by absolute seek to byte 69 (or 65 for variant meshes).
    # Yeah, sure we trust our code to this point, do absolute seek anyway.
    if VARIANTMESHFLAG == True :
        fidin.seek( 65, 0 )                                     
    else :
        fidin.seek( 69, 0 )                                     

    num_groups             = getuint( fidin )

    if WRITESTRINGS == True :
        print 'num_groups = ' + str(num_groups)
        fidout.write( 'num_groups = ' + str( num_groups ) + '\n' )

    # Optional byte count check. Should be at 73 for regular units and 69 for variants.
    nbytes                 = fidin.tell()
    fidout.write( 'number of bytes from tell()    ' + str( nbytes ) + '\n' )
    fidout.write( 'number of bytes from bytecount ' + str( bytecount ) + '\n' )
    fidout.flush()
    print 'At byte number ' + str( nbytes )

    return


# -----------------------------------------------------------------------------------
#    Reads the first triangle or face group.  Requires a custom def because the
#    first block has unusual skips and header/footer sizes.
#    mesh type name, mesh group name, and the array of triangle indicies are 
#    accumulated into trianglelist.
# -----------------------------------------------------------------------------------
def processfirstpart( fidin, fidout ) :
    import string
    global bytecount
    global WRITEDATA
    global WRITESTRINGS
    global VARIANTMESHFLAG
    global trianglelist

    # There are two 0 bytes after the num_groups integer.  Process them
    # to get to the start of the header for the first group of triangles.
    outstring              = formatbytes( fidin, 2 )
    fidout.write( outstring + '\n' )
    print outstring
    # Output a blank line to make it clear the two 0's have nothing to do
    # with the first header.
    fidout.write( '\n' )
    print ' '

    # This header has the 1 0 two byte variant code in it after the 11 0 
    # code.  Ouput these four bytes on a separate line for clarity.
    outstring              = formatbytes( fidin, 4 )
    fidout.write( outstring + '\n' )
    print outstring

    # Rest of header. Accumulate everything in outstring.
    outstring              = '      ' + formatbytes( fidin, 4 )
    outstring              = outstring.ljust(22)

    # Get count for the mesh type name string and then read name.
    nname                  = getuint( fidin )
    nname                  = int( nname )
    meshtypename           = fidin.read( nname )
    bytecount              = bytecount + nname
    if WRITESTRINGS == True :
        outstring          = outstring + meshtypename.ljust(18)

    # Save mesh type name.
    trianglelist.append( meshtypename )
    
    # Now get count for the mesh group name string and then read mesh group name.
    nname                  = getuint( fidin )
    meshgroupname          = fidin.read( nname )
    bytecount              = bytecount + nname
    if WRITESTRINGS == True :
        outstring          = outstring + meshgroupname.ljust(18)

    # Save mesh group name.
    trianglelist.append( meshgroupname )
    
    # Skip 2 bytes of dummy data between group name and integer with number of triangles.  (Only applies to the first group.) 
    fidin.seek( 2, 1 )
    bytecount              = bytecount + 2
    
    # Read integer with number of triangles.
    n                      = getuint( fidin )
    nstr                   = str( n )
    if WRITESTRINGS == True :
        outstring          = outstring + ' ' + nstr.rjust(4) + ' (tris:) '

    # Allocate an array of unsigned shorts and read in 3 * n of them from fidin.
    triarr                     = array.array( 'H' )
    triarr.fromfile( fidin, 3*n )
    bytecount              = bytecount + 3*2*n

    trianglelist.append( triarr )

    # This block is an extra four bytes 0 0 0 0 that only occurs with this first triangle entry.
    # We break it this way to still show how this footer aligns with all the others.
    outstring              = outstring + formatbytes( fidin, 4 )
    fidout.write( outstring + '\n' )
    print outstring

    # Write data if requested.
    if WRITEDATA == True :
        for ii in range( n ) :
            fidout.write( ash( triarr[3*ii] ) + '  ' + ash( triarr[3*ii+1] ) + ' ' + ash( triarr[3*ii+2] ) + '\n' )

        fidout.flush()

    # Since this is a 1 0 code variant we break the footer line to make this more apparent.
    # This long blank string aligns the rest of the footer with the other entry's footers. Messy but it works!
    outstring              = '                                                                        ' + formatbytes( fidin, 9 )
    fidout.write( outstring + '\n' )
    print outstring

    # Rest of footer bytes.
    outstring              = '                                                                                             ' + formatbytes( fidin, 10 )
    fidout.write( outstring + '\n' )
    print outstring

    # Set it off with a blank line.
    fidout.write( '\n' )
    print ' '

    # Optional byte count check.
    writebytecount( fidin, fidout )

    return


# -----------------------------------------------------------------------------------
#    Reads all the remaining triangle blocks which have uniform formatting.
#    mesh type name, mesh group name, and the array of triangle indicies are 
#    accumulated into trianglelist.
# -----------------------------------------------------------------------------------
def processregularpart( fidin, fidout ) :
    global bytecount
    global WRITEDATA
    global WRITESTRINGS
    global trianglelist

    # Get the header. All of these entries are regular and have 6 byte headers.
    outstring              = formatbytes( fidin, 6 )
    outstring              = outstring.ljust( 22 )

    # Get count for the mesh type name string and then read name.
    nname                  = getuint( fidin )
    meshtypename           = fidin.read( nname )
    bytecount              = bytecount + nname
    if WRITESTRINGS == True :
        outstring          = outstring + meshtypename.ljust(18)

    # Save mesh type name.
    trianglelist.append( meshtypename )
    
    # Now get count for the mesh group name string and then read mesh group name.
    nname                  = getuint( fidin )
    meshgroupname          = fidin.read( nname )
    bytecount              = bytecount + nname
    if WRITESTRINGS == True :
        outstring          = outstring + meshgroupname.ljust(18)

    # Save mesh group name.
    trianglelist.append( meshgroupname )
    
    # Read integer with number of triangles.
    n                      = getuint( fidin )
    nstr                   = str( n )
    if WRITESTRINGS == True :
        outstring          = outstring + ' ' + nstr.rjust(4) + ' (tris:) '
    
    # Allocate an array of unsigned shorts and read in 3 * n of them from fidin.
    triarr                     = array.array( 'H' )
    triarr.fromfile( fidin, 3*n )
    bytecount              = bytecount + 3*2*n

    # Save the data.
    trianglelist.append( triarr )

    # Footers are 17 bytes for regular triangle entries.
    outstring              = outstring + formatbytes( fidin, 17 )
    fidout.write( outstring + '\n' )
    print outstring

    # Write data if requested.
    if WRITEDATA == True :
        for ii in range( n ) :
            fidout.write( ash( triarr[3*ii] ) + '  ' + ash( triarr[3*ii+1] ) + ' ' + ash( triarr[3*ii+2] ) + '\n' )

        fidout.flush()

    # Optional byte count check.
    writebytecount( fidin, fidout )

    return


# -----------------------------------------------------------------------------------
#    Reads the uv coordinates block.  2 floats (x,y) for each vertex.           
#    The array is accumulated in the list vertexdatalist as the first entry.
# -----------------------------------------------------------------------------------
def processtexturecoordinates( fidin, fidout ) :
    global bytecount
    global WRITEDATA
    global WRITESTRINGS
    global vertexdatalist

    # uv coords usually have that 1 0 extra two bytes.
    outstring              = formatbytes( fidin, 4 )
    fidout.write( outstring + '\n' )
    print outstring

    # Read 10 more bytes for a total of 14 bytes in the header.
    outstring              = '      ' + formatbytes( fidin, 10 )
    outstring              = outstring.ljust(40)

    # Get count for the number of vertices.
    nvert                  = getuint( fidin )
    nvertstr               = str( nvert )
    if WRITESTRINGS == True :
        outstring          = outstring + nvertstr.ljust(5) + ' (uv coords: float pairs)  ' 
    
    # Allocate an array of floats and read in 2 * n of them from fidin.
    triarr                 = array.array( 'f' )
    triarr.fromfile( fidin, 2*nvert )
    bytecount              = bytecount + 2*4*nvert

    # Save data in list.
    vertexdatalist.append( triarr )

    # Display first 8 bytes of footer to show the 1 0 extra two bytes.
    outstring              = outstring + formatbytes( fidin, 8 )
    fidout.write( outstring + '\n' )
    print outstring

    # Write data if requested.
    if WRITEDATA == True :
        for ii in range( nvert ) :
            fidout.write( str( triarr[2*ii] ) + '  ' + str(triarr[2*ii+1]) + '\n' )

    # Output rest of footer for a total of 18 bytes.
    outstring              = '                                                                                          ' + formatbytes( fidin, 10 )
    fidout.write( outstring + '\n' )
    print outstring

    # Set off with a blank line.
    fidout.write( '\n' )
    print ' '

    # Optional byte count check.
    writebytecount( fidin, fidout )

    return


# -----------------------------------------------------------------------------------
#    Reads the vertex weights block.  2 floats (primary, secondary) for each vertex.           
#    The array is accumulated in the list vertexdatalist as the second entry.
# -----------------------------------------------------------------------------------
def processvertexweights( fidin, fidout ) :
    global bytecount
    global WRITEDATA
    global WRITESTRINGS

    # Vertex weights have a regular header of 10 bytes with no extra 1 0 codes.
    outstring              = formatbytes( fidin, 10 )
    outstring              = outstring.ljust(40)

    # Get count for the number of vertices.
    nvert                  = getuint( fidin )
    nvertstr               = str( nvert )
    if WRITESTRINGS == True :
        outstring          = outstring + nvertstr.ljust(5) + ' (vert. wts: float pairs)  ' 
    
    # Allocate an array of floats and read in 2 * n of them from fidin.
    triarr                 = array.array( 'f' )
    triarr.fromfile( fidin, 2*nvert )
    bytecount              = bytecount + 2*4*nvert

    # Save data in list.
    vertexdatalist.append( triarr )

    # Footer string of 24 bytes.
    outstring              = outstring + formatbytes( fidin, 24 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring

    # Write data if requested.
    if WRITEDATA == True :
        for ii in range( nvert ) :
            fidout.write( str( triarr[2*ii] ) + '  ' + str(triarr[2*ii+1]) + '\n' )

        fidout.flush()

    # Set off with blank line.
    fidout.write( '\n' )
    print ' '

    # Optional byte count check.
    writebytecount( fidin, fidout )

    return


# -----------------------------------------------------------------------------------
#    Reads the vertex vectors block.  3 floats (x,y,z) for each vertex.           
#    The array is accumulated in the list vertexdatalist as the third entry.
# -----------------------------------------------------------------------------------
def processvertexvectors( fidin, fidout ) :
    global bytecount
    global WRITEDATA
    global WRITESTRINGS
    global VARIANTMESHFLAG
    global TANGENTSPACEFLAG

    # Vectex vectors have the 1 0 code so we set it off by itself.
    outstring              = formatbytes( fidin, 4 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring

    # Rest of header for a total of 14 bytes.
    outstring              = '      ' + formatbytes( fidin, 10 )
    outstring              = outstring.ljust(40)

    # Get count for the number of vertices.
    nvert                  = getuint( fidin )
    nvertstr               = str( nvert )
    if WRITESTRINGS == True :
        outstring          = outstring + nvertstr.ljust(5) + ' (vert. vecs: float triple)' 
    
    # Allocate an array of floats and read in 3 * n of them from fidin.
    triarr                 = array.array( 'f' )
    triarr.fromfile( fidin, 3*nvert )
    bytecount              = bytecount + 3*4*nvert

    # Save data in list.
    vertexdatalist.append( triarr )

    # Set of the 1 0 code by writing first 8 bytes of footer.
    outstring              = outstring + formatbytes( fidin, 8 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring

    # Write data if requested.
    if WRITEDATA == True :
        for ii in range( nvert ) :
            fidout.write( str( triarr[3*ii] ) + '  ' + str(triarr[3*ii+1]) + '  ' + str(triarr[3*ii+2]) + '\n' )

        fidout.flush()

    # Rest of footer is 18 bytes for a total of 26 (or 10 more for total of 18 bytes for variant meshes).
    if VARIANTMESHFLAG == False :
        outstring          = '                                                                                          ' + formatbytes( fidin, 18 )
        # Need to grab the 4 0 0 0 or 2 0 0 0 int and put it in a global to do mounts.
        fidin.seek( -6, 1 )
        TANGENTSPACEFLAG   = getuint( fidin )
        fidin.seek(  2, 1 )
        bytecount          = bytecount - 4
    else :
        outstring          = '                                                                                          ' + formatbytes( fidin, 10 )

    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring

    # Set off with a blank line.
    fidout.write( '\n' )
    print ' '

    # Optional byte count check.
    writebytecount( fidin, fidout )

    return


# -----------------------------------------------------------------------------------
#    Reads the vertex bone assignment block.  4 bytes (0,n,m,0) for each vertex.           
#    The array is accumulated in the list vertexdatalist as the fourth entry.
# -----------------------------------------------------------------------------------
def processvertexbones( fidin, fidout ) :
    global bytecount
    global WRITEDATA
    global WRITESTRINGS
    global VARIANTMESHFLAG
    global variantboundingsphere

    # Header for vertex bones has a 1 0 code so set it off.
    outstring              = formatbytes( fidin, 4 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring

    # Rest of header for a total of 14 bytes.
    outstring              = '      ' + formatbytes( fidin, 10 )
    outstring              = outstring.ljust(40)

    # Get count for the number of vertices.
    nvert                      = getuint( fidin )
    nvertstr               = str( nvert )
    if WRITESTRINGS == True :
        outstring          = outstring + nvertstr.ljust(5) + ' (vert. bones: byte quads) ' 
    
    # Allocate an array of unsigned bytes and read in 4 * n of them from fidin.
    triarr                     = array.array( 'B' )
    triarr.fromfile( fidin, 4*nvert )
    bytecount                  = bytecount + 4*nvert

    # Save data in list.
    vertexdatalist.append( triarr )

    # The 1 0 code showing up again, set it off by writing 8 bytes.
    outstring              = outstring + formatbytes( fidin, 8 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring

    # Rest of footer for a total of 8+10=18 bytes.
    outstring              = '                                                                                          ' + formatbytes( fidin, 10 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring

    # Set it off with a blank line.
    fidout.write( '\n' )
    print ' '

    # Write data if requested.
    if WRITEDATA == True :
        for ii in range( nvert ) :
            fidout.write( str( triarr[4*ii] ) + '  ' + str(triarr[4*ii+1]) + '  ' + str(triarr[4*ii+2]) + '  ' + str(triarr[4*ii+3]) + '\n' )

        fidout.flush()

    # Variant meshes have the bone assignments AFTER the mystery block so we have
    # to grab the bounding sphere data here.
    if VARIANTMESHFLAG == True :
        # These 10 bytes are always zeros.
        gobblebytes( fidin, fidout, 10 )
        # Have to grab the four floats of the bounding sphere here if its a variant mesh.
        f1                 = getfloat( fidin )
        f2                 = getfloat( fidin )
        f3                 = getfloat( fidin )
        f4                 = getfloat( fidin )
        variantboundingsphere = [ f1, f2, f3, f4 ]
        fidout.write( str( f1 ) + ' ' + str( f2 ) + ' ' + str( f3 ) + ' ' + str( f4 ) + '\n' )
        print str(f1) + ' ' + str(f2) + ' ' + str(f3) + ' ' + str(f4)

        # Skip 4 more bytes, these are also always zero.
        gobblebytes( fidin, fidout, 4 )

    # Optional byte count check.
    writebytecount( fidin, fidout )

    return


# -----------------------------------------------------------------------------------
#    Reads the mystery block.  4 bytes (j,k,n,m) for each vertex.           
#    Usually the fourth entry is 0.  Latest speculation from Caliban is that this
#    is a leftover boolean stream that isn't needed.  For now just read it and
#    save it.  Note that the hi-res lod0 mesh files have three blocks and the 
#    other lod1, lod2, lod3 SEEM to only have 1 block.  (Last statement based
#    on a very limited survey.)
#    
#    Multiple block processing is signaled through the global flag LOD0FLAG.
#    The main routine checks the input filename for the presence of the string
#    'lod0' and sets the flag True if found.
#
#    The array is NOT saved into vertexdatalist at this point.  If this data
#    turns out to be needed, then this must be programmed in.
# -----------------------------------------------------------------------------------
def processmysteryblock( fidin, fidout ) :
    import math

    global tangentspacedatalist
    global bytecount
    global WRITEDATA
    global WRITESTRINGS
    global LOD0FLAG
    global TANGENTSPACEFLAG
    if TANGENTSPACEFLAG == 4 :
        LOD0FLAG           = True  
    else :
        LOD0FLAG           = False  

    vnorms                 = tangentspacedatalist[0]
    vtans                  = tangentspacedatalist[1]
    vbinorms               = tangentspacedatalist[2]
    vrefs                  = tangentspacedatalist[3]

    # Mystery blocks are regular, no 1 0 codes.  10 bytes headers.
    outstring              = formatbytes( fidin, 10 )
    outstring              = outstring.ljust(40)

    # Get count for the number of vertices.
    nvert                      = getuint( fidin )
    nvertstr               = str( nvert )
    if WRITESTRINGS == True :
        outstring          = outstring + nvertstr.ljust(5) + ' (mystery strm: byte quads) =============='

    # Allocate an array of unsigned bytes and read in 4*nvert of them from fidin.
    quadarr                = array.array( 'B' )
    quadarr.fromfile( fidin, 4*nvert )
    bytecount              = bytecount + 4*1*nvert

    # 16 byte footer. 
    outstring              = outstring + formatbytes( fidin, 16 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring

    # Write data if requested.
    if WRITEDATA == True :
        fidout.write( 'Original RGB  Computed RGB  Ref.            Original tangent space vector               Computed tangent space vector         \n' )
        fidout.write( '  z   y   x     z   y   x   count    mag          z         y        x           mag       z          y           x    Angle  \n' )
        for ii in range( nvert ) :
            btriple        = f2rgb( vnorms[3*ii], vnorms[3*ii+1], vnorms[3*ii+2] )
            ftriple        = rgb2f( quadarr[4*ii], quadarr[4*ii+1], quadarr[4*ii+2] )
            dotprod        = ftriple[0] * vnorms[3*ii+2] + ftriple[1] * vnorms[3*ii+1] + ftriple[2] * vnorms[3*ii]            
            if ( dotprod >= -1.0 ) & ( dotprod <= 1.0 ) :
                ang        = 180.0 * math.acos( dotprod ) / math.acos(-1.0)
            else :
                ang        = 0.0
            fidout.write( formatbytetriples( quadarr[4*ii], quadarr[4*ii+1], quadarr[4*ii+2] ) + formatbytetriples( btriple[2], btriple[1], btriple[0] ) + '  [' + "%2s" % vrefs[ii] + '] '  + formatfloatvector( ftriple[0], ftriple[1], ftriple[2] ) + '  ' + formatfloatvector( vnorms[3*ii+2], vnorms[3*ii+1], vnorms[3*ii+0] ) + "%7.5f" % ang + '\n' )

        fidout.flush()

    if LOD0FLAG == True :
        # 10 byte header.
        outstring          = formatbytes( fidin, 10 )
        outstring          = outstring.ljust(40)

        # Get count for the number of vertices.
        nvert              = getuint( fidin )
        nvertstr           = str( nvert )
        if WRITESTRINGS == True :
            outstring      = outstring + nvertstr.ljust(5) + ' (mystery strm: byte quads) ================' 

        # Allocate an array of unsigned bytes and read in 4*nvert of them from fidin.
        quadarr            = array.array( 'B' )
        quadarr.fromfile( fidin, 4*nvert )
        bytecount          = bytecount + 4*1*nvert

        # 16 byte footer.
        outstring          = outstring + formatbytes( fidin, 16 )
        fidout.write( outstring + '\n' )
        fidout.flush()
        print outstring

        # Write data if requested.
        if WRITEDATA == True :
            for ii in range( nvert ) :
                btriple    = f2rgb( vtans[3*ii], vtans[3*ii+1], vtans[3*ii+2] )
                ftriple    = rgb2f( quadarr[4*ii], quadarr[4*ii+1], quadarr[4*ii+2] )
                dotprod    = ftriple[0] * vtans[3*ii+2] + ftriple[1] * vtans[3*ii+1] + ftriple[2] * vtans[3*ii]            
                if ( dotprod >= -1.0 ) & ( dotprod <= 1.0 ) :
                    ang    = 180.0 * math.acos( dotprod ) / math.acos(-1.0)
                else :
                    ang    = 0.0
                fidout.write( formatbytetriples( quadarr[4*ii], quadarr[4*ii+1], quadarr[4*ii+2] ) + formatbytetriples( btriple[2], btriple[1], btriple[0] ) + '  [' + "%2s" % vrefs[ii] + '] '  + formatfloatvector( ftriple[0], ftriple[1], ftriple[2] ) + '  ' + formatfloatvector( vtans[3*ii+2], vtans[3*ii+1], vtans[3*ii+0] ) + "%7.5f" % ang + '\n' )

            fidout.flush()

        # 10 byte header for third block.
        outstring          = formatbytes( fidin, 10 )
        outstring          = outstring.ljust(40)

        # Get count for the number of vertices.
        nvert              = getuint( fidin )
        nvertstr           = str( nvert )
        if WRITESTRINGS == True :
            outstring      = outstring + nvertstr.ljust(5) + ' (mystery strm: byte quads) ================' 

        # Allocate an array of unsigned bytes and read in 4*nvert of them from fidin.
        quadarr            = array.array( 'B' )
        quadarr.fromfile( fidin, 4*nvert )
        bytecount          = bytecount + 4*1*nvert

        # 16 byte footer.
        outstring          = outstring + formatbytes( fidin, 16 )
        fidout.write( outstring + '\n' )
        fidout.flush()
        print outstring

        # Write data if requested.
        if WRITEDATA == True :
            for ii in range( nvert ) :
                btriple    = f2rgb( vbinorms[3*ii], vbinorms[3*ii+1], vbinorms[3*ii+2] )
                ftriple    = rgb2f( quadarr[4*ii], quadarr[4*ii+1], quadarr[4*ii+2] )
                dotprod    = ftriple[0] * vbinorms[3*ii+2] + ftriple[1] * vbinorms[3*ii+1] + ftriple[2] * vbinorms[3*ii]            
                if ( dotprod >= -1.0 ) & ( dotprod <= 1.0 ) :
                    ang    = 180.0 * math.acos( dotprod ) / math.acos(-1.0)
                else :
                    ang    = 0.0
                fidout.write( formatbytetriples( quadarr[4*ii], quadarr[4*ii+1], quadarr[4*ii+2] ) + formatbytetriples( btriple[2], btriple[1], btriple[0] ) + '  [' + "%2s" % vrefs[ii] + '] '  + formatfloatvector( ftriple[0], ftriple[1], ftriple[2] ) + '  ' + formatfloatvector( vbinorms[3*ii+2], vbinorms[3*ii+1], vbinorms[3*ii+0] ) + "%7.5f" % ang + '\n' )

            fidout.flush()

    # Done with processing blocks 2 and 3.

    # 10 bytes of zeros.
    outstring          = formatbytes( fidin, 10 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring

    # Try to yank four floats (bounding sphere) at the end of the mystery block.
    f1                     = getfloat( fidin )
    f2                     = getfloat( fidin )
    f3                     = getfloat( fidin )
    f4                     = getfloat( fidin )
    fidout.write( 'Bounding sphere data: 4 floats\n' )
    fidout.write( str( f1 ) + ' ' + str( f2 ) + ' ' + str( f3 ) + ' ' + str( f4 ) + '\n' )
    print 'Bounding sphere data: 4 floats'
    print str(f1) + ' ' + str(f2) + ' ' + str(f3) + ' ' + str(f4)

    # A final 4 byte footer.
    gobblebytes( fidin, fidout, 4 )

    # Optional byte count check.
    writebytecount( fidin, fidout )

    return


# -----------------------------------------------------------------------------------
#    Reads the variant mystery block.   DEAD CODE: WE DON'T DO THE AZTEC VARIANTS.
#    
# -----------------------------------------------------------------------------------
def processvariantmysteryblock( fidin, fidout ) :
    global bytecount
    global WRITEDATA
    global WRITESTRINGS
    global LOD0FLAG

    LOD0FLAG               = True

    # 10 byte header.
    outstring              = formatbytes( fidin, 10 )
    outstring              = outstring.ljust(40)

    # Get count for the number of vertices.
    nvert                  = getuint( fidin )
    nvertstr               = str( nvert )
    if WRITESTRINGS == True :
        outstring          = outstring + nvertstr.ljust(5) + ' (var myst: float triples) ' 

    # Allocate an array of floats and read in 3*nvert of them from fidin.
    triarr                 = array.array( 'f' )
    triarr.fromfile( fidin, 3*nvert )
    bytecount              = bytecount + 3*4*nvert
    if WRITEDATA == True :
        for ii in range( nvert ) :
            fidout.write( str( triarr[3*ii] ) + ' ' + str(triarr[3*ii+1]) + ' ' + str(triarr[3*ii+2]) + '\n' )

        fidout.flush()

    # 16 byte footer. 
    outstring              = outstring + formatbytes( fidin, 16 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring


    if LOD0FLAG == True :
        # Get 10 bytes of header.
        outstring          = formatbytes( fidin, 10 )
        outstring          = outstring.ljust(40)
        # Get count for the number of vertices.
        nvert              = getuint( fidin )
        nvertstr           = str( nvert )
        if WRITESTRINGS == True :
            outstring      = outstring + nvertstr.ljust(5) + ' (var myst: float triples) ' 

        # Allocate an array of floats and read in 3*nvert of them from fidin.
        triarr             = array.array( 'f' )
        triarr.fromfile( fidin, 3*nvert )
        bytecount          = bytecount + 3*4*nvert
        if WRITEDATA == True :
            for ii in range( nvert ) :
                fidout.write( str( triarr[3*ii] ) + ' ' + str(triarr[3*ii+1]) + ' ' + str(triarr[3*ii+2]) + '\n' )

            fidout.flush()

        # 16 byte footer.
        outstring          = outstring + formatbytes( fidin, 16 )
        fidout.write( outstring + '\n' )
        fidout.flush()
        print outstring

        # Get 10 bytes of header.
        outstring          = formatbytes( fidin, 10 )
        outstring          = outstring.ljust(40)

        # Get count for the number of vertices.
        nvert              = getuint( fidin )
        nvertstr           = str( nvert )
        if WRITESTRINGS == True :
            outstring      = outstring + nvertstr.ljust(5) + ' (var myst: float triples) ' 

        # Allocate an array of unsigned bytes and read in 4*nvert of them from fidin.
        triarr                 = array.array( 'f' )
        triarr.fromfile( fidin, 3*nvert )
        bytecount              = bytecount + 3*4*nvert
        if WRITEDATA == True :
            for ii in range( nvert ) :
                fidout.write( str( triarr[3*ii] ) + ' ' + str(triarr[3*ii+1]) + ' ' + str(triarr[3*ii+2]) + '\n' )

            fidout.flush()

        # 16 byte footer.
        outstring          = outstring + formatbytes( fidin, 16 )
        fidout.write( outstring + '\n' )
        fidout.flush()
        print outstring

        # 8 byte footer.
        fidout.write( '\n' )
        print ' '
        gobblebytes( fidin, fidout, 8 )
        fidout.write( '\n' )
        print ' '

#        # And another 12 for lod0.
#        gobblebytes( fidin, fidout, 10 )

    # Optional byte count check.
    writebytecount( fidin, fidout )

    return


# -----------------------------------------------------------------------------------
#    Reads the bone strings and the rest of the file to the end.                           
#    Bone strings have a character count as an int, an ASCII string of that length
#    that is the bone string, and a hierarchy assignment number as an int.
#    Then a 30 byte header/footer followed by a unit name?  Then a bunch more
#    bytes, and finally the same bounding sphere four floats as read in the 
#    mystery block.
#
#    The data is accumulated in the list bonestringslist.
# -----------------------------------------------------------------------------------
def processbonestrings( fidin, fidout ) :
    global bytecount
    global bonestringslist
    global VARIANTMESHFLAG
    global texturefilelist

    # Get number of bone strings.
    nstrings               = getuint( fidin )
    fidout.write( 'Processing bone strings, number = ' + str(nstrings) + '\n' )
    print 'Processing bone strings, number = ' + str(nstrings)

    # Skip 2 bytes.
    fidin.seek( 2, 1 )
    bytecount          = bytecount + 2

    for ii in range( nstrings ) :
        nchars             = getuint( fidin )
        bonestring         = fidin.read( nchars )
        bonestringslist.append( bonestring )
        bytecount          = bytecount + nchars
        cnt                = getuint( fidin )
        bonestringslist.append( cnt )
        outstring          = str(nchars) + ' ' + bonestring + ' ' + str( cnt )
        fidout.write( outstring + '\n' )
        print outstring
        fidout.flush()

    # Is this a termination string?
    gobblebytes( fidin, fidout, 30 )

    # Another char string?
    nchars                 = getuint( fidin )
    fidout.write( 'number of chars ' + str( nchars ) + '\n' )
    fidout.flush()
    print 'number of chars ' + str( nchars )
    bonestring         = fidin.read( nchars )
    bonestringslist.append( bonestring )
    bytecount          = bytecount + nchars
    fidout.write( bonestring + '\n' )
    print bonestring

    if VARIANTMESHFLAG == False :
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout,  8 )
    else :
        gobblebytes( fidin, fidout, 4 )                    # This is the 17 0 0 0 int that ends the bonestrings
        nchar              = getuint( fidin )              # String length for first texture file.
        texfile1           = fidin.read( nchar )           # Reads the file name.
        bytecount          = bytecount + nchar               
        fidout.write( str( nchar ) + ' ' + texfile1 + '\n' )
        nchar              = getuint( fidin )              # String length for second texture file.
        texfile2           = fidin.read( nchar )           # Reads the file name.
        gobblebytes( fidin, fidout, 4 )                    # Zero bytes have to be processes.
        bytecount          = bytecount + nchar               
        fidout.write( str( nchar ) + ' ' + texfile2 + '\n' )
        nchar              = getuint( fidin )              # String length for third texture file.
        texfile3           = fidin.read( nchar )           # Reads the file name.
        bytecount          = bytecount + nchar               
        fidout.write( str( nchar ) + ' ' + texfile3 + '\n' )
        gobblebytes( fidin, fidout, 24 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 28 )
        gobblebytes( fidin, fidout, 20 )
        texturefilelist    = []
        texturefilelist.append( texfile1 )
        texturefilelist.append( texfile2 )
        texturefilelist.append( texfile3 )
        print 'texfile1 = ' + texfile1
        print 'texfile2 = ' + texfile2
        print 'texfile3 = ' + texfile3
        
    
    # Allocate an array of floats and read in 4 element bounding sphere vector.
    boundingsphere         = array.array( 'f' )
    boundingsphere.fromfile( fidin, 4 )        
    bytecount              = bytecount + 16
    bonestringslist.append( boundingsphere )

    fidout.write( str( boundingsphere[0] ) + ' ' + str( boundingsphere[1] ) + ' ' + str( boundingsphere[2] ) + ' ' + str( boundingsphere[3] ) + '\n' )
    print str( boundingsphere[0] ) + ' ' + str( boundingsphere[1] ) + ' ' + str( boundingsphere[2] ) + ' ' + str( boundingsphere[3] )  

    # Optional byte count check.
    writebytecount( fidin, fidout )

    return


# -----------------------------------------------------------------------------------
#    This def computes vertex normals.  It goes through every triangle and computes |
#    a face normal and then it uses that normal to add to each of its vertices.     |
#    After every face is processed you have the average normal of each triangle     |
#    sharing a given vertex.  Normalize this vector and, hey presto!, an ersatz     |
#    vertex normal, a more unmathematical concept I could not imagine.  It then     |
#    computes the tangents and binormals to have a complete tangent space basis.    |
# -----------------------------------------------------------------------------------
def computetangentspacebasis( fidout ) :
    import math

    # globals.
    global trianglelist
    global vertexdatalist
    global tangentspacedatalist

    # This holds the tangent space basis vectors.
    tangentspacedatalist   = []

    # Unpack out the uvcoords 2-vectors.
    uvcoords               = vertexdatalist[0]

    # Unpack out the vertex vectors.
    vvecs                  = vertexdatalist[2]
    nitems                 = len( vvecs )
    nvert                  = nitems / 3
    print 'Computing tangent space basis'
    print 'nvert    = ' + str( nvert )

    # Need to make 3 float arrays of size nvert x 3 to hold the normals, tangents, and binormals. 
    vnorms                 = array.array( 'f' )
    vtans                  = array.array( 'f' )
    vbinorms               = array.array( 'f' )
    for ii in range(nitems) :
        vnorms.append( 0.0 )
        vtans.append( 0.0 )
        vbinorms.append( 0.0 )

    # Need a byte array of size Nvert to hold the reference counts.
    vrefs                  = array.array( 'B' )
    for ii in range( nvert ) :
        vrefs.append( 0 )

    # Start processing the triangles
    nitems                 = len( trianglelist )
    ntrilists              = nitems / 3
    for ii in range( ntrilists ) :
        trilist            = trianglelist[3*ii+2]
        nitems             = len( trilist )
        ntris              = nitems / 3
        for jj in range( ntris ) :
            idx1           = trilist[3*jj]        # Curse Python for not having a built-in 2-D array!
            idx2           = trilist[3*jj+1]
            idx3           = trilist[3*jj+2]
            vrefs[idx1]    = vrefs[idx1] + 1
            vrefs[idx2]    = vrefs[idx2] + 1
            vrefs[idx3]    = vrefs[idx3] + 1
            v1x            = vvecs[3*idx1]
            v1y            = vvecs[3*idx1+1]
            v1z            = vvecs[3*idx1+2]
            v2x            = vvecs[3*idx2]
            v2y            = vvecs[3*idx2+1]
            v2z            = vvecs[3*idx2+2]
            v3x            = vvecs[3*idx3]
            v3y            = vvecs[3*idx3+1]
            v3z            = vvecs[3*idx3+2]
            edge12x        = v2x - v1x            #
            edge12y        = v2y - v1y            #   | i   j   k  |
            edge12z        = v2z - v1z            #   | ux  uy  uz | = i(uy*vz-uz*vy) + j(uz*vx-ux*vz) + k(ux*vy-uy*vx)
            edge23x        = v3x - v2x            #   | vx  vy  vz |
            edge23y        = v3y - v2y            #
            edge23z        = v3z - v2z            #
            # Take cross product to get a normal vector (normal to face, but it won't be normalized).
            unx            = edge12y * edge23z - edge12z * edge23y
            uny            = edge12z * edge23x - edge12x * edge23z
            unz            = edge12x * edge23y - edge12y * edge23x
            # Normalize it.
            norm           = unx*unx + uny*uny + unz*unz 
            norm           = norm**0.5 
            if norm != 0.0 :
                nx         = unx / norm     
                ny         = uny / norm
                nz         = unz / norm
            else :
                nx         = unx
                ny         = uny
                nz         = unz

            # Add it to vnorms for each index.
            vnorms[3*idx1]   = vnorms[3*idx1]   + nx 
            vnorms[3*idx1+1] = vnorms[3*idx1+1] + ny 
            vnorms[3*idx1+2] = vnorms[3*idx1+2] + nz 
            vnorms[3*idx2]   = vnorms[3*idx2]   + nx 
            vnorms[3*idx2+1] = vnorms[3*idx2+1] + ny 
            vnorms[3*idx2+2] = vnorms[3*idx2+2] + nz 
            vnorms[3*idx3]   = vnorms[3*idx3]   + nx 
            vnorms[3*idx3+1] = vnorms[3*idx3+1] + ny 
            vnorms[3*idx3+2] = vnorms[3*idx3+2] + nz 

            # Another algorithm.
            E1             = [ v2x-v1x, v2y-v1y, v2z-v1z ]
            E2             = [ v3x-v2x, v3y-v2y, v3z-v2z ]
            E3             = [ v1x-v3x, v1y-v3y, v1z-v3z ]
            E1mag          = E1[0]*E1[0] + E1[1]*E1[1] + E1[2]*E1[2] 
            E2mag          = E2[0]*E2[0] + E2[1]*E2[1] + E2[2]*E2[2] 
            E3mag          = E3[0]*E3[0] + E3[1]*E3[1] + E3[2]*E3[2] 
            E1mag          = E1mag**0.5
            E2mag          = E2mag**0.5
            E3mag          = E3mag**0.5
            # Do vertex 1.
            E3XE1          = [ E3[1]*E1[2] - E3[2]*E1[1], E3[2]*E1[0] - E3[0]*E1[2], E3[0]*E1[1] - E3[1]*E1[0] ]
            E3XE1mag       = E3XE1[0]*E3XE1[0] + E3XE1[1]*E3XE1[1] + E3XE1[2]*E3XE1[2]  
            E3XE1mag       = E3XE1mag**0.5
            #wt1            = E3XE1mag / ( E3mag * E3mag * E1mag * E1mag )
            #wt1            = E3XE1mag 
            #wt1            = 1.0 / ( E3mag * E1mag )
            #wt1            = ( E3[0]*E1[0] + E3[1]*E1[1] + E3[2]*E1[2] ) / ( E3mag * E1mag )
            #wt1            = math.acos( wt1 ) 

            # Do vertex 2.
            E1XE2          = [ E1[1]*E2[2] - E1[2]*E2[1], E1[2]*E2[0] - E1[0]*E2[2], E1[0]*E2[1] - E1[1]*E2[0] ]
            E1XE2mag       = E1XE2[0]*E1XE2[0] + E1XE2[1]*E1XE2[1] + E1XE2[2]*E1XE2[2]  
            E1XE2mag       = E1XE2mag**0.5
            #wt2            = E1XE2mag / ( E1mag * E1mag * E2mag * E2mag )
            #wt2            = E1XE2mag 
            #wt2            = 1.0 / ( E1mag * E2mag )
            #wt2            = ( E1[0]*E2[0] + E1[1]*E2[1] + E1[2]*E2[2] ) / ( E1mag * E2mag )
            #wt2            = math.acos( wt2 ) 

            # Do vertex 3.
            E2XE3          = [ E2[1]*E3[2] - E2[2]*E3[1], E2[2]*E3[0] - E2[0]*E3[2], E2[0]*E3[1] - E2[1]*E3[0] ]
            E2XE3mag       = E2XE3[0]*E2XE3[0] + E2XE3[1]*E2XE3[1] + E2XE3[2]*E2XE3[2]  
            E2XE3mag       = E2XE3mag**0.5
            #wt3            = E2XE3mag / ( E2mag * E2mag * E3mag * E3mag )
            #wt3            = E2XE3mag 
            #wt3            = 1.0 / ( E2mag * E3mag )
            #wt3            = ( E2[0]*E3[0] + E2[1]*E3[1] + E2[2]*E3[2] ) / ( E2mag * E3mag )
            #wt3            = math.acos( wt3 ) 


            # Add it to vnorms for each index.
            #vnorms[3*idx1]   = vnorms[3*idx1]   + nx * wt1
            #vnorms[3*idx1+1] = vnorms[3*idx1+1] + ny * wt1
            #vnorms[3*idx1+2] = vnorms[3*idx1+2] + nz * wt1
            #vnorms[3*idx2]   = vnorms[3*idx2]   + nx * wt2
            #vnorms[3*idx2+1] = vnorms[3*idx2+1] + ny * wt2
            #vnorms[3*idx2+2] = vnorms[3*idx2+2] + nz * wt2
            #vnorms[3*idx3]   = vnorms[3*idx3]   + nx * wt3
            #vnorms[3*idx3+1] = vnorms[3*idx3+1] + ny * wt3
            #vnorms[3*idx3+2] = vnorms[3*idx3+2] + nz * wt3

            # Now process for the tangent vector.
            edge13x        = v3x - v1x           
            edge13y        = v3y - v1y            
            edge13z        = v3z - v1z            
            uv1x           = uvcoords[2*idx1]
            uv1y           = uvcoords[2*idx1+1]
            uv2x           = uvcoords[2*idx2]
            uv2y           = uvcoords[2*idx2+1]
            uv3x           = uvcoords[2*idx3]
            uv3y           = uvcoords[2*idx3+1]
            edge12uvx      = uv2x - uv1x
            edge12uvy      = uv2y - uv1y
            edge13uvx      = uv3x - uv1x
            edge13uvy      = uv3y - uv1y
            tx             = ( -edge12x * edge13uvy + edge13x * edge12uvy )
            ty             = ( -edge12y * edge13uvy + edge13y * edge12uvy )
            tz             = ( -edge12z * edge13uvy + edge13z * edge12uvy )
            bx             = ( -edge12x * edge13uvx + edge13x * edge12uvx )
            by             = ( -edge12y * edge13uvx + edge13y * edge12uvx )
            bz             = ( -edge12z * edge13uvx + edge13z * edge12uvx )
            cp             = edge12uvy * edge13uvx - edge12uvx * edge13uvy
            if cp != 0.0 :
                mul        = 1.0 / cp
                tx         = -mul * tx
                ty         = -mul * ty
                tz         = -mul * tz
                bx         = mul * bx
                by         = mul * by
                bz         = mul * bz

            # Normalize the tangent vector.
            norm           = tx*tx + ty*ty + tz*tz
            norm           = norm**0.5
            if norm != 0.0 :
                tx         = tx / norm
                ty         = ty / norm
                tz         = tz / norm
            vtans[3*idx1]  = vtans[3*idx1]   + tx
            vtans[3*idx2]  = vtans[3*idx2]   + tx
            vtans[3*idx3]  = vtans[3*idx3]   + tx
            vtans[3*idx1+1]= vtans[3*idx1+1] + ty
            vtans[3*idx2+1]= vtans[3*idx2+1] + ty
            vtans[3*idx3+1]= vtans[3*idx3+1] + ty
            vtans[3*idx1+2]= vtans[3*idx1+2] + tz
            vtans[3*idx2+2]= vtans[3*idx2+2] + tz
            vtans[3*idx3+2]= vtans[3*idx3+2] + tz

            # Normalize the binormal vector.
            norm           = bx*bx + by*by + bz*bz
            norm           = norm**0.5
            if norm != 0.0 :
                bx         = bx / norm
                by         = by / norm
                bz         = bz / norm
            vbinorms[3*idx1]  = vbinorms[3*idx1]   + bx
            vbinorms[3*idx2]  = vbinorms[3*idx2]   + bx
            vbinorms[3*idx3]  = vbinorms[3*idx3]   + bx
            vbinorms[3*idx1+1]= vbinorms[3*idx1+1] + by
            vbinorms[3*idx2+1]= vbinorms[3*idx2+1] + by
            vbinorms[3*idx3+1]= vbinorms[3*idx3+1] + by
            vbinorms[3*idx1+2]= vbinorms[3*idx1+2] + bz
            vbinorms[3*idx2+2]= vbinorms[3*idx2+2] + bz
            vbinorms[3*idx3+2]= vbinorms[3*idx3+2] + bz


    # Now we have to renormalize our vnorms.
    for ii in range( nvert ) :
        norm               = vnorms[3*ii]*vnorms[3*ii] + vnorms[3*ii+1]*vnorms[3*ii+1] + vnorms[3*ii+2]*vnorms[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0 :
            vnorms[3*ii]   = vnorms[3*ii]   / norm
            vnorms[3*ii+1] = vnorms[3*ii+1] / norm
            vnorms[3*ii+2] = vnorms[3*ii+2] / norm

    # Now we have to renormalize our vtans.
    for ii in range( nvert ) :
        norm               = vtans[3*ii]*vtans[3*ii] + vtans[3*ii+1]*vtans[3*ii+1] + vtans[3*ii+2]*vtans[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0 :
            vtans[3*ii]    = vtans[3*ii]   / norm
            vtans[3*ii+1]  = vtans[3*ii+1] / norm
            vtans[3*ii+2]  = vtans[3*ii+2] / norm

    # Now we have to renormalize our vbinorms.
    for ii in range( nvert ) :
        norm               = vbinorms[3*ii]*vbinorms[3*ii] + vbinorms[3*ii+1]*vbinorms[3*ii+1] + vbinorms[3*ii+2]*vbinorms[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0 :
            vbinorms[3*ii]   = vbinorms[3*ii]   / norm
            vbinorms[3*ii+1] = vbinorms[3*ii+1] / norm
            vbinorms[3*ii+2] = vbinorms[3*ii+2] / norm

    # First order Gram-Schmidt to orthogonalize vtans against vnorms.
    for ii in range( nvert ):
        dotprod            = vnorms[3*ii] * vtans[3*ii] + vnorms[3*ii+1] * vtans[3*ii+1] + vnorms[3*ii+2] * vtans[3*ii+2]
        vtans[3*ii]        = vtans[3*ii]   - vnorms[3*ii]   * dotprod
        vtans[3*ii+1]      = vtans[3*ii+1] - vnorms[3*ii+1] * dotprod
        vtans[3*ii+2]      = vtans[3*ii+2] - vnorms[3*ii+2] * dotprod

    # Now we have to renormalize AGAIN our vtans.
    for ii in range( nvert ) :
        norm               = vtans[3*ii]*vtans[3*ii] + vtans[3*ii+1]*vtans[3*ii+1] + vtans[3*ii+2]*vtans[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0 :
            vtans[3*ii]    = vtans[3*ii]   / norm
            vtans[3*ii+1]  = vtans[3*ii+1] / norm
            vtans[3*ii+2]  = vtans[3*ii+2] / norm

    # Second order Gram-Schmidt to orthogonalize vbinorms against vnorms AND vtans.
    for ii in range( nvert ):
        dotprod            = vbinorms[3*ii] * vnorms[3*ii] + vbinorms[3*ii+1] * vnorms[3*ii+1] + vbinorms[3*ii+2] * vnorms[3*ii+2]
        dotprod2           = vbinorms[3*ii] * vtans[3*ii]  + vbinorms[3*ii+1] * vtans[3*ii+1]  + vbinorms[3*ii+2] * vtans[3*ii+2]
        vbinorms[3*ii]     = vbinorms[3*ii]   - vtans[3*ii]   * dotprod2 - vnorms[3*ii]   * dotprod
        vbinorms[3*ii+1]   = vbinorms[3*ii+1] - vtans[3*ii+1] * dotprod2 - vnorms[3*ii+1] * dotprod
        vbinorms[3*ii+2]   = vbinorms[3*ii+2] - vtans[3*ii+2] * dotprod2 - vnorms[3*ii+2] * dotprod

    # Now we have to renormalize AGAIN our vbinorms.
    for ii in range( nvert ) :
        norm               = vbinorms[3*ii]*vbinorms[3*ii] + vbinorms[3*ii+1]*vbinorms[3*ii+1] + vbinorms[3*ii+2]*vbinorms[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0 :
            vbinorms[3*ii]   = vbinorms[3*ii]   / norm
            vbinorms[3*ii+1] = vbinorms[3*ii+1] / norm
            vbinorms[3*ii+2] = vbinorms[3*ii+2] / norm


    # Save data.
    tangentspacedatalist.append( vnorms )
    tangentspacedatalist.append( vtans )
    tangentspacedatalist.append( vbinorms )
    tangentspacedatalist.append( vrefs )

    return  


# -----------------------------------------------------------------------------------
#    This def writes data out in ASCII SMD format.  
# -----------------------------------------------------------------------------------
def writesmdfile( fn, vnorms ) :
    # globals.
    global trianglelist
    global vertexdatalist
    global bonestringslist

    uvcoords               = vertexdatalist[0]
    vwts                   = vertexdatalist[1]
    vvecs                  = vertexdatalist[2]
    vbones                 = vertexdatalist[3]

    # Open filename for writing ASCII.
    fidsmd                 = open( fn, 'w' )

    # Required tag on first line.
    fidsmd.write( 'version 1\n' )

    # Write out the nodes or bones.
    fidsmd.write( 'nodes\n' )
    nitems                 = len( bonestringslist )
    nbones                 = nitems / 2
    parentid               = -1
    for ii in range( nbones-1 ) :
        bonestring         = bonestringslist[2*ii]
        id                 = bonestringslist[2*ii+1]
        fidsmd.write( str( id ) + ' "' + bonestring + '" ' + str( parentid ) + '\n' )
        parentid           = id

    fidsmd.write( 'end\n' )

    # Write out the skeleton pose data.
    fidsmd.write( 'skeleton\n' )
    fidsmd.write( 'time 0\n' )

    for ii in range( nbones - 1 ) :
        id                 = bonestringslist[2*ii+1]
        fidsmd.write( str(id) + ' ' + str( 0.0 ) + ' ' + str( 0.0 ) + ' ' + str( 0.0 ) + ' ' + str( 0.0 ) + ' ' + str( 0.0 ) + ' ' + str( 0.0 ) + '\n' )

    fidsmd.write( 'end\n' )
        
    # Okay, the only dicey part is pulling out indices from tris 
    # and replacing them with the actual vectors.
    fidsmd.write( 'triangles\n' )
    nitems                 = len( trianglelist )
    ntrilists              = nitems / 3
    for ii in range( ntrilists ) :
        trilist            = trianglelist[3*ii+2]
        nitems             = len( trilist )
        ntris              = nitems / 3
#        print 'ntris = ' + str( ntris )
        for jj in range( ntris ) :
            idx1           = trilist[3*jj]        # Curse Python for not having a built-in 2-D array!
            idx2           = trilist[3*jj+1]
            idx3           = trilist[3*jj+2]
            fidsmd.write( 'texturemap.tga\n' )
            # First vertex.
            vecx           = vvecs[3*idx1]
            vecy           = vvecs[3*idx1+1]
            vecz           = vvecs[3*idx1+2]
            wtprimary      = vwts[2*idx1]
            wtsecondary    = vwts[2*idx1+1]
            vnormx         = vnorms[3*idx1]
            vnormy         = vnorms[3*idx1+1]
            vnormz         = vnorms[3*idx1+2]
            U              = uvcoords[2*idx1]
            V              = uvcoords[2*idx1+1]
            boneIDsecondary= vbones[4*idx1+1]
            boneIDprimary  = vbones[4*idx1+2]
            outstring      = str(boneIDprimary) + ' ' + str(vecy) + ' ' + str(vecz) + ' ' + str(vecx) + ' ' + str(vnormx) + ' ' + str(vnormy) + ' ' + str(vnormz) + ' '
            outstring      = outstring + str(U) + ' ' + str(V) + ' ' + str(1) + ' ' + str(boneIDsecondary) + ' ' + str(wtsecondary) + '\n'
            fidsmd.write( outstring ) 

            # Second vertex.
            vecx           = vvecs[3*idx2]
            vecy           = vvecs[3*idx2+1]
            vecz           = vvecs[3*idx2+2]
            wtprimary      = vwts[2*idx2]
            wtsecondary    = vwts[2*idx2+1]
            vnormx         = vnorms[3*idx2]
            vnormy         = vnorms[3*idx2+1]
            vnormz         = vnorms[3*idx2+2]
            U              = uvcoords[2*idx2]
            V              = uvcoords[2*idx2+1]
            boneIDsecondary= vbones[4*idx2+1]
            boneIDprimary  = vbones[4*idx2+2]
            outstring      = str(boneIDprimary) + ' ' + str(vecy) + ' ' + str(vecz) + ' ' + str(vecx) + ' ' + str(vnormx) + ' ' + str(vnormy) + ' ' + str(vnormz) + ' '
            outstring      = outstring + str(U) + ' ' + str(V) + ' ' + str(1) + ' ' + str(boneIDsecondary) + ' ' + str(wtsecondary) + '\n'
            fidsmd.write( outstring ) 

            # Third vertex.
            vecx           = vvecs[3*idx3]
            vecy           = vvecs[3*idx3+1]
            vecz           = vvecs[3*idx3+2]
            wtprimary      = vwts[2*idx3]
            wtsecondary    = vwts[2*idx3+1]
            vnormx         = vnorms[3*idx3]
            vnormy         = vnorms[3*idx3+1]
            vnormz         = vnorms[3*idx3+2]
            U              = uvcoords[2*idx3]
            V              = uvcoords[2*idx3+1]
            boneIDsecondary= vbones[4*idx3+1]
            boneIDprimary  = vbones[4*idx3+2]
            outstring      = str(boneIDprimary) + ' ' + str(vecy) + ' ' + str(vecz) + ' ' + str(vecx) + ' ' + str(vnormx) + ' ' + str(vnormy) + ' ' + str(vnormz) + ' '
            outstring      = outstring + str(U) + ' ' + str(V) + ' ' + str(1) + ' ' + str(boneIDsecondary) + ' ' + str(wtsecondary) + '\n'
            fidsmd.write( outstring ) 


    fidsmd.write( 'end\n' )
    fidsmd.close()

    return


# -----------------------------------------------------------------------------------
#    This def writes data out in ASCII obj format.  Only the vertices, uv coords, 
#    and the triangles are supported.                                        
# -----------------------------------------------------------------------------------
def writeobjfile( fn ) :
    # globals.
    global trianglelist
    global vertexdatalist

    # Open it for writing ASCII.
    fidobj                 = open( fn, 'w' )

    # Vertices first.
    verts                  = vertexdatalist[2]
    nverts                 = len( verts ) / 3
    for ii in range(nverts) :
        fidobj.write( 'v ' + str( verts[3*ii] ) + ' ' + str( verts[3*ii+1] ) + ' ' + str( verts[3*ii+2] ) + '\n' )

    # Now uv coords.
    uvcoords               = vertexdatalist[0]
    ncoords                = len( uvcoords ) / 2
    for ii in range(ncoords) :
        fidobj.write( 'vt ' + str( uvcoords[2*ii] ) + ' ' + str( uvcoords[2*ii+1] ) + '\n' )

    # Finally the triangles with group names but leaving off the group types.
    # Note that the obj format wants 1-based indices so we add one to the triangle indices.
    numlen                 = len( trianglelist )
    numgroups              = numlen / 3
    for ii in range(numgroups) :
        gtype              = trianglelist[3*ii]
        gname              = trianglelist[3*ii+1]
        triarr             = trianglelist[3*ii+2]
        numel              = len( triarr )
        numtris            = numel / 3
        fidobj.write( 'g ' + gname + '\n' )
        for jj in range(numtris) :
            fidobj.write( 'f ' + str( triarr[3*jj]+1 ) + '/' + str( triarr[3*jj]+1 ) + '  ' + str( triarr[3*jj+1]+1 ) + '/' + str( triarr[3*jj+1]+1 ) + ' ' + str( triarr[3*jj+2]+1 ) + '/' + str( triarr[3*jj+2]+1 ) + '\n' )

    # C'est fini!
    fidobj.close()

    return


# -----------------------------------------------------------------------------------
#    This def writes data from an external ASCII skeleton file to the binary ms3d format.  
# -----------------------------------------------------------------------------------
def writeskeleton2ms3dfile( fidms3d, fnskeleton ) :
    global nNumJoints

    # Have to count the number of entries first.
    fidskel                = open( fnskeleton, 'r' )

    ilines                 = 0
    for eachline in fidskel :
        token              = eachline.split(',')
        if len(token) > 1 :
            ilines         = ilines + 1
        else :
            break

    fidskel.close()

    # Output short with bone count.
    putushort( ilines, fidms3d )
    nNumJoints             = ilines
    print 'number of bones = ' + str( ilines ) 


    # Reopen again to actually output the bones.
    # Use the csv (Comma Separated Values) reader to make dealing with bone names with spaces easier.
    fidskel                = open( fnskeleton, 'r' ) 

    for eachline in fidskel :

        # Tokenize eachline for processing.
        token                  = eachline.split(',')
        ntoks                  = len( token )
        if ntoks < 2 :
            break

        # Get bone_name and parent_name.
        bonename               = token[0].strip()
        parentname             = token[1].strip()
        print 'bonename =' + bonename + ', parentname =' + parentname

        # Get rotation angles.
        xangle                 = float( token[2] )
        yangle                 = float( token[3] )
        zangle                 = float( token[4] )

        # Get positions.
        xpos                   = float( token[5] )
        ypos                   = float( token[6] )
        zpos                   = float( token[7] )

        # Get time.
        rottime                = float( token[8] )
        transtime              = float( token[9] )

        # Now we're ready to write.
        putubyte( 8, fidms3d )                             # Dirty flag.
        fidms3d.write( bonename )
        putzerobytes( 32-len( bonename ), fidms3d )
        if parentname == 'none' :
            putzerobytes( 32, fidms3d )                    # Blank string means no parent.
        else :
            fidms3d.write( parentname )
            putzerobytes( 32-len( parentname), fidms3d )

        putfloat( xangle,     fidms3d )
        putfloat( yangle,     fidms3d )
        putfloat( zangle,     fidms3d )
        putfloat( xpos,       fidms3d )
        putfloat( ypos,       fidms3d )
        putfloat( zpos,       fidms3d )
        putushort( 1,         fidms3d )
        putushort( 1,         fidms3d )
        putfloat( rottime,    fidms3d )
        putfloat( 0.0,        fidms3d )
        putfloat( 0.0,        fidms3d )
        putfloat( 0.0,        fidms3d )
        putfloat( transtime,  fidms3d )
        putfloat( 0.0,        fidms3d )
        putfloat( 0.0,        fidms3d )
        putfloat( 0.0,        fidms3d )


    # Done with this, close the skeleton file.
    fidskel.close()

    return 


# -----------------------------------------------------------------------------------
#    This def writes data out in binary ms3d (1.8.0) format.  
# -----------------------------------------------------------------------------------
def writems3dfile( fn, vnorms, vrefs ) :

    # Globals.
    global ms3dbytecount
    global trianglelist
    global vertexdatalist
    global bonestringslist
    global texturefilelist
    global nNumJoints
    global variantboundingsphere
    global VARIANTMESHFLAG
    global TANGENTSPACEFLAG

    STOREMESHBONESTRINGS   = False                         # Determines if we store bone strings in the model comments.

    uvcoords               = vertexdatalist[0]
    vwts                   = vertexdatalist[1]
    vvecs                  = vertexdatalist[2]
    vbones                 = vertexdatalist[3]

    # Open filename for writing binary.
    fidms3d                = open( fn, 'wb' )
    ms3dbytecount          = 0

    # -------------------------------------------------------------------------------
    # Required tag and version number to start the file.                            |
    # -------------------------------------------------------------------------------
    fidms3d.write( 'MS3D000000' )
    ms3dbytecount          = ms3dbytecount + 10
    putuint( 4, fidms3d )

    # -------------------------------------------------------------------------------
    # Vertex data. This includes flags, boneId, and reference count.                |
    # -------------------------------------------------------------------------------
    numentries             = len( vvecs )
    nvert                  = numentries / 3
    putushort( nvert, fidms3d )                            # Write number of vertices. 
    for ii in range(nvert) :
        putubyte( 1, fidms3d )                             # Flag is 1 for selected.
        putfloat(-vvecs[3*ii+0], fidms3d )                 # Vertex vector, x, y, and z. NOTE THE MINUS SIGN ON x!!!!!
        putfloat( vvecs[3*ii+1], fidms3d )
        putfloat( vvecs[3*ii+2], fidms3d )
        putbyte( vbones[4*ii+2], fidms3d )                 # Bone Id (we use the primary).
#        putbyte( vrefs[ii], fidms3d )                      # Reference count, we do this in computing the vertex normals.
        putbyte( 1, fidms3d )                              # GrumpyOldMan's value.


    # -------------------------------------------------------------------------------
    # Triangle data as one large block.                                             |
    # -------------------------------------------------------------------------------
    numentries             = len( trianglelist )
    num_groups             = numentries / 3
    ntriangles             = 0
    for ii in range( num_groups ) :                        # Count number of triangles. 
        tris               = trianglelist[3*ii+2]
        ntriangles         = ntriangles + len(tris) / 3

    putushort( ntriangles, fidms3d )                       # Write out number of triangles.
    for jj in range( num_groups ) :
        tris               = trianglelist[3*jj+2]
        numentries         = len( tris )
        ntris              = numentries / 3
        for ii in range(ntris) :
            putushort( 1, fidms3d )                  # Flag is 1 for selected (a short here). 
            idx1           = tris[3*ii]              # Index of triangle's first vertex.
            idx2           = tris[3*ii+1]            # Index of triangle's second vertex.
            idx3           = tris[3*ii+2]            # Index of triangle's third vertex.
            putushort( idx1, fidms3d )               
            putushort( idx3, fidms3d )               # NOTE THAT idx3 and idx2 ARE FLIPPED!!!!!!!!!
            putushort( idx2, fidms3d )               
            putfloat(-vnorms[3*idx1],   fidms3d )    # x value of first vertex vector.
            putfloat( vnorms[3*idx1+1], fidms3d )    # y value of first vertex vector.
            putfloat( vnorms[3*idx1+2], fidms3d )    # z value of first vertex vector.
            putfloat(-vnorms[3*idx2],   fidms3d )    # x value of second vertex vector.
            putfloat( vnorms[3*idx2+1], fidms3d )    # y value of second vertex vector.
            putfloat( vnorms[3*idx2+2], fidms3d )    # z value of second vertex vector.
            putfloat(-vnorms[3*idx3],   fidms3d )    # x value of third vertex vector.
            putfloat( vnorms[3*idx3+1], fidms3d )    # y value of third vertex vector.
            putfloat( vnorms[3*idx3+2], fidms3d )    # z value of third vertex vector.
            putfloat( uvcoords[2*idx1], fidms3d )    # u or s value for first vertex.     NOTE: SWITCH UV COORDS TOO!
            putfloat( uvcoords[2*idx3], fidms3d )    # u or s value for second vertex.
            putfloat( uvcoords[2*idx2], fidms3d )    # u or s value for third vertex.
            putfloat( uvcoords[2*idx1+1], fidms3d )  # v or t value for first vertex.
            putfloat( uvcoords[2*idx3+1], fidms3d )  # v or t value for second vertex.
            putfloat( uvcoords[2*idx2+1], fidms3d )  # v or t value for third vertex.
            putubyte( 1, fidms3d )                   # Smoothing group?
            putubyte( jj, fidms3d )                  # Group index.

    # -------------------------------------------------------------------------------
    # Group data. This organizes groups of triangles like in the mesh.              |
    # -------------------------------------------------------------------------------
    putushort( num_groups, fidms3d ) 

    ntriidx                = 0
    for jj in range(num_groups) :
        grouptype          = trianglelist[3*jj+0]
        groupname          = trianglelist[3*jj+1]
        tris               = trianglelist[3*jj+2]
        numentries         = len( tris )
        ntris              = numentries / 3
        putubyte( 1, fidms3d )                       # Flag, 1 for selected.
        fidms3d.write( groupname )
        nname              = len( groupname )
        putzerobytes( 32-nname, fidms3d )
        ms3dbytecount      = ms3dbytecount + 32   
        putushort( ntris, fidms3d )
        for ii in range( ntris ) :
            putushort( ntriidx, fidms3d )
            ntriidx        = ntriidx + 1

        nprimary           = grouptype.find( 'primary' )
        nshield            = grouptype.find( 'shield' )
       
        if ( nprimary > -1  ) | ( nshield > -1 ) :
            putbyte( 1, fidms3d )                          # 1-index Attachments material.
        else :
            putbyte( 0, fidms3d )                          # 0-index Figure material.

    # -------------------------------------------------------------------------------
    # Materials section.  We follow GOM in this.                                    |
    # -------------------------------------------------------------------------------
    putushort( 2, fidms3d )
    # Figure material.
    fidms3d.write( 'Figure' )
    putzerobytes( 32-6, fidms3d )
    ms3dbytecount          = ms3dbytecount + 32   
    putfloat( 1.0, fidms3d )                               # ambient (4 floats)
    putfloat( 1.0, fidms3d )
    putfloat( 1.0, fidms3d )
    putfloat( 1.0, fidms3d )

    putfloat( 0.8, fidms3d )                               # diffuse (4 floats)
    putfloat( 0.8, fidms3d )
    putfloat( 0.8, fidms3d )
    putfloat( 1.0, fidms3d )

    putfloat( 0.0, fidms3d )                               # specular (4 floats)
    putfloat( 0.0, fidms3d )
    putfloat( 0.0, fidms3d )
    putfloat( 1.0, fidms3d )

    putfloat( 0.0, fidms3d )                               # emissive (4 floats)
    putfloat( 0.0, fidms3d )
    putfloat( 0.0, fidms3d )
    putfloat( 1.0, fidms3d )

    putfloat( 0.0, fidms3d )                               # shininess

    putfloat( 1.0, fidms3d )                               # transparency

    putbyte(  0.0, fidms3d )                               # mode (unused)

    putzerobytes( 128, fidms3d )                           # texture map 
    putzerobytes( 128, fidms3d )                           # alpha map
    ms3dbytecount          = ms3dbytecount + 256  

    # Attachment material.
    fidms3d.write( 'Attachments' )
    putzerobytes( 32-11, fidms3d )
    ms3dbytecount          = ms3dbytecount + 32   
    putfloat( 1.0, fidms3d )                               # ambient (4 floats)
    putfloat( 1.0, fidms3d )
    putfloat( 1.0, fidms3d )
    putfloat( 1.0, fidms3d )

    putfloat( 0.8, fidms3d )                               # diffuse (4 floats)
    putfloat( 0.8, fidms3d )
    putfloat( 0.8, fidms3d )
    putfloat( 1.0, fidms3d )

    putfloat( 0.0, fidms3d )                               # specular (4 floats)
    putfloat( 0.0, fidms3d )
    putfloat( 0.0, fidms3d )
    putfloat( 1.0, fidms3d )

    putfloat( 0.0, fidms3d )                               # emissive (4 floats)
    putfloat( 0.0, fidms3d )
    putfloat( 0.0, fidms3d )
    putfloat( 1.0, fidms3d )

    putfloat( 0.0, fidms3d )                               # shininess

    putfloat( 1.0, fidms3d )                               # transparency

    putbyte(  0.0, fidms3d )                               # mode (unused)

    putzerobytes( 128, fidms3d )                           # texture map 
    putzerobytes( 128, fidms3d )                           # alpha map
    ms3dbytecount          = ms3dbytecount + 256  


    # -------------------------------------------------------------------------------
    # Animation data.                                                               |
    # -------------------------------------------------------------------------------
    putfloat( 1.0, fidms3d )                               # Animation FPS.
    putfloat( 0.0, fidms3d )                               # Current time.
    putuint( 1, fidms3d )                                  # Number of frames.

    # -------------------------------------------------------------------------------
    # Joints section.  Just read in external .skems3d file as appropriate:          |
    #                  skeleton_standard.skems3d for regular units.                 |
    #                  skeleton_horse.skems3d for horse mounts.                     |
    #                  skeleton_camel.skems3d for camel mounts.                     |
    #                  skeleton_elephant.skems3d for elephant mounts.               |
    # -------------------------------------------------------------------------------
    nmount                 = fn.find( 'mount_' )
    ncamel                 = fn.find( 'mount_camel' )
    nelephant              = fn.find( 'mount_elephant' )
    ncatapult              = fn.find( 'catapult.ms3d' )
    ncatapult1             = fn.find( 'catapult_med.ms3d' )
    ncatapult2             = fn.find( 'catapult_low.ms3d' )
    nbasilisk              = fn.find( 'basilisk_idle.ms3d' )
    nbasilisk1             = fn.find( 'basilisk_med.ms3d' )
    nbasilisk2             = fn.find( 'basilisk_low.ms3d' )
    nbombard               = fn.find( '\\bombard.ms3d' )
    nbombard1              = fn.find( '\\bombard_med.ms3d' )
    nbombard2              = fn.find( '\\bombard_low.ms3d' )
    nculverin              = fn.find( 'culverin.ms3d' )
    nculverin1             = fn.find( 'culverin_med.ms3d' )
    nculverin2             = fn.find( 'culverin_low.ms3d' )
    ngrand_bombard         = fn.find( 'grand_bombard_rig.ms3d' )
    ngrand_bombard1        = fn.find( 'grand_bombard_med.ms3d' )
    ngrand_bombard2        = fn.find( 'grand_bombard_low.ms3d' )
    ngreat_cross_standard  = fn.find( 'great_cross_standard.ms3d' )
    ngreat_cross_standard1 = fn.find( 'great_cross_standard_med.ms3d' )
    ngreat_cross_standard2 = fn.find( 'great_cross_standard_low.ms3d' )
    nram_model             = fn.find( 'ram_model.ms3d' )
    nram_model1            = fn.find( 'ram_model_med.ms3d' )
    nram_model2            = fn.find( 'ram_model_low.ms3d' )
    nrocket_launcher       = fn.find( 'rocket_launcher.ms3d' )
    nrocket_launcher1      = fn.find( 'rocket_launcher_med.ms3d' )
    nrocket_launcher2      = fn.find( 'rocket_launcher_low.ms3d' )
    nsiege_ballista        = fn.find( 'siege_ballista.ms3d' )
    nsiege_ballista1       = fn.find( 'siege_ballista_med.ms3d' )
    nsiege_ballista2       = fn.find( 'siege_ballista_low.ms3d' )
    nsiege_bombard         = fn.find( 'siege_bombard.ms3d' )
    nsiege_bombard1        = fn.find( 'siege_bombard_med.ms3d' )
    nsiege_bombard2        = fn.find( 'siege_bombard_low.ms3d' )
    nsiege_cannon          = fn.find( 'siege_cannon.ms3d' )
    nsiege_cannon1         = fn.find( 'siege_cannon_med.ms3d' )
    nsiege_cannon2         = fn.find( 'siege_cannon_low.ms3d' )
    nsiege_huge_bombard    = fn.find( 'siege_huge_bombard.ms3d' )
    nsiege_huge_bombard1   = fn.find( 'siege_huge_bombard_med.ms3d' )
    nsiege_huge_bombard2   = fn.find( 'siege_huge_bombard_low.ms3d' )
    nsiege_monster_ribault  = fn.find( 'siege_monster_ribault.ms3d' )
    nsiege_monster_ribault1 = fn.find( 'siege_monster_ribault_med.ms3d' )
    nsiege_monster_ribault2 = fn.find( 'siege_monster_ribault_low.ms3d' )
    nsiege_mortar          = fn.find( 'siege_mortar.ms3d' )
    nsiege_mortar1         = fn.find( 'siege_mortar_med.ms3d' )
    nsiege_mortar2         = fn.find( 'siege_mortar_low.ms3d' )
    nsiege_serpentine      = fn.find( 'siege_serpentine.ms3d' )
    nsiege_serpentine1     = fn.find( 'siege_serpentine_med.ms3d' )
    nsiege_serpentine2     = fn.find( 'siege_serpentine_low.ms3d' )
    nsiege_tower_big       = fn.find( 'siege_tower_big.ms3d' )
    nsiege_tower_small     = fn.find( 'siege_tower_small.ms3d' )
    nsiege_tower_town_high = fn.find( 'siege_tower_town_high.ms3d' )
    nsiege_trebuchet       = fn.find( 'siege_trebuchet.ms3d' )
    nsiege_trebuchet1      = fn.find( 'siege_trebuchet_med.ms3d' )
    nsiege_trebuchet2      = fn.find( 'siege_trebuchet_low.ms3d' )
    nsmall_ribault         = fn.find( 'small_ribault.ms3d' )
    nsmall_ribault1        = fn.find( 'small_ribault_med.ms3d' )
    nsmall_ribault2        = fn.find( 'small_ribault_low.ms3d' )
    if ( nmount > 1 ) & ( ncamel > 1 ) :
        fnskeleton         = 'skeleton_camel.skems3d' 
    elif ( nmount > 1 ) & ( nelephant > 1 ) : 
        fnskeleton         = 'skeleton_elephant.skems3d' 
    elif nmount > 1 :
        fnskeleton         = 'skeleton_horse.skems3d' 
    elif ( ncatapult > -1 ) | ( ncatapult1 > -1 ) | ( ncatapult2 > -1 ) :
        fnskeleton         = 'skeleton_catapult.skems3d' 
    elif ( nbasilisk > -1 ) | ( nbasilisk1 > -1 ) | ( nbasilisk2 > -1 ) :
        fnskeleton         = 'skeleton_basilisk.skems3d' 
    elif ( nbombard > -1 ) | ( nbombard1 > -1 ) | ( nbombard2 > -1 ) :
        fnskeleton         = 'skeleton_bombard.skems3d' 
    elif ( nculverin > -1 ) | ( nculverin1 > -1 ) | ( nculverin2 > -1 ) :
        fnskeleton         = 'skeleton_culverin.skems3d' 
    elif ( ngrand_bombard > -1 ) | ( ngrand_bombard1 > -1 ) | ( ngrand_bombard2 > -1 ) :
        fnskeleton         = 'skeleton_great_bombard.skems3d' 
    elif ( ngreat_cross_standard > -1 ) | ( ngreat_cross_standard1 > -1 ) | ( ngreat_cross_standard2 > -1 ) :
        fnskeleton         = 'skeleton_great_cross_standard.skems3d' 
    elif ( nram_model > -1 ) | ( nram_model1 > -1 ) | ( nram_model2 > -1 ) :
        fnskeleton         = 'skeleton_siege_ram.skems3d' 
    elif ( nrocket_launcher > -1 ) | ( nrocket_launcher1 > -1 ) | ( nrocket_launcher2 > -1 ) :
        fnskeleton         = 'skeleton_siege_rocket_launcher.skems3d' 
    elif ( nsiege_ballista > -1 ) | ( nsiege_ballista1 > -1 ) | ( nsiege_ballista2 > -1 ) :
        fnskeleton         = 'skeleton_siege_ballista.skems3d' 
    elif ( nsiege_bombard > -1 ) | ( nsiege_bombard1 > -1 ) | ( nsiege_bombard2 > -1 ) :
        fnskeleton         = 'skeleton_siege_bombard.skems3d' 
    elif ( nsiege_cannon > -1 ) | ( nsiege_cannon1 > -1 ) | ( nsiege_cannon2 > -1 ) :
        fnskeleton         = 'skeleton_siege_cannon.skems3d' 
    elif ( nsiege_huge_bombard > -1 ) | ( nsiege_huge_bombard1 > -1 ) | ( nsiege_huge_bombard2 > -1 ) :
        fnskeleton         = 'skeleton_siege_huge_bombard.skems3d' 
    elif ( nsiege_monster_ribault > -1 ) | ( nsiege_monster_ribault1 > -1 ) | ( nsiege_monster_ribault2 > -1 ) :
        fnskeleton         = 'skeleton_siege_monster_ribault.skems3d' 
    elif ( nsiege_mortar > -1 ) | ( nsiege_mortar1 > -1 ) | ( nsiege_mortar2 > -1 ) :
        fnskeleton         = 'skeleton_siege_mortar.skems3d' 
    elif ( nsiege_serpentine > -1 ) | ( nsiege_serpentine1 > -1 ) | ( nsiege_serpentine2 > -1 ) :
        fnskeleton         = 'skeleton_siege_serpentine.skems3d' 
    elif ( nsiege_tower_big > -1 ) :
        fnskeleton         = 'skeleton_siege_tower_big.skems3d' 
    elif ( nsiege_tower_small > -1 ) :
        fnskeleton         = 'skeleton_siege_tower_small.skems3d' 
    elif ( nsiege_tower_town_high > -1 ) :
        fnskeleton         = 'skeleton_siege_tower_town.skems3d' 
    elif ( nsiege_trebuchet > -1 ) | ( nsiege_trebuchet1 > -1 ) | ( nsiege_trebuchet2 > -1 ) :
        fnskeleton         = 'skeleton_siege_trebuchet.skems3d' 
    elif ( nsmall_ribault > -1 ) | ( nsmall_ribault1 > -1 ) | ( nsmall_ribault2 > -1 ) :
        fnskeleton         = 'skeleton_small_ribault.skems3d' 
    else :
        fnskeleton         = 'skeleton_standard.skems3d' 

    print 'Using skeleton: ' + fnskeleton

    # Writes joint section the way Milkshape 3D wants it.
    writeskeleton2ms3dfile( fidms3d, fnskeleton ) 

    # -------------------------------------------------------------------------------
    # SubVersion entry, must be 1.                                                  |
    # -------------------------------------------------------------------------------
    putuint( 1 , fidms3d )
    print 'Writing subversion entry 1, tell says ' + str( fidms3d.tell() )

    # -------------------------------------------------------------------------------
    # Group comments. This is where the triangle group types and names are stored.  |
    # -------------------------------------------------------------------------------
    putuint( num_groups, fidms3d )
    for jj in range(num_groups) :
        typename           = trianglelist[3*jj]
        groupname          = trianglelist[3*jj+1]
        entrystring        = typename + '\r\n' + groupname + '\r\n' + '0'
        nchars             = len( entrystring )
        putuint( jj, fidms3d )
        putuint( nchars, fidms3d )
        fidms3d.write( entrystring )
        ms3dbytecount      = ms3dbytecount + nchars

    print 'Finished writing group comments, tell says ' + str( fidms3d.tell() )

    # -------------------------------------------------------------------------------
    # Material comments.  We don't use these.                                       |
    # -------------------------------------------------------------------------------
    putuint( 0, fidms3d )
    print 'Finished writing no material comments, tell says ' + str( fidms3d.tell() )

    # -------------------------------------------------------------------------------
    # Joint comments.  We don't use these.                                          |
    # -------------------------------------------------------------------------------
    if VARIANTMESHFLAG == False :
        putuint( 0, fidms3d )
        print 'Finished writing no joint comments, tell says ' + str( fidms3d.tell() )
    else :
        putuint( 3, fidms3d )
        putuint( 0, fidms3d )                           # 0-index, first texturefile. 
        texfile1           = texturefilelist[0]
        nch                = len( texfile1 )
        putuint( nch, fidms3d )
        fidms3d.write( texfile1 )
        putuint( 1, fidms3d )                           # 1-index, second texturefile. 
        texfile2           = texturefilelist[1]
        nch                = len( texfile2 )
        putuint( nch, fidms3d )
        fidms3d.write( texfile2 )
        putuint( 2, fidms3d )                           # 2-index, third texturefile. 
        texfile3           = texturefilelist[2]
        nch                = len( texfile3 )
        putuint( nch, fidms3d )
        fidms3d.write( texfile3 )

#    putuint( 0, fidms3d )                              # Don't know what this is but ms3d wants it.
    # -------------------------------------------------------------------------------
    # Model comments. Store boundingsphere data in '/' delimited string.  Option    |
    # to also store the bone strings as they come out of the .mesh file.            |
    # -------------------------------------------------------------------------------
    nitems                 = len( bonestringslist )
    print 'nitems = ' + str( nitems )
    nstrings               = ( nitems - 2 ) / 2
    print 'nstrings = ' + str( nstrings )
    if VARIANTMESHFLAG == False :
        boundingsphere     = bonestringslist[nitems-1]
    else :
        boundingsphere     = variantboundingsphere
    f0                     = str( boundingsphere[0] )
    f1                     = str( boundingsphere[1] )
    f2                     = str( boundingsphere[2] )
    f3                     = str( boundingsphere[3] )
    if VARIANTMESHFLAG == False :
        boundingspherestring   = f0 + '\r\n' + f1 + '\r\n' + f2 + '\r\n' + f3 + '\r\n' + str( TANGENTSPACEFLAG ) + '\r\n/'
    else :
        boundingspherestring   = f0 + '\r\n' + f1 + '\r\n' + f2 + '\r\n' + f3 + '\r\n' + str( 4 ) + '\r\n/'
    nchars                 = len( boundingspherestring )
    print 'boundingsphere string = ' + boundingspherestring + ', of length ' + str( nchars )

    if STOREMESHBONESTRINGS == False :
        putuint( 1, fidms3d )                              # Only have the boundingsphere string.
#        putuint( 0, fidms3d )                              # Index.
        putuint( nchars, fidms3d )                         # Char count.
        fidms3d.write( boundingspherestring )              # Floats as / delimited string.
    else :
        putuint( nstrings + 1, fidms3d )                   # nstrings comments for bone list + 1 comment to hold bounding sphere.

        # First model comment is boundingsphere string.
        putuint( 0, fidms3d )                              # Index.
        putuint( nchars, fidms3d )                         # Char count.
        fidms3d.write( boundingspherestring ) 

        # Bone strings and node number? stuff from the mesh just as it came out.
        for ii in range( nstrings ) :
            bonename       = bonestringslist[2*ii]         # bone name.
            cnt            = bonestringslist[2*ii+1]       # count number that follows it.
            nchars         = len( bonename )               # characters in bone name.
            commentlength  = 4 + nchars + 4                # int for nchars, the bonename in chars, and int for cnt.
            putuint(    0, fidms3d )                       # index (not used).
            putuint( commentlength, fidms3d )              # length of comment.
            putuint( nchars, fidms3d )                     # Note that we store the string count as well as the bone name
            fidms3d.write( bonename )                      # and node number, just like the mesh format.
            putuint( cnt, fidms3d )
            print 'commentlength = ' + str( commentlength )
            print 'nchars        = ' + str( nchars )
            print 'bonename      = ' + bonename
            print 'cnt           = ' + str( cnt )

    # -------------------------------------------------------------------------------
    # Second subversion number, must be 2.                                          |
    # -------------------------------------------------------------------------------
    subversion             = 2
    putuint( subversion, fidms3d )                                  

    # -------------------------------------------------------------------------------
    # Vertex weighting.                                                             |
    # -------------------------------------------------------------------------------
    for ii in range(nvert) :
        wtprim             = vwts[2*ii]
        wtsecond           = vwts[2*ii+1]
        boneIdprim         = vbones[4*ii+2]
        boneIdsecond       = vbones[4*ii+1]
        putbyte( boneIdsecond, fidms3d )                   # Secondary bone Id.
        putbyte( -1, fidms3d )                             # Not used, only 2 weightings allowed in .mesh file.
        putbyte( -1, fidms3d )                             # Not used, only 2 weightings allowed in .mesh file. 
        wtprimpercent      = int( 100.0 * wtprim )
        wtsecondpercent    = int( 100.0 * wtsecond )
        putbyte( wtprimpercent, fidms3d )
        putbyte( wtsecondpercent, fidms3d )
        putbyte( 0, fidms3d )
        if subversion == 2 :
            putuint( 0, fidms3d )                          # This is version 2 extra int field.

    # -------------------------------------------------------------------------------
    # Second subversion number, must be 2.                                          |
    # -------------------------------------------------------------------------------
#    putuint( 2, fidms3d )                                  

    # -------------------------------------------------------------------------------
    # Joint colors.                                                                 |
    # -------------------------------------------------------------------------------
#    for ii in range(nNumJoints) :
#        putfloat( 0.0, fidms3d )
#        putfloat( 0.0, fidms3d )
#        putfloat( 1.0, fidms3d )
     

    fidms3d.close()

    return


# -----------------------------------------------------------------------------------
#    This def reads data out of the binary ms3d format and converts it to the  
# .mesh format.  Also outputs to a text file to compare data from each way of the
# conversion.
# -----------------------------------------------------------------------------------
def ms3dtomeshconverter( fnin, fnouttxt, fnoutmesh ) :

    # Globals.
    global ms3dbytecount
    global bytecount
    global MOUNTFLAG
    global LOD0FLAG
    global TANGENTSPACEFLAG

    fout                   = open( fnouttxt, 'w' )

    # Open filename for reading.
    fidms3d                = open( fnin, 'rb' )
    bytecount              = 0

    # -------------------------------------------------------------------------------
    # Get header tag and version number.                                            |
    # -------------------------------------------------------------------------------
    header                 = fidms3d.read( 10 )
    bytecount              = bytecount + 10
    versionnum             = getuint( fidms3d )               
    string                 = header + ', version number ' + str( versionnum )
    fout.write( string + '\n' )
    fout.flush()
    print string

    # -------------------------------------------------------------------------------
    # Vertex data.                                                                  |
    # -------------------------------------------------------------------------------
    nvert                  = getushort( fidms3d )
    string                 = 'Number of vertices = ' + str( nvert )
    fout.write( string + '\n' )
    fout.flush()
    print string

    # Allocate arrays.
    vvecs                  = array.array( 'f' )
    vbones                 = array.array( 'B' )
    vnorms                 = array.array( 'f' )
    vtans                  = array.array( 'f' )
    vbinorms               = array.array( 'f' )
    # Initialize.
    for ii in range( 3 * nvert ) :
        vvecs.append( 0.0 )
        vnorms.append( 0.0 )
        vtans.append( 0.0 )
        vbinorms.append( 0.0 )
    # Initialize.
    for ii in range( 4 * nvert ) :
        vbones.append( 0 )

    # Read in vertex data. This includes flags, boneId, and reference count.
    for ii in range( nvert ) :
        flag               = getubyte( fidms3d )     # Flag value we don't care about.
        vvecs[3*ii]        = -getfloat( fidms3d )    # Vertex vector, x, y, and z
        vvecs[3*ii+1]      = getfloat( fidms3d )
        vvecs[3*ii+2]      = getfloat( fidms3d )
        vbones[4*ii+2]     = getubyte( fidms3d )     # Bone Id, put in the primary slot.
#        putbyte( vrefs[ii], fidms3d )                # Reference count.
        dummy              = getubyte( fidms3d )     # Don't care about reference count.
    

    # -------------------------------------------------------------------------------
    # Triangle data.  This is a mess because the groups are all messed up.          |
    # We also compute our vertex normals at this point using the triangles.         |
    # -------------------------------------------------------------------------------
    ntriangles             = getushort( fidms3d )
    print 'ntriangles = ' + str( ntriangles )
    tris                   = array.array( 'H' )
    s                      = array.array( 'f' )
    t                      = array.array( 'f' )
    groupidx               = array.array( 'B' )
    for ii in range( ntriangles ) :
        tris.append( 0 )
        tris.append( 0 )
        tris.append( 0 )
        s.append( 0.0 )
        s.append( 0.0 )
        s.append( 0.0 )
        t.append( 0.0 )
        t.append( 0.0 )
        t.append( 0.0 )
        groupidx.append( 0 )

    for ii in range( ntriangles ) :
        dummy              = getushort( fidms3d )
        tris[3*ii]         = getushort( fidms3d )
        tris[3*ii+2]       = getushort( fidms3d )     # Note reversed triangle vertices! Important!
        tris[3*ii+1]       = getushort( fidms3d )
        idx1               = tris[3*ii]
        idx2               = tris[3*ii+1]
        idx3               = tris[3*ii+2]
#        vnorms[3*idx1]     = -getfloat(  fidms3d )
#        vnorms[3*idx1+1]   = getfloat(  fidms3d )
#        vnorms[3*idx1+2]   = getfloat(  fidms3d )
#        vnorms[3*idx2]     = -getfloat(  fidms3d )
#        vnorms[3*idx2+1]   = getfloat(  fidms3d )
#        vnorms[3*idx2+2]   = getfloat(  fidms3d )
#        vnorms[3*idx3]     = -getfloat(  fidms3d )
#        vnorms[3*idx3+1]   = getfloat(  fidms3d )
#        vnorms[3*idx3+2]   = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        s[3*ii]            = getfloat(  fidms3d )          # NOTE THE SWITCH IN UV COORDS TO MATCH THE TRIANGLE SWITCH!
        s[3*ii+2]          = getfloat(  fidms3d )
        s[3*ii+1]          = getfloat(  fidms3d )
        t[3*ii]            = getfloat(  fidms3d )
        t[3*ii+2]          = getfloat(  fidms3d )
        t[3*ii+1]          = getfloat(  fidms3d )
        bdummy             = getubyte(  fidms3d )
        groupidx[ii]       = getubyte(  fidms3d )
        # Don't believe vertex normals, going to roll my own.
        v1x                = vvecs[3*idx1]
        v1y                = vvecs[3*idx1+1]
        v1z                = vvecs[3*idx1+2]
        v2x                = vvecs[3*idx2]
        v2y                = vvecs[3*idx2+1]
        v2z                = vvecs[3*idx2+2]
        v3x                = vvecs[3*idx3]
        v3y                = vvecs[3*idx3+1]
        v3z                = vvecs[3*idx3+2]
        edge12x            = v2x - v1x            #
        edge12y            = v2y - v1y            #   | i   j   k  |
        edge12z            = v2z - v1z            #   | ux  uy  uz | = i(uy*vz-uz*vy) + j(uz*vx-ux*vz) + k(ux*vy-uy*vx)
        edge23x            = v3x - v2x            #   | vx  vy  vz |
        edge23y            = v3y - v2y            #
        edge23z            = v3z - v2z            #
        # Take cross product to get a normal vector (normal to face, but it won't be normalized).
        unx                = edge12y * edge23z - edge12z * edge23y
        uny                = edge12z * edge23x - edge12x * edge23z
        unz                = edge12x * edge23y - edge12y * edge23x
        # Normalize it.
        norm               = unx*unx + uny*uny + unz*unz 
        norm               = norm**0.5 
        if norm != 0.0 :
            nx             = unx / norm     
            ny             = uny / norm
            nz             = unz / norm
        else :
            nx             = unx
            ny             = uny
            nz             = unz
        # Add it to vnorms for each index.
        vnorms[3*idx1]     = vnorms[3*idx1]   + nx
        vnorms[3*idx2]     = vnorms[3*idx2]   + nx
        vnorms[3*idx3]     = vnorms[3*idx3]   + nx
        vnorms[3*idx1+1]   = vnorms[3*idx1+1] + ny
        vnorms[3*idx2+1]   = vnorms[3*idx2+1] + ny
        vnorms[3*idx3+1]   = vnorms[3*idx3+1] + ny
        vnorms[3*idx1+2]   = vnorms[3*idx1+2] + nz
        vnorms[3*idx2+2]   = vnorms[3*idx2+2] + nz
        vnorms[3*idx3+2]   = vnorms[3*idx3+2] + nz


    # Now we have to normalize our vnorms.
    for ii in range( nvert ) :
    #        fidout.write( str( vnorms[3*ii] ) + ' ' + str( vnorms[3*ii+1] ) + ' ' + str( vnorms[3*ii+2] ) + '\n' )
        norm               = vnorms[3*ii]*vnorms[3*ii] + vnorms[3*ii+1]*vnorms[3*ii+1] + vnorms[3*ii+2]*vnorms[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0 :
            vnorms[3*ii]   = vnorms[3*ii]   / norm
            vnorms[3*ii+1] = vnorms[3*ii+1] / norm
            vnorms[3*ii+2] = vnorms[3*ii+2] / norm


    # -------------------------------------------------------------------------------
    # Group data.  This allows us to reconstruct the .mesh format groups.           |
    # -------------------------------------------------------------------------------
    num_groups             = getushort( fidms3d )
    print 'num_groups = ' + str( num_groups )
    group_names            = []
    tri_groups             = []
    for ii in range( num_groups ) :
        bdummy             = getubyte(  fidms3d )
        tempname           = fidms3d.read( 32 )
        bytecount          = bytecount + 32
        groupname          = zipstrip( tempname )
        print groupname
        group_names.append( groupname )

        ntris              = getushort( fidms3d )
        triidx             = array.array( 'H' )
        for jj in range( ntris ) :
            triidx.append( getushort( fidms3d ) )

        bchar              = getbyte( fidms3d )              # Don't care about material index.
        # Now reorganize the triangle data into the mesh format.
        triarr             = array.array( 'H' )
        for jj in range( ntris ) :
            idx            = triidx[jj]
            triarr.append( tris[3*idx] )
            triarr.append( tris[3*idx+1] )
            triarr.append( tris[3*idx+2] )

        tri_groups.append( triarr )

    # Print out a little face data to check.
    tritest                = tri_groups[0]
    print 'face: ' + str( tritest[0] ) + ' ' + str( tritest[1] ) + ' ' + str( tritest[2] )
    print 'face: ' + str( tritest[3] ) + ' ' + str( tritest[4] ) + ' ' + str( tritest[5] )
    print 'face: ' + str( tritest[6] ) + ' ' + str( tritest[7] ) + ' ' + str( tritest[8] )
    print 'face: ' + str( tritest[9] ) + ' ' + str( tritest[10] ) + ' ' + str( tritest[11] )


    # -------------------------------------------------------------------------------
    # Read nMaterials (the materials struct is 361 bytes).  Just skip if more than 0.
    # -------------------------------------------------------------------------------
    nMaterials             = getushort( fidms3d )
    print 'nMaterials = ' + str( nMaterials )

    if nMaterials > 0 :
        for jj in range( nMaterials ) :
            fidms3d.seek( 361, 1 )                         # This just relative skips past it.

    # -------------------------------------------------------------------------------
    # Keyframer data. (Just junk to us.)                                            |
    # -------------------------------------------------------------------------------
    fAnimationFPS          = getfloat(  fidms3d )
    fCurrentTime           = getfloat(  fidms3d )
    iTotalFrames           = getuint(   fidms3d )
    print 'fAnimationFPS = ' + str( fAnimationFPS )
    print 'fCurrentTime  = ' + str( fCurrentTime  )
    print 'iTotalFrames  = ' + str( iTotalFrames  )

    print 'Passed the iTotalFrames read, tell says' + str( fidms3d.tell() )


    # -------------------------------------------------------------------------------
    # Joints or bone data.                                                          |
    # -------------------------------------------------------------------------------
    njoints                = getushort( fidms3d )
    print 'njoints = ' + str( njoints )
    bonenames              = []
    parentnames            = []
    for ii in range( njoints ) :
        dummy              = getubyte(  fidms3d )
        tbone              = fidms3d.read( 32 )
        bonename           = zipstrip( tbone )
        pbone              = fidms3d.read( 32 )
        bytecount          = bytecount + 64
        parentname         = zipstrip( pbone )
        bonenames.append( bonename )
        parentnames.append( parentname )
        print bonename + ' from ' + parentname
        # rotation[3] float.
        fdummy             = getfloat(  fidms3d )
        fdummy             = getfloat(  fidms3d )
        fdummy             = getfloat(  fidms3d )
        # position[3] float.
        fdummy             = getfloat(  fidms3d )
        fdummy             = getfloat(  fidms3d )
        fdummy             = getfloat(  fidms3d )
        # numKeyFramesRot.
        numKeyFramesRot    = getshort(  fidms3d )
        # numKeyFramesTrans
        numKeyFramesTrans  = getshort(  fidms3d )
        for ii in range( numKeyFramesRot ) :
            fdummy         = getfloat(  fidms3d )          # Just dump the animation matrices.
            fdummy         = getfloat(  fidms3d )
            fdummy         = getfloat(  fidms3d )
            fdummy         = getfloat(  fidms3d )

        for ii in range( numKeyFramesTrans ) :
            fdummy         = getfloat(  fidms3d )          # Ditto.
            fdummy         = getfloat(  fidms3d )
            fdummy         = getfloat(  fidms3d )
            fdummy         = getfloat(  fidms3d )

    # -------------------------------------------------------------------------------
    # Subversion number (always 1).                                                 |
    # -------------------------------------------------------------------------------
    dummy                  = getuint(   fidms3d )
    print 'First subversion number = ' + str( dummy )
    print 'Passed the first subversion number, tell says' + str( fidms3d.tell() )


    # -------------------------------------------------------------------------------
    # Group comments.  This is where the group types live along with group names.   |
    # -------------------------------------------------------------------------------
    num_comments           = getuint(   fidms3d )
    typenames              = []
    for ii in range( num_comments ) :
        idx                = getuint(   fidms3d )
        nchars             = getuint(   fidms3d )
        comment            = fidms3d.read( nchars )
        bytecount          = bytecount + nchars
        nch                = comment.find( '\r\n' )
        typename           = comment[0:nch]
        typenames.append( typename )
        print typename + ' from ' + comment

    print 'Passed the group comments, tell says' + str( fidms3d.tell() )


    # -------------------------------------------------------------------------------
    # Material comments.  Not used by me, should be zero.                           |
    # -------------------------------------------------------------------------------
    nNumMaterialComments   = getuint(   fidms3d )
    print 'nNumMaterialComments = ' + str( nNumMaterialComments )

    if nNumMaterialComments > 0 :
        for ii in range( nNumMaterialComments ) :
            idummy         = getuint(   fidms3d ) 
            nchars         = getuint(   fidms3d )
            cdummy         = fidms3d.read( nchars )

    print 'Passed the material comments, tell says' + str( fidms3d.tell() )


    # -------------------------------------------------------------------------------
    # Joint comments.  Not used by me, should be zero.                              |
    # -------------------------------------------------------------------------------
    nNumJointComments      = getuint(   fidms3d )
    print 'nNumJointComments = ' + str( nNumJointComments )

    if nNumJointComments > 0 :
        for ii in range( nNumJointComments ) :
            idummy         = getuint(   fidms3d ) 
            nchars         = getuint(   fidms3d )
            cdummy         = fidms3d.read( nchars )

    print 'Passed the joint comments, tell says' + str( fidms3d.tell() )

    # -------------------------------------------------------------------------------
    # Model comments. This is where we store the bounding sphere data and,          |
    # optionally, the bone strings data that comes directly out of the .mesh file.  |
    # -------------------------------------------------------------------------------
    nHasModelComment       = getuint(   fidms3d )
    print 'nHasModelComment = ' + str( nHasModelComment )
    bonedata               = []
    boundingsphere         = array.array( 'f' )
    for ii in range( 4 ) :
        boundingsphere.append( 0.0 )

    # boundingsphere data comes as ASCII in the first comment.
#    nGOM                   = fnin.find( 'GOM' )
#    if nGOM == -1 :
#        idummy             = getuint(   fidms3d )          # GOM skips the index int here.
    nchars                 = getuint(   fidms3d )
    boundingspherestring   = fidms3d.read( nchars )
    print 'boundingspherestring = ' + boundingspherestring
    flist                  = boundingspherestring.split( '\r\n' )
    boundingsphere[0]      = float( flist[0] )
    boundingsphere[1]      = float( flist[1] )
    boundingsphere[2]      = float( flist[2] )
    boundingsphere[3]      = float( flist[3] )
    if MOUNTFLAG == True :
        TANGENTSPACEFLAG   = int(   flist[4] )

    print 'nchars            = ' + str( nchars )
    print 'boundingsphere[0] = ' + str( boundingsphere[0] )
    print 'boundingsphere[1] = ' + str( boundingsphere[1] )
    print 'boundingsphere[2] = ' + str( boundingsphere[2] )
    print 'boundingsphere[3] = ' + str( boundingsphere[3] )

    # My format has bone strings here.
    if nHasModelComment > 1 :
        for ii in range( nHasModelComment - 1 ) :
            idummy         = getuint(   fidms3d ) 
            nchars         = getuint(   fidms3d )
            nch            = getuint(   fidms3d )
            bonedata.append( nch )
            bonename       = fidms3d.read( nch )
            bonedata.append( bonename )
            cnt            = getuint(   fidms3d )
            bonedata.append( cnt )       
            print 'nch      = ' + str( nch )
            print 'bonename = ' + bonename
            print 'cnt      = ' + str( cnt )
    else :
        # Hard code the standard skeleton if need be.
        nbones             = len( bonenames )
        nodecnt            = 0
        for ii in range( nbones ) :
            bonename       = bonenames[ii]
            nch            = len( bonename )
            if bonename == 'bone_weapon01' :
                cnt        = nodecnt
            elif bonename == 'bone_weapon' :
                cnt        = nodecnt
                nodecnt    = nodecnt + 1 
            elif bonename == 'bone_weapon02' :
                cnt        = nodecnt
                nodecnt    = nodecnt + 1 
            elif bonename == 'bone_weapon03' :
                cnt        = nodecnt
            elif bonename == 'bone_shield01' :
                cnt        = nodecnt
            elif bonename == 'bone_shield' :
                cnt        = nodecnt
            else :
                cnt        = nodecnt
                nodecnt    = nodecnt + 1 
            bonedata.append( nch )
            bonedata.append( bonename )
            bonedata.append( cnt )       


    # -------------------------------------------------------------------------------
    # Vertex weighting data.                                                        |
    # -------------------------------------------------------------------------------
    nv                     = getuint(   fidms3d )
    print 'Second subversion number = ' + str( nv )
    vwts                   = array.array( 'f' )
    for ii in range( nvert ) :
        vbones[4*ii+1]     = getubyte(  fidms3d ) 
        dummy              = getubyte(  fidms3d ) 
        dummy              = getubyte(  fidms3d ) 
        if ( vbones[4*ii+1] > njoints-1 ) :
            vbones[4*ii+1] = vbones[4*ii+2]
            wtprim         = 1.0
            wtsecond       = 0.0
            dummy          = getubyte( fidms3d )
            dummy          = getubyte( fidms3d )
        else:
            wtprim         = float( getubyte(  fidms3d ) ) / 100.0   # Convert percentage to float.
            wtsecond       = float( getubyte(  fidms3d ) ) / 100.0   
            if ( wtsecond > 1.00001 ) | ( wtprim > 1.00001 ) :
                wtprim     = 1.0
                wtsecond   = 0.0
            else:
                wtprim     = wtprim + ( 1 - wtprim - wtsecond )
        vwts.append( wtprim )
        vwts.append( wtsecond )
        dummy              = getubyte(  fidms3d )
        if nv == 2 :
            dummy          = getuint(  fidms3d )                     # Version 2 has extra int field.

    # That's all the data, close now.
    fidms3d.close()

    # -------------------------------------------------------------------------------
    # Now to tidy up and package it all.  vvecs, vwts, and vbones are done          |
    # but we need uvcoords. The triangle data is already sort of organized.         |
    # -------------------------------------------------------------------------------
    uvcoords               = array.array( 'f' )
    for ii in range( nvert ) :
        uvcoords.append( 0.0 ) 
        uvcoords.append( 0.0 ) 

    for ii in range( ntriangles ) :
        idx1               = tris[3*ii]      # Indices to vertices.
        idx2               = tris[3*ii+1]
        idx3               = tris[3*ii+2]
        s1                 = s[3*ii]
        s2                 = s[3*ii+1]
        s3                 = s[3*ii+2]
        t1                 = t[3*ii]
        t2                 = t[3*ii+1]
        t3                 = t[3*ii+2]
        uvcoords[2*idx1]   = s1
        uvcoords[2*idx1+1] = t1
        uvcoords[2*idx2]   = s2
        uvcoords[2*idx2+1] = t2
        uvcoords[2*idx3]   = s3
        uvcoords[2*idx3+1] = t3

    # -------------------------------------------------------------------------------
    # First we have to calculate the tangent basis.  Then we have to do the cube    |
    # normalization calculation before we are ready to output to the .mesh format.  |
    # -------------------------------------------------------------------------------
    for ii in range( ntriangles ) :
        idx1               = tris[3*ii]      # Indices to vertices.
        idx2               = tris[3*ii+1]
        idx3               = tris[3*ii+2]
        # Edges.
        e1x                = vvecs[3*idx2]      - vvecs[3*idx1]
        e1y                = vvecs[3*idx2+1]    - vvecs[3*idx1+1]
        e1z                = vvecs[3*idx2+2]    - vvecs[3*idx1+2]
        e2x                = vvecs[3*idx3]      - vvecs[3*idx1]
        e2y                = vvecs[3*idx3+1]    - vvecs[3*idx1+1]
        e2z                = vvecs[3*idx3+2]    - vvecs[3*idx1+2]
        euv1x              = uvcoords[2*idx2]   - uvcoords[2*idx1]
        euv1y              = uvcoords[2*idx2+1] - uvcoords[2*idx1+1]
        euv2x              = uvcoords[2*idx3]   - uvcoords[2*idx1]
        euv2y              = uvcoords[2*idx3+1] - uvcoords[2*idx1+1]
        cp                 = euv1y * euv2x - euv1x * euv2y
        if cp != 0.0 :
            mul            = 1.0 / cp
            tanx           = ( -e1x * euv2y + e2x * euv1y ) * mul
            tany           = ( -e1y * euv2y + e2y * euv1y ) * mul
            tanz           = ( -e1z * euv2y + e2z * euv1y ) * mul
            norm           = tanx*tanx + tany*tany + tanz*tanz
            norm           = norm**0.5
            tanx           = tanx / norm
            tany           = tany / norm
            tanz           = tanz / norm
            vtans[3*idx1]  = vtans[3*idx1]   + tanx
            vtans[3*idx1+1]= vtans[3*idx1+1] + tany
            vtans[3*idx1+2]= vtans[3*idx1+2] + tanz
            vtans[3*idx2]  = vtans[3*idx2]   + tanx
            vtans[3*idx2+1]= vtans[3*idx2+1] + tany
            vtans[3*idx2+2]= vtans[3*idx2+2] + tanz
            vtans[3*idx3]  = vtans[3*idx3]   + tanx
            vtans[3*idx3+1]= vtans[3*idx3+1] + tany
            vtans[3*idx3+2]= vtans[3*idx3+2] + tanz

    for ii in range( nvert ) :
        norm               = vtans[3*ii]*vtans[3*ii] + vtans[3*ii+1]*vtans[3*ii+1] + vtans[3*ii+2]*vtans[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0:
            vtans[3*ii]    = vtans[3*ii]   / norm
            vtans[3*ii+1]  = vtans[3*ii+1] / norm
            vtans[3*ii+2]  = vtans[3*ii+2] / norm
        # Now orthogonalize with respect to vertex normals.
        dotprod            = vnorms[3*ii]*vtans[3*ii] + vnorms[3*ii+1]*vtans[3*ii+1]  + vnorms[3*ii+2]*vtans[3*ii+2]
        vtans[3*ii]        = vtans[3*ii]   - vnorms[3*ii]   * dotprod
        vtans[3*ii+1]      = vtans[3*ii+1] - vnorms[3*ii+1] * dotprod
        vtans[3*ii+2]      = vtans[3*ii+2] - vnorms[3*ii+2] * dotprod
        # Renormalize.
        norm               = vtans[3*ii]*vtans[3*ii] + vtans[3*ii+1]*vtans[3*ii+1] + vtans[3*ii+2]*vtans[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0:
            vtans[3*ii]    = -vtans[3*ii]   / norm
            vtans[3*ii+1]  = -vtans[3*ii+1] / norm
            vtans[3*ii+2]  = -vtans[3*ii+2] / norm
        # Finally we get the binorms by taking a cross product and we have the tangent space system.
        vbinorms[3*ii]     = -( vnorms[3*ii+1] * vtans[3*ii+2] - vnorms[3*ii+2] * vtans[3*ii+1] )
        vbinorms[3*ii+1]   = -( vnorms[3*ii+2] * vtans[3*ii]   - vnorms[3*ii]   * vtans[3*ii+2] )
        vbinorms[3*ii+2]   = -( vnorms[3*ii]   * vtans[3*ii+1] - vnorms[3*ii+1] * vtans[3*ii] )

    # Now create the normalization cube.
    bnorms                 = array.array( 'B' )
    btans                  = array.array( 'B' )
    bbinorms               = array.array( 'B' )
    for ii in range( 3*nvert ) :
        bnorms.append( 0 )
        btans.append( 0 )
        bbinorms.append( 0 )

    for ii in range( nvert ) :
        bnorms[3*ii]       = int( round( 255.0 * 0.5 * ( vnorms[3*ii]     + 1.0 ) ) )
        bnorms[3*ii+1]     = int( round( 255.0 * 0.5 * ( vnorms[3*ii+1]   + 1.0 ) ) )
        bnorms[3*ii+2]     = int( round( 255.0 * 0.5 * ( vnorms[3*ii+2]   + 1.0 ) ) )
        btans[3*ii]        = int( round( 255.0 * 0.5 * ( vtans[3*ii]      + 1.0 ) ) )
        btans[3*ii+1]      = int( round( 255.0 * 0.5 * ( vtans[3*ii+1]    + 1.0 ) ) )
        btans[3*ii+2]      = int( round( 255.0 * 0.5 * ( vtans[3*ii+2]    + 1.0 ) ) )
        bbinorms[3*ii]     = int( round( 255.0 * 0.5 * ( vbinorms[3*ii]   + 1.0 ) ) )
        bbinorms[3*ii+1]   = int( round( 255.0 * 0.5 * ( vbinorms[3*ii+1] + 1.0 ) ) )
        bbinorms[3*ii+2]   = int( round( 255.0 * 0.5 * ( vbinorms[3*ii+2] + 1.0 ) ) )


    # ===============================================================================
    #                Output to a .txt. file like its a mesh file to be able to      |
    #                see the data.                                                  |
    # ===============================================================================
    ngr                    = len( tri_groups )
    fout.write( 'num_groups = ' + str( ngr ) + '\n' )
    for ii in range( ngr ) :
        fout.write( typenames[ii] + '\n' )
        fout.write( group_names[ii] + '\n' )
        triarr             = tri_groups[ii]
        ntris              = len( triarr ) / 3
        fout.write( 'number of tris ' + str( ntris ) + '\n' )
        for jj in range( ntris ) :
            fout.write( str( triarr[3*jj] ) + '  ' + str( triarr[3*jj+1] ) + '  ' + str( triarr[3*jj+2] ) + '\n' )

    nv                     = len( uvcoords ) / 2
    fout.write( 'Number of vertex uv coords ' + str( nv ) + '\n' )
    for ii in range(nv) :
        fout.write( str( uvcoords[2*ii] ) + '  ' + str( uvcoords[2*ii+1] ) + '\n' )

    nv                     = len( vwts ) / 2
    fout.write( 'Number of vertex weights ' + str( nv ) + '\n' )
    for ii in range(nv) :
        fout.write( str( vwts[2*ii] ) + '  ' + str( vwts[2*ii+1] ) + '\n' )

    nv                     = len( vvecs ) / 3
    fout.write( 'Number of vertex vectors ' + str( nv ) + '\n' )
    for ii in range(nv) :
        fout.write( str( vvecs[3*ii] ) + '  ' + str( vvecs[3*ii+1] ) + '  ' + str( vvecs[3*ii+2] ) + '\n' )

    nv                     = len( vbones ) / 4
    fout.write( 'Number of bone assigments ' + str( nv ) + '\n' )
    for ii in range(nv) :
        fout.write( str( vbones[4*ii] ) + '  ' + str( vbones[4*ii+1] ) + '  ' + str( vbones[4*ii+2] ) + '  ' + str( vbones[4*ii+3] ) + '\n' )

    nv                     = len( bnorms ) / 3
    fout.write( 'Number of tangent space normals ' + str( nv ) + '\n' )
#    for ii in range(nv) :
#        fout.write( aby( bnorms[3*ii+2] ) + '  ' + aby( bnorms[3*ii+1] ) + '  ' + aby( bnorms[3*ii+0] ) + '  ' + aby( 0 ) + '    ' + rgb2f( bnorms[3*ii+2], bnorms[3*ii+1], bnorms[3*ii+0] ) + '\n' )

    nv                     = len( btans ) / 3
    fout.write( 'Number of tangent space tangents ' + str( nv ) + '\n' )
#    for ii in range(nv) :
#        fout.write( aby( btans[3*ii+2] ) + '  ' + aby( btans[3*ii+1] ) + '  ' + aby( btans[3*ii+0] ) + '  ' + aby( 0 ) + '    ' + rgb2f( btans[3*ii+2], btans[3*ii+1], btans[3*ii+0] ) + '\n' )

    nv                     = len( bbinorms ) / 3
    fout.write( 'Number of tangent space binormals ' + str( nv ) + '\n' )
#    for ii in range(nv) :
#        fout.write( aby( bbinorms[3*ii+2] ) + '  ' + aby( bbinorms[3*ii+1] ) + '  ' + aby( bbinorms[3*ii+0] ) + '  ' + aby( 0 ) + '    ' + rgb2f( bbinorms[3*ii+2], bbinorms[3*ii+1], bbinorms[3*ii+0] ) + '\n' )

    nb                     = len( bonenames )
    for ii in range( nb ) :
        fout.write( bonenames[ii] + '\n' )

    fout.close()


    # ===============================================================================
    # Finally!   We write the data out to a mesh file for real.                     |
    # ===============================================================================


    ms3dbytecount          = 0
    fout                   = open( fnoutmesh, 'wb' )


    # -------------------------------------------------------------------------------
    # Header granny strings.                                                        |
    # -------------------------------------------------------------------------------
    putuint( 22, fout )
    fout.write( 'serialization::archive' )
    header = [ 3,4,4,4,  8,1,0,0,  0,1,0,0,  0,0,0,0,  0,1,0,0,  0,0,0,4,  0,1,3,1,  0,0,0,0,  0,7,0,1,  0,2,0,0,  0,0,0 ]
    nheader                = len( header )
    for ii in range( nheader ) :
        putubyte( header[ii], fout )

    putuint( num_groups, fout )
    putushort( 0, fout ) # Dummy two bytes.


    # -------------------------------------------------------------------------------
    # Triangle data.  This is first irregular group.                                |
    # -------------------------------------------------------------------------------
    serialodd              = 3
    serialeven             = 4
    triarr                 = tri_groups[0]
    ntriangles             = len( tri_groups )
    print 'ntriangles should be num_groups ' + str( ntriangles ) + ' ' + str( num_groups )

    # Header.
    putushort( 11, fout )                                  # Code for triangles.
    putubyte( 1, fout )
    putubyte( 3, fout )
    putuint( serialodd, fout )                             # Serial count int. 

    # Type, group, and number of triangles in first group.
    typename               = typenames[0]
    groupname              = group_names[0]
    nch                    = len( typename )
    putuint( nch, fout )
    fout.write( typename )
    nch                    = len( groupname )
    putuint( nch, fout )
    fout.write( groupname )
    putushort( 0, fout )                                   # Dummy two 0 bytes.
    ntris                  = len( triarr ) / 3
    putuint( ntris, fout )

    # Write triangle data.
    for ii in range( ntris ) :
        putushort( triarr[3*ii],   fout )
        putushort( triarr[3*ii+1], fout )
        putushort( triarr[3*ii+2], fout )

    # Footer.
    putuint( 0, fout )
    putuint( 0, fout )
    putubyte( 0, fout )
    putushort( 10, fout )
    putushort( 1, fout )
    putuint( serialeven, fout )
    putushort( 11, fout )
    putuint( serialodd, fout )

    # Update sequence counts.
    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Now do the regular triangle groups.                                           |
    # -------------------------------------------------------------------------------
    for igroup in range( 1, num_groups ) :
        # Get triangle data.
        triarr                 = tri_groups[igroup]

        # Header.
        putushort( 11, fout )                              # Code for triangles.
        putuint( serialodd, fout )                         # Serial count int. 

        # Type, group, and number of triangles in first group.
        typename               = typenames[igroup]
        groupname              = group_names[igroup]
        nch                    = len( typename )
        putuint( nch, fout )
        fout.write( typename )
        nch                    = len( groupname )
        putuint( nch, fout )
        fout.write( groupname )
        ntris                  = len( triarr ) / 3
        putuint( ntris, fout )

        # Write triangle data.
        for ii in range( ntris ) :
            putushort( triarr[3*ii],   fout )
            putushort( triarr[3*ii+1], fout )
            putushort( triarr[3*ii+2], fout )

        # Footer.
        putuint( 0, fout )
        putubyte( 0, fout )
        putushort( 10, fout )
        putuint( serialeven, fout )
        putushort( 11, fout )
        putuint( serialodd, fout )

        # Update sequence counts.
        serialodd              = serialodd + 2
        serialeven             = serialeven + 2


    # Have to finish with weird, long granny string here.
    # Example: feudal_knights_lod0:  6  0  1  0 71  0  0  0  7  0  2  0  0  0 142 17  0  0  0  0  2  0  0  0  0  0
    putushort( 6, fout )
    putushort( 1, fout )
    putuint( serialodd, fout )
    putushort( 7, fout )
    putushort( 2, fout )
    putushort( 0, fout )
    putuint( nvert, fout )
    putushort( 0, fout )
    putuint( 2, fout )
    putushort( 0, fout )

    # -------------------------------------------------------------------------------
    # UV coordinates.                                                               |
    # -------------------------------------------------------------------------------
    print 'Starting the uv coordinates: tell(0) says ' + str( fout.tell() )
    serialodd              = serialodd + 2
    # float pairs command string is 18 0.
    # This is the header.
    putushort( 18, fout )
    putushort( 1,  fout )                                  # This is one of those 1 0 code cases.
    putuint( serialeven, fout )                            # We switch and put out even ones now.
    putuint( 4, fout )                                     # A 4 0 0 0  for no apparent reason.
    putushort( 0, fout )                                   # The extra 0 0 for the 1 0 code case.
    putuint( nvert, fout )                                 # The number of vertices.

    # Write uv coordinates data.
    for ii in range( nvert ) :
        putfloat( uvcoords[2*ii],   fout )
        putfloat( uvcoords[2*ii+1], fout )

    # This is the footer.
    putuint( 0, fout )
    putushort( 17, fout )
    putushort( 1, fout )                                   # This is the 1 0 repeat.
    putuint( serialodd, fout )
    putushort( 18, fout )
    putuint( serialeven, fout )

    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Vertex weights.                                                               |
    # -------------------------------------------------------------------------------
    print 'Starting the vertex weights: tell(0) says ' + str( fout.tell() )
    # float pairs command string is 18 0.
    putushort( 18, fout )
    putuint( serialeven, fout )                            # We switch and put out even ones now.
    putuint( 1, fout )                                     # A 1 0 0 0  for no apparent reason.
    putuint( nvert, fout )                                 # The number of vertices.

    # Write vertex weights data.
    for ii in range( nvert ) :
        total              = vwts[2*ii] + vwts[2*ii+1]
        if total > 1.0 :
            print 'Index ' + str(ii) + ' wt total = ' + str( total )
            diff           = total - 1.0
            vwts[2*ii]     = vwts[2*ii] - 1.00001 * diff
        putfloat( vwts[2*ii],   fout )
        putfloat( vwts[2*ii+1], fout )

#    for ii in range( 750 ) :
#        putfloat( 0.0,   fout )
#        putfloat( 0.0, fout )

    # This is the footer.
    putuint( 0, fout )
    putushort( 17, fout )
    putuint( serialodd, fout )
    putushort( 18, fout )
    putuint( serialeven, fout )
    # Weird extra granny stuff here.
    putushort( 0, fout )
    putuint(   1, fout )
    putushort( 0, fout )

    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Vertex vectors.                                                               |
    # -------------------------------------------------------------------------------
    print 'Starting the vertex vectors: tell(0) says ' + str( fout.tell() )
    # float triples command string is 23 0.
    putushort( 23, fout )
    putushort( 1,  fout )                                  # This is one of those 1 0 code cases.
    putuint( serialeven, fout )                            # We switch and put out even ones now.
    putuint( 0, fout )                                     
    putushort( 0, fout )                                   # The extra 0 0 for the 1 0 code case.
    putuint( nvert, fout )                                 # The number of vertices.

    # Write vertex vectors data.
    for ii in range( nvert ) :
        putfloat( vvecs[3*ii],   fout )
        putfloat( vvecs[3*ii+1], fout )
        putfloat( vvecs[3*ii+2], fout )

    # This is the footer.
    putuint( 0, fout )
    putushort( 22, fout )
    putushort( 1, fout )                                   # This is the 1 0 repeat.
    putuint( serialodd, fout )
    putushort( 23, fout )
    putuint( serialeven, fout )
    # Weird extra granny stuff here.
    putushort( 0, fout )
    if MOUNTFLAG == True :
        putuint( TANGENTSPACEFLAG, fout )
    else :
        if LOD0FLAG == True :
            putuint(   4, fout )
        else :
            putuint(   2, fout )                           # GOM found this for higher LODs.

    putushort( 0, fout )

    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Now do the bone assignments.                                                  |
    # -------------------------------------------------------------------------------
    print 'Starting the bone assignments: tell(0) says ' + str( fout.tell() )
    # byte quads command string is 28 0.
    putushort( 28, fout )
    putushort( 1,  fout )                                  # This is one of those 1 0 code cases.
    putuint( serialeven, fout )                            # We switch and put out even ones now.
    putuint( 2, fout )                                     # A 2 0 0 0  for no apparent reason.
    putushort( 0, fout )                                   # The extra 0 0 for the 1 0 code case.
    putuint( nvert, fout )                                 # The number of vertices.

    # Write bone assignmets data.
    for ii in range( nvert ) :
        putubyte( vbones[4*ii],   fout )
        putubyte( vbones[4*ii+1], fout )
        putubyte( vbones[4*ii+2], fout )
        putubyte( vbones[4*ii+3], fout )

    # This is the footer.
    putuint( 0, fout )
    putushort( 27, fout )
    putushort( 1, fout )                                   # This is the 1 0 repeat.
    putuint( serialodd, fout )
    putushort( 28, fout )
    putuint( serialeven, fout )
    # No weird extra granny stuff here.

    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Tangent space basis (the mystery blocks).  Note that lod0's have normals,     |
    # tangents, and binormals. lod1, lod2, and lod3 just have vertex normals.       |
    # Must do this in reverse order z, y, x, 0                                      |
    # -------------------------------------------------------------------------------
    print 'Starting the mystery blocks: tell(0) says ' + str( fout.tell() )
    if MOUNTFLAG == True :
        if TANGENTSPACEFLAG == 4 :
            nblocks        = 3
        else :
            nblocks        = 1
    else :
        if LOD0FLAG == True :
            nblocks        = 3
        else :
            nblocks        = 1


    for imystery in range( nblocks ) :
        # byte quads command string is 28 0.
        putushort( 28, fout )
        putuint( serialeven, fout )           
        if imystery == 0 :
            putuint(  3, fout )                    
        elif imystery == 1 :
            putuint( 10, fout )                    
        elif imystery == 2 :
            putuint( 11, fout )                    

        # Output the vertex count.
        putuint( nvert, fout )

        # Real byte data.
        for ii in range( nvert ) :
            if imystery == 0 :
                putubyte( bnorms[3*ii+2], fout )
                putubyte( bnorms[3*ii+1], fout )
                putubyte( bnorms[3*ii],   fout )
                putubyte(              0, fout )
            elif imystery == 1 :
                putubyte( btans[3*ii+2], fout )
                putubyte( btans[3*ii+1], fout )
                putubyte( btans[3*ii],   fout )
                putubyte(              0, fout )
            elif imystery == 2 :
                putubyte( bbinorms[3*ii+2], fout )
                putubyte( bbinorms[3*ii+1], fout )
                putubyte( bbinorms[3*ii],   fout )
                putubyte(              0, fout )

        # This is the footer.
        putuint( 0, fout )
        putushort( 27, fout )
        putuint( serialodd, fout )
        putushort( 28, fout )
        putuint( serialeven, fout )

        # Increment serials.
        serialodd          = serialodd + 2
        serialeven         = serialeven + 2

    # Write 10 zero bytes then the bounding sphere data.
    putuint( 0, fout )
    putuint( 0, fout )
    putushort( 0, fout )
    # First bounding sphere data written here (also at end).
    putfloat( boundingsphere[0], fout )
    putfloat( boundingsphere[1], fout )
    putfloat( boundingsphere[2], fout )
    putfloat( boundingsphere[3], fout )
    putuint( 0, fout )                                     # Final four 0's.

    # -------------------------------------------------------------------------------
    # Bone strings section.                                                         |
    # -------------------------------------------------------------------------------
    nitems                 = len( bonedata )
    nbones                 = nitems / 3

    putuint( nbones, fout )
    putushort( 0, fout )                                   # Two skip bytes.
    for ii in range( nbones ) :
        nch                = bonedata[3*ii]
        bonename           = bonedata[3*ii+1]
        cnt                = bonedata[3*ii+2]
        putuint( nch, fout )
        fout.write( bonename )
        putuint( cnt, fout )


    # Bizarre granny string hard coded here.  Code this like I grouped it but put in right sequence numbers.
    #   3 0    1 0    86 0 0 0    4 0   1 0    0 0 0 0    1 0 0 0    0 0    39 0    1     4    87 0 0 0
    putushort( 3, fout )
    putushort( 1, fout )
    putuint( serialeven, fout )
    putushort( 4, fout )
    putushort( 1, fout )
    putuint(   0, fout )
    putuint(   1, fout )
    putushort( 0, fout )
    putushort(39, fout )
    putubyte(  1, fout )
    putubyte(  4, fout )
    putuint( serialodd, fout )

    putuint(  13, fout )
    n0                     = fnin.find( 'lod0' )
    n1                     = fnin.find( 'lod1' )
    n2                     = fnin.find( 'lod2' )
    n3                     = fnin.find( 'lod3' )
    if n0 > 0 : 
        fout.write( 'characterlod0' )
    elif n1 > 0 :
        fout.write( 'characterlod1' )
    elif n2 > 0 :
        fout.write( 'characterlod2' )
    elif n3 > 0 :
        fout.write( 'characterlod2' )

    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Really odd closing section.  All hard coded except for sequence numbers.      |
    # -------------------------------------------------------------------------------
    #  17   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     4   0   0   0     1   0   0   0     3   0   0   0  
    #  18 192  63   0     4   0   0   0   255 255 255 255   255 255 255 255   255 255 255 255   255 255 255 255     9   0   0   0  
    #   0   0   0  64     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0     8   0   0   0     2   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0  
    # 128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0  
    #   0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    # 128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0  
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0  
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0  
    #   0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0  38   0     1   0  88   0  
    #   0   0  39   0    87   0   0   0  
    #0.00446363957599 0.432539612055 0.793390989304 1.68704903126

    # First line.
    #  17   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    putuint(   17, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Second line.
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Third line.
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     4   0   0   0     1   0   0   0     3   0   0   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    4, fout )
    putuint(    1, fout )
    putuint(    3, fout )

    # Fourth line.
    #  18 192  63   0     4   0   0   0   255 255 255 255   255 255 255 255   255 255 255 255   255 255 255 255     9   0   0   0  
    putubyte(  18, fout )
    putubyte( 192, fout )
    putubyte(  63, fout )
    putubyte(   0, fout )
    putuint(    4, fout )
    putint(    -1, fout )
    putint(    -1, fout )
    putint(    -1, fout )
    putint(    -1, fout )
    putuint(    9, fout )

    # Fifth line.
    #   0   0   0  64     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    putushort(  0, fout )
    putubyte(   0, fout )
    putubyte(  64, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Sixth line.
    #   0   0   0   0     0   0   0   0     8   0   0   0     2   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    8, fout )
    putuint(    2, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )

    # Seventh line.
    #   0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )

    # Eighth line.
    # 128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0  
    putuint(16256, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )

    # Ninth line.
    #   0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Tenth line.
    # 128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0  
    putuint(16256, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Eleventh line.
    #   0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putuint(16256, fout )

    # Twelth line.
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )

    # Thirteenth line.
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0  
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(16256, fout )

    # Fourteenth line.
    #   0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Fifteenth line.
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0  
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )

    # Sixteenth line.
    #   0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0  38   0     1   0  88   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(16256, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort( 38, fout )
    putushort(  1, fout )
    putushort( serialeven, fout )

    # Seventeenth line.
    #   0   0  39   0    87   0   0   0  
    putushort(  0, fout )
    putushort( 39, fout )
    putuint( serialodd, fout )

    # Bounding sphere.
    #0.00446363957599 0.432539612055 0.793390989304 1.68704903126
    putfloat( boundingsphere[0], fout )
    putfloat( boundingsphere[1], fout )
    putfloat( boundingsphere[2], fout )
    putfloat( boundingsphere[3], fout )

    # Son of a gun, we are done!
    fout.close()

    return


# -----------------------------------------------------------------------------------
#    This def reads data out of the binary ms3d format and converts it to the  
# .mesh format.  This function does the variant mesh that is used by siege weapons.
# Also outputs to a text file to compare data from each way of the
# conversion.
# -----------------------------------------------------------------------------------
def ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh ) :

    # Globals.
    global ms3dbytecount
    global bytecount
    global LOD0FLAG
    global variantboundingsphere

    fout                   = open( fnouttxt, 'w' )

    # Open filename for reading.
    fidms3d                = open( fnin, 'rb' )
    bytecount              = 0

    # -------------------------------------------------------------------------------
    # Get header tag and version number.                                            |
    # -------------------------------------------------------------------------------
    header                 = fidms3d.read( 10 )
    bytecount              = bytecount + 10
    versionnum             = getuint( fidms3d )               
    string                 = header + ', version number ' + str( versionnum )
    fout.write( string + '\n' )
    fout.flush()
    print string

    # -------------------------------------------------------------------------------
    # Vertex data.                                                                  |
    # -------------------------------------------------------------------------------
    nvert                  = getushort( fidms3d )
    string                 = 'Number of vertices = ' + str( nvert )
    fout.write( string + '\n' )
    fout.flush()
    print string

    # Allocate arrays.
    vvecs                  = array.array( 'f' )
    vbones                 = array.array( 'B' )
    vnorms                 = array.array( 'f' )
    vtans                  = array.array( 'f' )
    vbinorms               = array.array( 'f' )
    # Initialize.
    for ii in range( 3 * nvert ) :
        vvecs.append( 0.0 )
        vnorms.append( 0.0 )
        vtans.append( 0.0 )
        vbinorms.append( 0.0 )
    # Initialize.
    for ii in range( 4 * nvert ) :
        vbones.append( 0 )

    # Read in vertex data. This includes flags, boneId, and reference count.
    for ii in range( nvert ) :
        flag               = getubyte( fidms3d )     # Flag value we don't care about.
        vvecs[3*ii]        = -getfloat( fidms3d )    # Vertex vector, x, y, and z
        vvecs[3*ii+1]      = getfloat( fidms3d )
        vvecs[3*ii+2]      = getfloat( fidms3d )
        vbones[4*ii+2]     = getubyte( fidms3d )     # Bone Id, put in the primary slot.
#        putbyte( vrefs[ii], fidms3d )                # Reference count.
        dummy              = getubyte( fidms3d )     # Don't care about reference count.
    

    # -------------------------------------------------------------------------------
    # Triangle data.  This is a mess because the groups are all messed up.          |
    # We also compute our vertex normals at this point using the triangles.         |
    # -------------------------------------------------------------------------------
    ntriangles             = getushort( fidms3d )
    print 'ntriangles = ' + str( ntriangles )
    tris                   = array.array( 'H' )
    s                      = array.array( 'f' )
    t                      = array.array( 'f' )
    groupidx               = array.array( 'B' )
    for ii in range( ntriangles ) :
        tris.append( 0 )
        tris.append( 0 )
        tris.append( 0 )
        s.append( 0.0 )
        s.append( 0.0 )
        s.append( 0.0 )
        t.append( 0.0 )
        t.append( 0.0 )
        t.append( 0.0 )
        groupidx.append( 0 )

    for ii in range( ntriangles ) :
        dummy              = getushort( fidms3d )
        tris[3*ii]         = getushort( fidms3d )
        tris[3*ii+2]       = getushort( fidms3d )     # Note reversed triangle vertices! Important!
        tris[3*ii+1]       = getushort( fidms3d )
        idx1               = tris[3*ii]
        idx2               = tris[3*ii+1]
        idx3               = tris[3*ii+2]
#        vnorms[3*idx1]     = -getfloat(  fidms3d )
#        vnorms[3*idx1+1]   = getfloat(  fidms3d )
#        vnorms[3*idx1+2]   = getfloat(  fidms3d )
#        vnorms[3*idx2]     = -getfloat(  fidms3d )
#        vnorms[3*idx2+1]   = getfloat(  fidms3d )
#        vnorms[3*idx2+2]   = getfloat(  fidms3d )
#        vnorms[3*idx3]     = -getfloat(  fidms3d )
#        vnorms[3*idx3+1]   = getfloat(  fidms3d )
#        vnorms[3*idx3+2]   = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        dummynorm          = getfloat(  fidms3d )
        s[3*ii]            = getfloat(  fidms3d )          # NOTE THE SWITCH IN UV COORDS TO MATCH THE TRIANGLE SWITCH! 
        s[3*ii+2]          = getfloat(  fidms3d )
        s[3*ii+1]          = getfloat(  fidms3d )
        t[3*ii]            = getfloat(  fidms3d )
        t[3*ii+2]          = getfloat(  fidms3d )
        t[3*ii+1]          = getfloat(  fidms3d )
        bdummy             = getubyte(  fidms3d )
        groupidx[ii]       = getubyte(  fidms3d )
        # Don't believe vertex normals, going to roll my own.
        v1x                = vvecs[3*idx1]
        v1y                = vvecs[3*idx1+1]
        v1z                = vvecs[3*idx1+2]
        v2x                = vvecs[3*idx2]
        v2y                = vvecs[3*idx2+1]
        v2z                = vvecs[3*idx2+2]
        v3x                = vvecs[3*idx3]
        v3y                = vvecs[3*idx3+1]
        v3z                = vvecs[3*idx3+2]
        edge12x            = v2x - v1x            #
        edge12y            = v2y - v1y            #   | i   j   k  |
        edge12z            = v2z - v1z            #   | ux  uy  uz | = i(uy*vz-uz*vy) + j(uz*vx-ux*vz) + k(ux*vy-uy*vx)
        edge23x            = v3x - v2x            #   | vx  vy  vz |
        edge23y            = v3y - v2y            #
        edge23z            = v3z - v2z            #
        # Take cross product to get a normal vector (normal to face, but it won't be normalized).
        unx                = edge12y * edge23z - edge12z * edge23y
        uny                = edge12z * edge23x - edge12x * edge23z
        unz                = edge12x * edge23y - edge12y * edge23x
        # Normalize it.
        norm               = unx*unx + uny*uny + unz*unz 
        norm               = norm**0.5 
        if norm != 0.0 :
            nx             = unx / norm     
            ny             = uny / norm
            nz             = unz / norm
        else :
            nx             = unx
            ny             = uny
            nz             = unz
        # Add it to vnorms for each index.
        vnorms[3*idx1]     = vnorms[3*idx1]   + nx
        vnorms[3*idx2]     = vnorms[3*idx2]   + nx
        vnorms[3*idx3]     = vnorms[3*idx3]   + nx
        vnorms[3*idx1+1]   = vnorms[3*idx1+1] + ny
        vnorms[3*idx2+1]   = vnorms[3*idx2+1] + ny
        vnorms[3*idx3+1]   = vnorms[3*idx3+1] + ny
        vnorms[3*idx1+2]   = vnorms[3*idx1+2] + nz
        vnorms[3*idx2+2]   = vnorms[3*idx2+2] + nz
        vnorms[3*idx3+2]   = vnorms[3*idx3+2] + nz


    # Now we have to normalize our vnorms.
    for ii in range( nvert ) :
    #        fidout.write( str( vnorms[3*ii] ) + ' ' + str( vnorms[3*ii+1] ) + ' ' + str( vnorms[3*ii+2] ) + '\n' )
        norm               = vnorms[3*ii]*vnorms[3*ii] + vnorms[3*ii+1]*vnorms[3*ii+1] + vnorms[3*ii+2]*vnorms[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0 :
            vnorms[3*ii]   = vnorms[3*ii]   / norm
            vnorms[3*ii+1] = vnorms[3*ii+1] / norm
            vnorms[3*ii+2] = vnorms[3*ii+2] / norm


    # -------------------------------------------------------------------------------
    # Group data.  This allows us to reconstruct the .mesh format groups.           |
    # -------------------------------------------------------------------------------
    num_groups             = getushort( fidms3d )
    print 'num_groups = ' + str( num_groups )
    group_names            = []
    tri_groups             = []
    for ii in range( num_groups ) :
        bdummy             = getubyte(  fidms3d )
        tempname           = fidms3d.read( 32 )
        bytecount          = bytecount + 32
        groupname          = zipstrip( tempname )
        print groupname
        group_names.append( groupname )

        ntris              = getushort( fidms3d )
        triidx             = array.array( 'H' )
        for jj in range( ntris ) :
            triidx.append( getushort( fidms3d ) )

        bchar              = getbyte( fidms3d )              # Don't care about material index.
        # Now reorganize the triangle data into the mesh format.
        triarr             = array.array( 'H' )
        for jj in range( ntris ) :
            idx            = triidx[jj]
            triarr.append( tris[3*idx] )
            triarr.append( tris[3*idx+1] )
            triarr.append( tris[3*idx+2] )

        tri_groups.append( triarr )

    # Print out a little face data to check.
    tritest                = tri_groups[0]
    print 'face: ' + str( tritest[0] ) + ' ' + str( tritest[1] ) + ' ' + str( tritest[2] )
    print 'face: ' + str( tritest[3] ) + ' ' + str( tritest[4] ) + ' ' + str( tritest[5] )
    print 'face: ' + str( tritest[6] ) + ' ' + str( tritest[7] ) + ' ' + str( tritest[8] )
    print 'face: ' + str( tritest[9] ) + ' ' + str( tritest[10] ) + ' ' + str( tritest[11] )


    # -------------------------------------------------------------------------------
    # Read nMaterials (the materials struct is 361 bytes).  Just skip if more than 0.
    # -------------------------------------------------------------------------------
    nMaterials             = getushort( fidms3d )
    print 'nMaterials = ' + str( nMaterials )

    if nMaterials > 0 :
        for jj in range( nMaterials ) :
            fidms3d.seek( 361, 1 )                         # This just relative skips past it.

    # -------------------------------------------------------------------------------
    # Keyframer data. (Just junk to us.)                                            |
    # -------------------------------------------------------------------------------
    fAnimationFPS          = getfloat(  fidms3d )
    fCurrentTime           = getfloat(  fidms3d )
    iTotalFrames           = getuint(   fidms3d )
    print 'fAnimationFPS = ' + str( fAnimationFPS )
    print 'fCurrentTime  = ' + str( fCurrentTime  )
    print 'iTotalFrames  = ' + str( iTotalFrames  )

    print 'Passed the iTotalFrames read, tell says' + str( fidms3d.tell() )


    # -------------------------------------------------------------------------------
    # Joints or bone data.                                                          |
    # -------------------------------------------------------------------------------
    njoints                = getushort( fidms3d )
    print 'njoints = ' + str( njoints )
    bonenames              = []
    parentnames            = []
    for ii in range( njoints ) :
        dummy              = getubyte(  fidms3d )
        tbone              = fidms3d.read( 32 )
        bonename           = zipstrip( tbone )
        pbone              = fidms3d.read( 32 )
        bytecount          = bytecount + 64
        parentname         = zipstrip( pbone )
        bonenames.append( bonename )
        parentnames.append( parentname )
        print bonename + ' from ' + parentname
        # rotation[3] float.
        fdummy             = getfloat(  fidms3d )
        fdummy             = getfloat(  fidms3d )
        fdummy             = getfloat(  fidms3d )
        # position[3] float.
        fdummy             = getfloat(  fidms3d )
        fdummy             = getfloat(  fidms3d )
        fdummy             = getfloat(  fidms3d )
        # numKeyFramesRot.
        numKeyFramesRot    = getshort(  fidms3d )
        # numKeyFramesTrans
        numKeyFramesTrans  = getshort(  fidms3d )
        for ii in range( numKeyFramesRot ) :
            fdummy         = getfloat(  fidms3d )          # Just dump the animation matrices.
            fdummy         = getfloat(  fidms3d )
            fdummy         = getfloat(  fidms3d )
            fdummy         = getfloat(  fidms3d )

        for ii in range( numKeyFramesTrans ) :
            fdummy         = getfloat(  fidms3d )          # Ditto.
            fdummy         = getfloat(  fidms3d )
            fdummy         = getfloat(  fidms3d )
            fdummy         = getfloat(  fidms3d )

    # -------------------------------------------------------------------------------
    # Subversion number (always 1).                                                 |
    # -------------------------------------------------------------------------------
    dummy                  = getuint(   fidms3d )
    print 'First subversion number = ' + str( dummy )
    print 'Passed the first subversion number, tell says' + str( fidms3d.tell() )


    # -------------------------------------------------------------------------------
    # Group comments.  This is where the group types live along with group names.   |
    # -------------------------------------------------------------------------------
    num_comments           = getuint(   fidms3d )
    typenames              = []
    for ii in range( num_comments ) :
        idx                = getuint(   fidms3d )
        nchars             = getuint(   fidms3d )
        comment            = fidms3d.read( nchars )
        bytecount          = bytecount + nchars
        nch                = comment.find( '\r\n' )
        typename           = comment[0:nch]
        typenames.append( typename )
        print typename + ' from ' + comment

    print 'Passed the group comments, tell says' + str( fidms3d.tell() )


    # -------------------------------------------------------------------------------
    # Material comments.  Not used by me, should be zero.                           |
    # -------------------------------------------------------------------------------
    nNumMaterialComments   = getuint(   fidms3d )
    print 'nNumMaterialComments = ' + str( nNumMaterialComments )

    if nNumMaterialComments > 0 :
        for ii in range( nNumMaterialComments ) :
            idummy         = getuint(   fidms3d ) 
            nchars         = getuint(   fidms3d )
            cdummy         = fidms3d.read( nchars )

    print 'Passed the material comments, tell says' + str( fidms3d.tell() )


    # -------------------------------------------------------------------------------
    # Joint comments.  Not used by me, should be zero.                              |
    # -------------------------------------------------------------------------------
    nNumJointComments      = getuint(   fidms3d )
    print 'nNumJointComments = ' + str( nNumJointComments )

    if ( nNumJointComments > 0 ) & ( nNumJointComments != 3 ) :
        for ii in range( nNumJointComments ) :
            idummy         = getuint(   fidms3d ) 
            nchars         = getuint(   fidms3d )
            cdummy         = fidms3d.read( nchars )

    if nNumJointComments == 3 :
        idx                = getuint(   fidms3d )           
        nchars1            = getuint(   fidms3d )
        texfile1           = fidms3d.read( nchars1 )
        idx                = getuint(   fidms3d )           
        nchars2            = getuint(   fidms3d )
        texfile2           = fidms3d.read( nchars2 )
        idx                = getuint(   fidms3d )           
        nchars3            = getuint(   fidms3d )
        texfile3           = fidms3d.read( nchars3 )

    print 'Passed the joint comments, tell says' + str( fidms3d.tell() )

    # -------------------------------------------------------------------------------
    # Model comments. This is where we store the bounding sphere data and,          |
    # optionally, the bone strings data that comes directly out of the .mesh file.  |
    # -------------------------------------------------------------------------------
    nHasModelComment       = getuint(   fidms3d )
    print 'nHasModelComment = ' + str( nHasModelComment )
    bonedata               = []
    boundingsphere         = array.array( 'f' )
    for ii in range( 4 ) :
        boundingsphere.append( 0.0 )

    # boundingsphere data comes as ASCII in the first comment.
#    nGOM                   = fnin.find( 'GOM' )
#    if nGOM == -1 :
#        idummy             = getuint(   fidms3d )          # GOM skips the index int here.
    nchars                 = getuint(   fidms3d )
    boundingspherestring   = fidms3d.read( nchars )
    print 'boundingspherestring = ' + boundingspherestring
    flist                  = boundingspherestring.split( '\r\n' )
    boundingsphere[0]      = float( flist[0] )
    boundingsphere[1]      = float( flist[1] )
    boundingsphere[2]      = float( flist[2] )
    boundingsphere[3]      = float( flist[3] )

    print 'nchars            = ' + str( nchars )
    print 'boundingsphere[0] = ' + str( boundingsphere[0] )
    print 'boundingsphere[1] = ' + str( boundingsphere[1] )
    print 'boundingsphere[2] = ' + str( boundingsphere[2] )
    print 'boundingsphere[3] = ' + str( boundingsphere[3] )

    # My format has bone strings here.
    if nHasModelComment > 1 :
        for ii in range( nHasModelComment - 1 ) :
            idummy         = getuint(   fidms3d ) 
            nchars         = getuint(   fidms3d )
            nch            = getuint(   fidms3d )
            bonedata.append( nch )
            bonename       = fidms3d.read( nch )
            bonedata.append( bonename )
            cnt            = getuint(   fidms3d )
            bonedata.append( cnt )       
            print 'nch      = ' + str( nch )
            print 'bonename = ' + bonename
            print 'cnt      = ' + str( cnt )
    else :
        # Hard code the standard skeleton if need be.
        nbones             = len( bonenames )
        for ii in range( nbones ) :
            bonename       = bonenames[ii]
            nch            = len( bonename )
            if bonename == 'bone_weapon' :
                cnt        = 20
            elif bonename == 'bone_weapon02' :
                cnt        = 21
            elif bonename == 'bone_weapon03' :
                cnt        = 22
            elif bonename == 'bone_shield01' :
                cnt        = 22
            elif bonename == 'bone_shield' :
                cnt        = 22
            else :
                cnt        = ii
            bonedata.append( nch )
            bonedata.append( bonename )
            bonedata.append( cnt )       


    # -------------------------------------------------------------------------------
    # Vertex weighting data.                                                        |
    # -------------------------------------------------------------------------------
    nv                     = getuint(   fidms3d )
    print 'Second subversion number = ' + str( nv )
    vwts                   = array.array( 'f' )
    for ii in range( nvert ) :
        vbones[4*ii+1]     = getubyte(  fidms3d ) 
        dummy              = getubyte(  fidms3d ) 
        dummy              = getubyte(  fidms3d ) 
        if ( vbones[4*ii+1] > njoints-1 ) :
            vbones[4*ii+1] = vbones[4*ii+2]
            wtprim         = 1.0
            wtsecond       = 0.0 
            dummy          = getubyte( fidms3d )
            dummy          = getubyte( fidms3d )
        else :
            wtprim         = float( getubyte(  fidms3d ) ) / 100.0   # Convert percentage to float.
            wtsecond       = float( getubyte(  fidms3d ) ) / 100.0   
            if ( wtsecond > 1.00001 ) | ( wtprim > 1.00001 ) :
                wtprim     = 1.0
                wtsecond   = 0.0
            else:
                wtprim     = wtprim + ( 1 - wtprim - wtsecond )
        vwts.append( wtprim )
        vwts.append( wtsecond )
        dummy              = getubyte(  fidms3d )
        if nv == 2 :
            dummy          = getuint(  fidms3d )                     # SubVersion 2 has extra int field.

    # That's all the data, close now.
    fidms3d.close()

    # -------------------------------------------------------------------------------
    # Now to tidy up and package it all.  vvecs, vwts, and vbones are done          |
    # but we need uvcoords. The triangle data is already sort of organized.         |
    # -------------------------------------------------------------------------------
    uvcoords               = array.array( 'f' )
    for ii in range( nvert ) :
        uvcoords.append( 0.0 ) 
        uvcoords.append( 0.0 ) 

    for ii in range( ntriangles ) :
        idx1               = tris[3*ii]      # Indices to vertices.
        idx2               = tris[3*ii+1]
        idx3               = tris[3*ii+2]
        s1                 = s[3*ii]
        s2                 = s[3*ii+1]
        s3                 = s[3*ii+2]
        t1                 = t[3*ii]
        t2                 = t[3*ii+1]
        t3                 = t[3*ii+2]
        uvcoords[2*idx1]   = s1
        uvcoords[2*idx1+1] = t1
        uvcoords[2*idx2]   = s2
        uvcoords[2*idx2+1] = t2
        uvcoords[2*idx3]   = s3
        uvcoords[2*idx3+1] = t3

    # -------------------------------------------------------------------------------
    # First we have to calculate the tangent basis.  Then we have to do the cube    |
    # normalization calculation before we are ready to output to the .mesh format.  |
    # -------------------------------------------------------------------------------
    for ii in range( ntriangles ) :
        idx1               = tris[3*ii]      # Indices to vertices.
        idx2               = tris[3*ii+1]
        idx3               = tris[3*ii+2]
        # Edges.
        e1x                = vvecs[3*idx2]      - vvecs[3*idx1]
        e1y                = vvecs[3*idx2+1]    - vvecs[3*idx1+1]
        e1z                = vvecs[3*idx2+2]    - vvecs[3*idx1+2]
        e2x                = vvecs[3*idx3]      - vvecs[3*idx1]
        e2y                = vvecs[3*idx3+1]    - vvecs[3*idx1+1]
        e2z                = vvecs[3*idx3+2]    - vvecs[3*idx1+2]
        euv1x              = uvcoords[2*idx2]   - uvcoords[2*idx1]
        euv1y              = uvcoords[2*idx2+1] - uvcoords[2*idx1+1]
        euv2x              = uvcoords[2*idx3]   - uvcoords[2*idx1]
        euv2y              = uvcoords[2*idx3+1] - uvcoords[2*idx1+1]
        cp                 = euv1y * euv2x - euv1x * euv2y
        if cp != 0.0 :
            mul            = 1.0 / cp
            tanx           = ( -e1x * euv2y + e2x * euv1y ) * mul
            tany           = ( -e1y * euv2y + e2y * euv1y ) * mul
            tanz           = ( -e1z * euv2y + e2z * euv1y ) * mul
            norm           = tanx*tanx + tany*tany + tanz*tanz
            norm           = norm**0.5
            tanx           = tanx / norm
            tany           = tany / norm
            tanz           = tanz / norm
            vtans[3*idx1]  = vtans[3*idx1]   + tanx
            vtans[3*idx1+1]= vtans[3*idx1+1] + tany
            vtans[3*idx1+2]= vtans[3*idx1+2] + tanz
            vtans[3*idx2]  = vtans[3*idx2]   + tanx
            vtans[3*idx2+1]= vtans[3*idx2+1] + tany
            vtans[3*idx2+2]= vtans[3*idx2+2] + tanz
            vtans[3*idx3]  = vtans[3*idx3]   + tanx
            vtans[3*idx3+1]= vtans[3*idx3+1] + tany
            vtans[3*idx3+2]= vtans[3*idx3+2] + tanz

    for ii in range( nvert ) :
        norm               = vtans[3*ii]*vtans[3*ii] + vtans[3*ii+1]*vtans[3*ii+1] + vtans[3*ii+2]*vtans[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0:
            vtans[3*ii]    = vtans[3*ii]   / norm
            vtans[3*ii+1]  = vtans[3*ii+1] / norm
            vtans[3*ii+2]  = vtans[3*ii+2] / norm
        # Now orthogonalize with respect to vertex normals.
        dotprod            = vnorms[3*ii]*vtans[3*ii] + vnorms[3*ii+1]*vtans[3*ii+1]  + vnorms[3*ii+2]*vtans[3*ii+2]
        vtans[3*ii]        = vtans[3*ii]   - vnorms[3*ii]   * dotprod
        vtans[3*ii+1]      = vtans[3*ii+1] - vnorms[3*ii+1] * dotprod
        vtans[3*ii+2]      = vtans[3*ii+2] - vnorms[3*ii+2] * dotprod
        # Renormalize.
        norm               = vtans[3*ii]*vtans[3*ii] + vtans[3*ii+1]*vtans[3*ii+1] + vtans[3*ii+2]*vtans[3*ii+2]
        norm               = norm**0.5
        if norm != 0.0:
            vtans[3*ii]    = -vtans[3*ii]   / norm
            vtans[3*ii+1]  = -vtans[3*ii+1] / norm
            vtans[3*ii+2]  = -vtans[3*ii+2] / norm
        # Finally we get the binorms by taking a cross product and we have the tangent space system.
        vbinorms[3*ii]     = -( vnorms[3*ii+1] * vtans[3*ii+2] - vnorms[3*ii+2] * vtans[3*ii+1] )
        vbinorms[3*ii+1]   = -( vnorms[3*ii+2] * vtans[3*ii]   - vnorms[3*ii]   * vtans[3*ii+2] )
        vbinorms[3*ii+2]   = -( vnorms[3*ii]   * vtans[3*ii+1] - vnorms[3*ii+1] * vtans[3*ii] )

    # Now create the normalization cube.
    bnorms                 = array.array( 'B' )
    btans                  = array.array( 'B' )
    bbinorms               = array.array( 'B' )
    for ii in range( 3*nvert ) :
        bnorms.append( 0 )
        btans.append( 0 )
        bbinorms.append( 0 )

    for ii in range( nvert ) :
        bnorms[3*ii]       = int( round( 255.0 * 0.5 * ( vnorms[3*ii]     + 1.0 ) ) )
        bnorms[3*ii+1]     = int( round( 255.0 * 0.5 * ( vnorms[3*ii+1]   + 1.0 ) ) )
        bnorms[3*ii+2]     = int( round( 255.0 * 0.5 * ( vnorms[3*ii+2]   + 1.0 ) ) )
        btans[3*ii]        = int( round( 255.0 * 0.5 * ( vtans[3*ii]      + 1.0 ) ) )
        btans[3*ii+1]      = int( round( 255.0 * 0.5 * ( vtans[3*ii+1]    + 1.0 ) ) )
        btans[3*ii+2]      = int( round( 255.0 * 0.5 * ( vtans[3*ii+2]    + 1.0 ) ) )
        bbinorms[3*ii]     = int( round( 255.0 * 0.5 * ( vbinorms[3*ii]   + 1.0 ) ) )
        bbinorms[3*ii+1]   = int( round( 255.0 * 0.5 * ( vbinorms[3*ii+1] + 1.0 ) ) )
        bbinorms[3*ii+2]   = int( round( 255.0 * 0.5 * ( vbinorms[3*ii+2] + 1.0 ) ) )


    # ===============================================================================
    #                Output to a .txt. file like its a mesh file to be able to      |
    #                see the data.                                                  |
    # ===============================================================================
    ngr                    = len( tri_groups )
    fout.write( 'num_groups = ' + str( ngr ) + '\n' )
    for ii in range( ngr ) :
        fout.write( typenames[ii] + '\n' )
        fout.write( group_names[ii] + '\n' )
        triarr             = tri_groups[ii]
        ntris              = len( triarr ) / 3
        fout.write( 'number of tris ' + str( ntris ) + '\n' )
        for jj in range( ntris ) :
            fout.write( str( triarr[3*jj] ) + '  ' + str( triarr[3*jj+1] ) + '  ' + str( triarr[3*jj+2] ) + '\n' )

    nv                     = len( uvcoords ) / 2
    fout.write( 'Number of vertex uv coords ' + str( nv ) + '\n' )
    for ii in range(nv) :
        fout.write( str( uvcoords[2*ii] ) + '  ' + str( uvcoords[2*ii+1] ) + '\n' )

    nv                     = len( vwts ) / 2
    fout.write( 'Number of vertex weights ' + str( nv ) + '\n' )
    for ii in range(nv) :
        fout.write( str( vwts[2*ii] ) + '  ' + str( vwts[2*ii+1] ) + '\n' )

    nv                     = len( vvecs ) / 3
    fout.write( 'Number of vertex vectors ' + str( nv ) + '\n' )
    for ii in range(nv) :
        fout.write( str( vvecs[3*ii] ) + '  ' + str( vvecs[3*ii+1] ) + '  ' + str( vvecs[3*ii+2] ) + '\n' )

    nv                     = len( vbones ) / 4
    fout.write( 'Number of bone assigments ' + str( nv ) + '\n' )
    for ii in range(nv) :
        fout.write( str( vbones[4*ii] ) + '  ' + str( vbones[4*ii+1] ) + '  ' + str( vbones[4*ii+2] ) + '  ' + str( vbones[4*ii+3] ) + '\n' )

    nv                     = len( bnorms ) / 3
    fout.write( 'Number of tangent space normals ' + str( nv ) + '\n' )
#    for ii in range(nv) :
#        fout.write( aby( bnorms[3*ii+2] ) + '  ' + aby( bnorms[3*ii+1] ) + '  ' + aby( bnorms[3*ii+0] ) + '  ' + aby( 0 ) + '    ' + rgb2f( bnorms[3*ii+2], bnorms[3*ii+1], bnorms[3*ii+0] ) + '\n' )

    nv                     = len( btans ) / 3
    fout.write( 'Number of tangent space tangents ' + str( nv ) + '\n' )
#    for ii in range(nv) :
#        fout.write( aby( btans[3*ii+2] ) + '  ' + aby( btans[3*ii+1] ) + '  ' + aby( btans[3*ii+0] ) + '  ' + aby( 0 ) + '    ' + rgb2f( btans[3*ii+2], btans[3*ii+1], btans[3*ii+0] ) + '\n' )

    nv                     = len( bbinorms ) / 3
    fout.write( 'Number of tangent space binormals ' + str( nv ) + '\n' )
#    for ii in range(nv) :
#        fout.write( aby( bbinorms[3*ii+2] ) + '  ' + aby( bbinorms[3*ii+1] ) + '  ' + aby( bbinorms[3*ii+0] ) + '  ' + aby( 0 ) + '    ' + rgb2f( bbinorms[3*ii+2], bbinorms[3*ii+1], bbinorms[3*ii+0] ) + '\n' )

    nb                     = len( bonenames )
    for ii in range( nb ) :
        fout.write( bonenames[ii] + '\n' )

    fout.close()


    # ===============================================================================
    # Finally!   We write the data out to a mesh file for real.                     |
    # ===============================================================================


    ms3dbytecount          = 0
    fout                   = open( fnoutmesh, 'wb' )


    # -------------------------------------------------------------------------------
    # Header granny strings.                                                        |
    # -------------------------------------------------------------------------------
    putuint( 22, fout )
    fout.write( 'serialization::archive' )
    # Variant format,  this string is missing 4 bytes, 0 1 0 0, that are in the regular mesh.
    header = [ 3,4,4,4,  8,1,0,0,  0,0,0,0,  0,1,0,0,  0,0,0,4,  0,1,3,0,  0,0,0,0,  0,7,0,1,  0,1,0,0,  0,0,0 ]
    nheader                = len( header )
    for ii in range( nheader ) :
        putubyte( header[ii], fout )

    putuint( num_groups, fout )
    putushort( 0, fout ) # Dummy two bytes.


    # -------------------------------------------------------------------------------
    # Triangle data.  This is first irregular group.                                |
    # -------------------------------------------------------------------------------
    serialodd              = 3
    serialeven             = 2
    triarr                 = tri_groups[0]
    ntriangles             = len( tri_groups )
    print 'ntriangles should be num_groups ' + str( ntriangles ) + ' ' + str( num_groups )

    # Header.
    putushort( 11, fout )                                  # Code for triangles.
    putubyte( 1, fout )
    putubyte( 3, fout )
    putuint( serialeven, fout )                            # Serial count int.  NOTE: even is backwards from regular. 

    # Type, group, and number of triangles in first group.
    typename               = typenames[0]
    groupname              = group_names[0]
    nch                    = len( typename )
    putuint( nch, fout )
    fout.write( typename )
    nch                    = len( groupname )
    putuint( nch, fout )
    fout.write( groupname )
    putushort( 0, fout )                                   # Dummy two 0 bytes.
    ntris                  = len( triarr ) / 3
    putuint( ntris, fout )
    print 'ntris for first block = ' + str( ntris )

    # Write triangle data.
    for ii in range( ntris ) :
        putushort( triarr[3*ii],   fout )
        putushort( triarr[3*ii+1], fout )
        putushort( triarr[3*ii+2], fout )

    # Footer.
    putuint( 0, fout )
    putuint( 0, fout )
    putubyte( 0, fout )
    putushort( 10, fout )
    putushort( 1, fout )
    putuint( serialodd, fout )                             # NOTE: backwards from regular mesh.
    putushort( 11, fout )
    putuint( serialeven, fout )                            # NOTE: backwards from regular mesh.

    # Update sequence counts.
    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Now do the regular triangle groups.                                           |
    # -------------------------------------------------------------------------------
    for igroup in range( 1, num_groups ) :
        # Get triangle data.
        triarr                 = tri_groups[igroup]

        # Header.
        putushort( 11, fout )                              # Code for triangles.
        putuint( serialodd, fout )                         # Serial count int. 

        # Type, group, and number of triangles in first group.
        typename               = typenames[igroup]
        groupname              = group_names[igroup]
        nch                    = len( typename )
        putuint( nch, fout )
        fout.write( typename )
        nch                    = len( groupname )
        putuint( nch, fout )
        fout.write( groupname )
        ntris                  = len( triarr ) / 3
        putuint( ntris, fout )

        # Write triangle data.
        for ii in range( ntris ) :
            putushort( triarr[3*ii],   fout )
            putushort( triarr[3*ii+1], fout )
            putushort( triarr[3*ii+2], fout )

        # Footer.
        putuint( 0, fout )
        putubyte( 0, fout )
        putushort( 10, fout )
        putuint( serialeven, fout )
        putushort( 11, fout )
        putuint( serialodd, fout )

        # Update sequence counts.
        serialodd              = serialodd + 2
        serialeven             = serialeven + 2


    # Have to finish with weird, long granny string here.
    # Example: feudal_knights_lod0:  6  0  1  0 71  0  0  0  7  0  2  0  0  0 142 17  0  0  0  0  2  0  0  0  0  0
    putushort( 6, fout )
    putushort( 1, fout )
    putuint( serialeven, fout )                            # NOTE: backwards from regular mesh.
    putushort( 7, fout )
    putushort( 1, fout )                                   # NOTE: 1 here and not a 2.
    putushort( 0, fout )
    putuint( nvert, fout )
    putushort( 0, fout )
    putuint( 2, fout )
    putushort( 0, fout )

    # -------------------------------------------------------------------------------
    # UV coordinates.                                                               |
    # -------------------------------------------------------------------------------
    print 'Starting the uv coordinates: tell(0) says ' + str( fout.tell() )
    serialeven             = serialeven + 2
    # float pairs command string is 18 0.
    # This is the header.
    putushort( 18, fout )
    putushort( 1,  fout )                                  # This is one of those 1 0 code cases.
    putuint( serialodd, fout )                             # NOTE: backwards, use odd here.
    putuint( 4, fout )                                     # A 4 0 0 0  for no apparent reason.
    putushort( 0, fout )                                   # The extra 0 0 for the 1 0 code case.
    putuint( nvert, fout )                                 # The number of vertices.

    # Write uv coordinates data.
    for ii in range( nvert ) :
        putfloat( uvcoords[2*ii],   fout )
        putfloat( uvcoords[2*ii+1], fout )

    # This is the footer.
    putuint( 0, fout )
    putushort( 17, fout )
    putushort( 1, fout )                                   # This is the 1 0 repeat.
    putuint( serialeven, fout )                            # NOTE: backwards from regular mesh.
    putushort( 18, fout )
    putuint( serialodd, fout )                             # NOTE: backwards from regular mesh.

    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Vertex weights.                                                               |
    # -------------------------------------------------------------------------------
    print 'Starting the vertex weights: tell(0) says ' + str( fout.tell() )
    # float pairs command string is 18 0.
    putushort( 18, fout )
    putuint( serialodd, fout )                             # NOTE: backwards from regular mesh.
    putuint( 1, fout )                                     # A 1 0 0 0  for no apparent reason.
    putuint( nvert, fout )                                 # The number of vertices.

    # Write vertex weights data.
    for ii in range( nvert ) :
        total              = vwts[2*ii] + vwts[2*ii+1]
        if total > 1.0 :
            print 'Index ' + str(ii) + ' wt total = ' + str( total )
            diff           = total - 1.0
            vwts[2*ii]     = vwts[2*ii] - 1.00001 * diff
        putfloat( vwts[2*ii],   fout )
        putfloat( vwts[2*ii+1], fout )

#    for ii in range( 750 ) :
#        putfloat( 0.0,   fout )
#        putfloat( 0.0, fout )

    # This is the footer.
    putuint( 0, fout )
    putushort( 17, fout )
    putuint( serialeven, fout )                            # NOTE: backwards from regular mesh.
    putushort( 18, fout )
    putuint( serialodd, fout )                             # NOTE: backwards from regular mesh.
    # Weird extra granny stuff here.
    putushort( 0, fout )
    putuint(   4, fout )                                   # NOTE: 4 instead of a 1 here.
    putushort( 0, fout )

    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Vertex vectors.                                                               |
    # -------------------------------------------------------------------------------
    print 'Starting the vertex vectors: tell(0) says ' + str( fout.tell() )
    # float triples command string is 23 0.
    putushort( 23, fout )
    putushort( 1,  fout )                                  # This is one of those 1 0 code cases.
    putuint( serialodd, fout )                             # NOTE: backwards from regular mesh.
    putuint( 0, fout )                                     
    putushort( 0, fout )                                   # The extra 0 0 for the 1 0 code case.
    putuint( nvert, fout )                                 # The number of vertices.

    # Write vertex vectors data.
    for ii in range( nvert ) :
        putfloat( vvecs[3*ii],   fout )
        putfloat( vvecs[3*ii+1], fout )
        putfloat( vvecs[3*ii+2], fout )

    # This is the footer.
    putuint( 0, fout )
    putushort( 22, fout )
    putushort( 1, fout )                                   # This is the 1 0 repeat.
    putuint( serialeven, fout )                            # NOTE: backwards from regular mesh.
    putushort( 23, fout )
    putuint( serialodd, fout )                             # NOTE: backwards from regular mesh.
#    # Weird extra granny stuff here.
#    putushort( 0, fout )
#    if LOD0FLAG == True :
#        putuint(   4, fout )
#    else :
#        putuint(   2, fout )                               # GOM found this for higher LODs.
#
#    putushort( 0, fout )

    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # -------------------------------------------------------------------------------
    # Tangent space basis (the mystery blocks).  
    # This is a variant location to do it here.
    # NOTE: We output floats not RGB encoded vectors.
    # -------------------------------------------------------------------------------
    print 'Starting the mystery blocks: tell(0) says ' + str( fout.tell() )
    nblocks                = 3                             # All lods have three blocks.
#    if LOD0FLAG == True :
#        nblocks            = 3
#    else :
#        nblocks            = 1
    for imystery in range( nblocks ) :
        # byte quads command string is 28 0.
        putushort( 23, fout )                              # NOTE: 23 NOT 28 because we output floats.
        putuint( serialodd, fout )                         # NOTE: backwards from regular mesh.
        if imystery == 0 :
            putuint(  3, fout )                    
        elif imystery == 1 :
            putuint( 10, fout )                    
        elif imystery == 2 :
            putuint( 11, fout )                    

        # Output the vertex count.
        putuint( nvert, fout )

        # Real float data.
        for ii in range( nvert ) :
            if imystery == 0 :
                putfloat( vnorms[3*ii+2], fout )
                putfloat( vnorms[3*ii+1], fout )
                putfloat( vnorms[3*ii],   fout )
#                putubyte(              0, fout )
            elif imystery == 1 :
                putfloat( vtans[3*ii+2], fout )
                putfloat( vtans[3*ii+1], fout )
                putfloat( vtans[3*ii],   fout )
#                putubyte(              0, fout )
            elif imystery == 2 :
                putfloat( vbinorms[3*ii+2], fout )
                putfloat( vbinorms[3*ii+1], fout )
                putfloat( vbinorms[3*ii],   fout )
#                putubyte(              0, fout )

        # This is the footer.
        putuint( 0, fout )
        putushort( 22, fout )                              # NOTE: 22 NOT 27 here for floats.
        putuint( serialeven, fout )                        # NOTE: backwards from regular mesh.
        putushort( 23, fout )                              # NOTE: 23 NOT 28 here for floats.
        putuint( serialodd, fout )                         # NOTE: backwards from regular mesh.

        # Increment serials.
        serialodd          = serialodd + 2
        serialeven         = serialeven + 2

#    # Write 10 zero bytes then the bounding sphere data.
#    putuint( 0, fout )
#    putuint( 0, fout )
#    putushort( 0, fout )
#    # First bounding sphere data written here (also at end).
#    putfloat( boundingsphere[0], fout )
#    putfloat( boundingsphere[1], fout )
#    putfloat( boundingsphere[2], fout )
#    putfloat( boundingsphere[3], fout )
#    putuint( 0, fout )                                     # Final four 0's.

    # Instead do this.
    putushort( 0, fout )
    putushort( 1, fout )
    putuint(   0, fout )  


    # -------------------------------------------------------------------------------
    # Now do the bone assignments.                                                  |
    # -------------------------------------------------------------------------------
    print 'Starting the bone assignments: tell(0) says ' + str( fout.tell() )
    # byte quads command string is 28 0.
    putushort( 28, fout )
    putushort( 1,  fout )                                  # This is one of those 1 0 code cases.
    putuint( serialodd, fout )                             # NOTE: backwards from regular mesh.
    putuint( 2, fout )                                     # A 2 0 0 0  for no apparent reason.
    putushort( 0, fout )                                   # The extra 0 0 for the 1 0 code case.
    putuint( nvert, fout )                                 # The number of vertices.

    # Write bone assignmets data.
    for ii in range( nvert ) :
        putubyte( vbones[4*ii],   fout )
        putubyte( vbones[4*ii+1], fout )
        putubyte( vbones[4*ii+2], fout )
        putubyte( vbones[4*ii+3], fout )

    # This is the footer.
    putuint( 0, fout )
    putushort( 27, fout )
    putushort( 1, fout )                                   # This is the 1 0 repeat.
    putuint( serialeven, fout )                            # NOTE: backwards from regular mesh.
    putushort( 28, fout )
    putuint( serialodd, fout )                             # NOTE: backwards from regular mesh. 
    # No weird extra granny stuff here.

    serialodd              = serialodd + 2
    serialeven             = serialeven + 2

    # NOW we put out the bounding sphere data.
    putuint(   0, fout )
    putuint(   0, fout )
    putushort( 0, fout )
    putfloat( boundingsphere[0], fout )
    putfloat( boundingsphere[1], fout )
    putfloat( boundingsphere[2], fout )
    putfloat( boundingsphere[3], fout )
    putuint(   0, fout )


    # -------------------------------------------------------------------------------
    # Bone strings section.                                                         |
    # -------------------------------------------------------------------------------
    nitems                 = len( bonedata )
    nbones                 = nitems / 3

    putuint( nbones, fout )
    putushort( 0, fout )                                   # Two skip bytes.
    for ii in range( nbones ) :
        nch                = bonedata[3*ii]
        bonename           = bonedata[3*ii+1]
        cnt                = bonedata[3*ii+2]
        putuint( nch, fout )
        fout.write( bonename )
        putuint( cnt, fout )


    # Bizarre granny string hard coded here.  Code this like I grouped it but put in right sequence numbers.
    #   3 0    1 0    86 0 0 0    4 0   1 0    0 0 0 0    1 0 0 0    0 0    39 0    1     4    87 0 0 0
    putushort( 3, fout )
    putushort( 1, fout )
    putuint( serialodd, fout )                             # NOTE: backwards from regular mesh.
    putuint( 4, fout )                                     # NOTE: 4 0 0 0 instead of 4 0 1 0 here.
#    putushort( 1, fout )
    putuint(   0, fout )
    putuint(   1, fout )
    putushort( 0, fout )
    putushort(39, fout )
    putubyte(  1, fout )
    putubyte(  4, fout )
    putuint( serialeven, fout )                            # NOTE: backwards from regular mesh. 

    putuint(  15, fout )                                   # NOTE: 15 chars for this string.
    n1                     = fnin.find( '_med' )
    n2                     = fnin.find( '_low' )
    if n1 > 0 :
        fout.write( 'siegeenginelod1' )
    elif n2 > 0 :
        fout.write( 'siegeenginelod2' )
    else :
        fout.write( 'siegeenginelod0' )

    serialodd              = serialodd + 2                 # NOTE: regular mesh increments the even one here.

    # -------------------------------------------------------------------------------
    # Really odd closing section.  All hard coded except for sequence numbers.      |
    # -------------------------------------------------------------------------------
    #  17   0   0   0  
    #45 siege_engines/textures/siege_catapult.texture
    #   0   0   0   0  
    #50 siege_engines/textures/siege_catapult_bump.texture
    #51 siege_engines/textures/siege_engine_overlay.texture
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    #   4   0   0   0     1   0   0   0     3   0   0   0    18 192  63   0     4   0   0   0   255 255 255 255   255 255 255 255  
    # 255 255 255 255   255 255 255 255     9   0   0   0     0   0   0  64     0   0   0   0     0   0   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     8   0   0   0     2   0   0   0  
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0  
    #   0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0  
    # 128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0  
    #   0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    # 128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0  
    #   0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0  
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0  
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0  
    #   0   0   0   0     0   0  38   0     1   0  21   0     0   0  39   0    20   0   0   0  
    #0.678406953812 1.1524848938 0.0618959590793 3.95701169968

    # First line.
    #  17   0   0   0  
    putuint( 17, fout )

    # Second line is first texture file with character count.
    putuint( nchars1, fout )
    fout.write( texfile1 )

    # Third line is a 0 int.
    putuint( 0, fout )

    # Fourth line is second texture file with character count.
    putuint( nchars2, fout )
    fout.write( texfile2 )

    # Fifth line is third texture file with character count.
    putuint( nchars3, fout )
    fout.write( texfile3 )

    # Sixth and seventh line are thirteen zeros.
    for ii in range( 13 ) :
        putuint( 0, fout )

    # Seventh line.
    #   4   0   0   0     1   0   0   0     3   0   0   0    18 192  63   0     4   0   0   0   255 255 255 255   255 255 255 255  
    putuint(    4, fout )
    putuint(    1, fout )
    putuint(    3, fout )
    putubyte(  18, fout )
    putubyte( 192, fout )
    putubyte(  63, fout )
    putubyte(   0, fout )
    putuint(    4, fout )
    putint(    -1, fout )
    putint(    -1, fout )

    # Eighth line.
    # 255 255 255 255   255 255 255 255     9   0   0   0     0   0   0  64     0   0   0   0     0   0   0   0     0   0   0   0  
    putint(    -1, fout )
    putint(    -1, fout )
    putuint(    9, fout )
    putushort(  0, fout )
    putubyte(   0, fout )
    putubyte(  64, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Ninth line.
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0     8   0   0   0     2   0   0   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    8, fout )
    putuint(    2, fout )

    # Tenth line.
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0  
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putshort(   0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putshort(   0, fout )

    # Eleventh line.
    #   0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Twelth line.
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0  
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )

    # Thirteenth line.
    #   0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )

    # Fourteenth line.
    # 128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0  
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )

    # Fifteenth line.
    #   0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0  
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Sixteenth line.
    # 128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0  
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putuint(    0, fout )

    # Seventeenth line.
    #   0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0     0   0   0   0   128  63   0   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )

    # Eighteenth line.
    #   0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0     0   0   0   0     0   0   2   0  
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0  
    #   0   0   0   0     0   0  38   0     1   0  21   0     0   0  39   0    20   0   0   0  
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putushort(  0, fout )
    putushort(  2, fout )

    # Nineteenth line.
    #   0   0   0   0   128  63   0   0     0   0   0   0     0   0   0   0     0   0   0   0     0   0   0   0   128  63   0   0  
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putuint(    0, fout )
    putubyte( 128, fout )
    putubyte(  63, fout )
    putushort(  0, fout )

    # Twentieth line.
    #   0   0   0   0     0   0  38   0     1   0  21   0     0   0  39   0    20   0   0   0  
    putuint(    0, fout )
    putushort(  0, fout )
    putushort( 38, fout )
    putushort(  1, fout )
    putushort(  serialodd, fout )
    putushort(  0, fout )
    putushort( 39, fout )
    putuint(    serialeven, fout )

    # Bounding sphere.
    #0.00446363957599 0.432539612055 0.793390989304 1.68704903126
    putfloat( boundingsphere[0], fout )
    putfloat( boundingsphere[1], fout )
    putfloat( boundingsphere[2], fout )
    putfloat( boundingsphere[3], fout )

    # Son of a gun, we are done!
    fout.close()

    return


# -----------------------------------------------------------------------------------
#    This def does the conversion from the binary .mesh format to the binary .ms3d format.  
# -----------------------------------------------------------------------------------
def meshtoms3dconverter( fnin, fnout, fnoutms3d ) :
    global tangentspacedatalist 
    global bytecount 
    global num_groups
    global LOD0FLAG
    global VARIANTMESHFLAG
    
    # Open the .mesh input file as read-only binary.
    fidin                  = open( fnin, 'rb' )   

    # Open the text output file for writing as ASCII.
    fidout                 = open( fnout, 'w' )
    
    bytecount              = 0
    # Read the main header.
    readmainheader( fidin, fidout )
    
    # Process the irregular first part.
    processfirstpart( fidin, fidout )
    
    # Now loop and get rest of triangle definitions.
    for ii in range( 1, num_groups ) :
        processregularpart( fidin, fidout )
    
    # Skip a 34 byte header. (Now a 26 byte header.)
    fidout.write( '\n' )
    print ' '
    outstring              = formatbytes( fidin, 26 )
    fidout.write( outstring + '\n' )
    fidout.flush()
    print outstring 
    fidout.write( '\n' )
    print ' '
    
    # Process the texture co-ordinates.
    processtexturecoordinates( fidin, fidout )
    
    # Process the vertex weights.
    processvertexweights( fidin, fidout )
    
    # Process the vertex vectors.
    processvertexvectors( fidin, fidout )

    # Pause to compute the vertex normals.
    computetangentspacebasis( fidout )
    vnorms                 = tangentspacedatalist[0]  
    vrefs                  = tangentspacedatalist[3]  
    
    if VARIANTMESHFLAG == False :
        # Process the vertex bone assignment section.
        processvertexbones( fidin, fidout )
    
        # Process the mystery block section.
        processmysteryblock( fidin, fidout )
    else :
        # Process the mystery block section.
        processvariantmysteryblock( fidin, fidout )
    
        # Process the vertex bone assignment section.
        processvertexbones( fidin, fidout )
    
    # Process the bone strings.
    processbonestrings( fidin, fidout )
    
    print 'Bytecount is now ' + str( bytecount ) 
    fidout.write( 'Bytecount is now ' + str( bytecount ) + '\n' )
    
    # Done!
    fidin.close()


    # -----------------------------------------------------------------------------------
    # Try writing to the .ms3d format. 
    # -----------------------------------------------------------------------------------
    writems3dfile( fnoutms3d, vnorms, vrefs )
   
    fidout.close()


    return


# ===================================================================================
#                            Start of main routine.                                 |
# ===================================================================================
global ms3dbytecount 
global bytecount 
global num_groups
global LOD0FLAG
global MOUNTFLAG
global VARIANTMESHFLAG
ms3dbytecount              = 0

# Get filename via a GUI.
root = Tk()
root.withdraw()
d                          = listdir.FileDialog(root)
fnin                       = d.go(os.curdir, "*.m*")
print fnin      
if fnin == None :
    exit()

# Test for a lod0 case.
nlod0                      = fnin.find( 'lod0' )
if nlod0 > 1 :
    LOD0FLAG               = True
else :
    LOD0FLAG               = False



# Check if user selected a mesh file.
nidx                       = fnin.find( ".mesh" )
if nidx > 0 :
    nchars                 = len( fnin )
    fnout                  = fnin[0:nchars-5] + '.txt'
    print 'Output text filename is ' + fnout   

    nchars                  = len( fnin )
    fnoutms3d               = fnout[0:nchars-5] + '.ms3d'
    print 'MS3D filename is ' + fnoutms3d

    # Call the main converter.
    meshtoms3dconverter( fnin, fnout, fnoutms3d )

#    # -----------------------------------------------------------------------------------
#    # Try writing out the data to an .obj file in object format as per Casuir.
#    # -----------------------------------------------------------------------------------
#    
#    # Grab output filename, remove .txt and put on .obj.
#    nchars                     = len( fnin )
#    fnoutobj                   = fnin[0:nchars-4] + '.obj'
#    print 'Object filename is ' + fnoutobj
#    
#    #writeobjfile( fnoutobj )
#    
#    # -----------------------------------------------------------------------------------
#    # Try writing to the .smd format. 
#    # -----------------------------------------------------------------------------------
#    nchars                     = len( fnin )
#    fnoutsmd                   = fnin[0:nchars-4] + '.smd'
#    print 'SMD filename is ' + fnoutsmd
#    vdata                      = computevertexnormals( fidout )
#    vnorms                     = vdata[0]
#    vrefs                      = vdata[1]
#    
#    #writesmdfile( fnoutsmd, vnorms )




# -----------------------------------------------------------------------------------
# Try reading from the .ms3d format and converting to the .mesh format. 
# -----------------------------------------------------------------------------------

nidx                       = fnin.find( ".ms3d" )
if nidx > 0 :
    nchars                  = len( fnin )
    fnouttxt                = fnin[0:nchars-5] + '_converted.txt'
    fnoutmesh               = fnin[0:nchars-5] + '_converted.mesh'
    print 'mesh output filename is ' + fnoutmesh
    print 'text output filename is ' + fnouttxt 
    nlen                    = len(fnin)
    nlastslash              = fnin.rfind( '\\' )            #
    testname                = fnin[(nlastslash+1):nlen]
    print 'testname is ' + testname 
    print 'testname = ' + testname
    if ( testname == 'catapult.ms3d' )   | ( testname == 'catapult_med.ms3d' ) | ( testname == 'catapult_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'basilisk_idle.ms3d' ) | ( testname == 'basilisk_med.ms3d' ) | ( testname == 'basilisk_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'bombard.ms3d' )  | ( testname == 'siege_bombard_med.ms3d' )  | ( testname == 'siege_bombard_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'culverin.ms3d' )  | ( testname == 'culverin_med.ms3d' )  | ( testname == 'culverin_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'bombard.ms3d' )  | ( testname == 'siege_bombard_med.ms3d' )  | ( testname == 'siege_bombard_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'grand_bombard_rig.ms3d' )  | ( testname == 'grand_bombard_med.ms3d' )  | ( testname == 'grand_bombard_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'great_cross_standard.ms3d' ) : 
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'ram_model.ms3d' )  | ( testname == 'ram_model_med.ms3d' )  | ( testname == 'ram_model_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'rocket_launcher.ms3d' )  | ( testname == 'rocket_launcher_med.ms3d' )  | ( testname == 'rocket_launcher_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_ballista.ms3d' )  | ( testname == 'siege_ballista_med.ms3d' )  | ( testname == 'siege_ballista_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_bombard.ms3d' )  | ( testname == 'siege_bombard_med.ms3d' )  | ( testname == 'siege_bombard_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_cannon.ms3d' )  | ( testname == 'siege_cannon_med.ms3d' )  | ( testname == 'siege_cannon_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_huge_bombard.ms3d' )  | ( testname == 'siege_huge_bombard_med.ms3d' )  | ( testname == 'siege_huge_bombard_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_monster_ribault.ms3d' )  | ( testname == 'siege_monster_ribault_med.ms3d' )  | ( testname == 'siege_monster_ribault_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_mortar.ms3d' )  | ( testname == 'siege_mortar_med.ms3d' )  | ( testname == 'siege_mortar_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_serpentine.ms3d' )  | ( testname == 'siege_serpentine_med.ms3d' )  | ( testname == 'siege_serpentine_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_tower_big.ms3d' )  | ( testname == 'siege_tower_big_med.ms3d' )  | ( testname == 'siege_tower_big_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_tower_small.ms3d' )  | ( testname == 'siege_tower_small_med.ms3d' )  | ( testname == 'siege_tower_small_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_tower_town_high.ms3d' )  | ( testname == 'siege_tower_town_high_med.ms3d' )  | ( testname == 'siege_tower_town_high_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'siege_trebuchet.ms3d' )  | ( testname == 'siege_trebuchet_med.ms3d' )  | ( testname == 'siege_trebuchet_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    elif ( testname == 'small_ribault.ms3d' )  | ( testname == 'small_ribault_med.ms3d' )  | ( testname == 'small_ribault_low.ms3d' ) :
        ms3dtovarmeshconverter( fnin, fnouttxt, fnoutmesh )
    else :

        # Test for a mount case.
        nmount             = fnin.find( 'mount_' )
        if nmount > -1 :
            MOUNTFLAG      = True
        else :
            MOUNTFLAG      = False

        ms3dtomeshconverter( fnin, fnouttxt, fnoutmesh )
